/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  DeSmet C is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; either version 2 of the License, or any
 *  later version.
 *
 *  DeSmet C is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/*  C88 COMPILER         C2.C */
/*
 *  Modification history:
 *    William Couture     8/26/05   Reformat/cleanup/comments
 *                                  changed 'heir' magic numbers to names
 */

#include "PASS1.H"
#include "NODES.H"
#if CHECK
#include "OBJ.H"
#endif

/* process LF (end-of-line), refill fileBuf[] as needed */
/* skipit == 0, skip leading spaces & do preprocessor */
dolf(char skipit)
   {
   char *memchr(), *mcp;
   int lsiz;

   if (cur > savEnd)  /* if parsing a macro expansion */
      {
      if (cur < macEnd)  /* if more expansion to parse */
         macxp = macfrom[nested];  /* get pointer to next piece */
      cur = nestfrom[--nested];  /* pop parse pointer */
      whitesp();
      }
   else
      {
      if (addparm == 0 && addproto == 0)  /* if not parsing "saved" function parameter list */
         cline += lineinc;
      lineinc = 0;
      newline = 1;  /* generate STMT node */
      last =
      lineBeg = ++cur;  /* remember start of line */
      if (eofptr && cur >= eofptr)  /* at EOF pointer, we're done */
         return cur = eofptr;
      while (1)
         {
         while (cur = memchr(cur, LF, bufEnd-cur))  /* if LF, a full line is in buffer */
            {
            lineinc++;
            if (*(cur-1) == '\\')  /* bug: what if LF is the 1st char? */
                           /* \LF == line continuation */
               {
               memcpy(cur-1, cur+1, (bufEnd-cur)-1);
               bufEnd -= 2;
               cur--;
               continue;
               }
            if (*(cur-2) == '\\')
                           /* \CRLF == line continuation */
               {
               memcpy(cur-2, cur+1, (bufEnd-cur)-1);
               bufEnd -= 3;
               cur -= 2;
               continue;
               }
            lineEnd = cur + 1;
            cur = lineBeg;
            if (trigraphs)  /* process trigraphs */
               {
               char *tp;

               while (tp = cur = memchr(cur, '?', lineEnd-cur))
                  {
                  if (*++cur == '?')
                     {
                     switch(*(cur+1))
                        {
                        case '=':
                           *tp = '#';
                           break;
                        case '(':
                           *tp = '[';
                           break;
                        case '/':
                           *tp = '\\';
                           break;
                        case ')':
                           *tp = ']';
                           break;
                        case '\'':
                           *tp = '^';
                           break;
                        case '<':
                           *tp = '{';  /* } */
                           break;
                        case '!':
                           *tp = '|';
                           break;
                        case '>':
/* { */                    *tp = '}';
                           break;
                        case '-':
                           *tp = '~';
                           break;
                        default:
                           continue;
                        }
                     memcpy(cur, cur+2, (bufEnd-cur)-2);
                     bufEnd -= 2;
                     lineEnd -= 2;
                     }
                  }
               cur = lineBeg;
               }
            if (skipit == 0)
               {
               while (ltype[*cur] == SPACE)  /* ^I, ^L, ^M, ' ' */
                  cur++;
               if (*cur == '#')
                  preproc();
               whitesp();
               }
            return;
            }
         /* if we are here, a full line is not in the buffer */
         if (addparm)
            mcp = addparm;
         else
            if (macproto)
               mcp = macproto;
            else
               if (addproto)
                  mcp = addproto;
               else
                  mcp = lineBeg;
         if ((lsiz = bufEnd - mcp) > 512)
            {
            error("line too long");
            real_exit(2);
            }
         else
            memcpy(&fileBuf[-lsiz], mcp, lsiz);  /* copy mcp area into lineBuf */
         lineBeg = &fileBuf[lineBeg-bufEnd];
         if (addparm)
            addparm = &fileBuf[-lsiz];
         else
            if (addproto)
               addproto = &fileBuf[-lsiz];
         if (see_exit && incnext == 0)
            bufEnd = fileBuf + see_call(0,fileBuf,2048);
         else
            bufEnd = fileBuf + read(file, fileBuf, 2048);
         if (bufEnd < fileBuf)  /* read was -1 == error */
            error("file read error");
         if (bufEnd < savEnd)  /* if at end of input file, force LF for EOL processing */
            {
            *bufEnd++ = LF;
            *bufEnd++ = LF;
            eofptr = bufEnd;
            *bufEnd++ = CONTZ;
            }
         cur = fileBuf;
         }
      }
   }

/* #pragma switches */
pswitch(char *sw, int def)
   {
   char c;

   while (ltype[*cur] == LETTER)  /* A .. Z, _, a .. z */
      cur++;  /* skip rest of pragma switch name */
   while (ltype[*cur] == SPACE)  /* ^I, ^L, ^M, ' ' */
      cur++;  /* skip any whitespace */
   if (*cur == '=')  /* i.e. #pragma warnings =
                      *      use passed default value
                      */
      *sw = def;
   else
      if (*cur == '-')  /* i.e. #pragma warnings -
                         *      turn off
                         */
         *sw = 0;
      else
         if (*cur == '!')  /* i.e. #pragma warnings !
                            *      toggle state
                            */
            *sw = !sw;
         else
            *sw = 1;  /* all others, turn on pragma */
   }


/* The current line starts with '#', feed it to the pre-processor */
preproc()
   {
   char want;

   cur++;  /* skip '#' */
   whitesp();  /* skip whitespace -- i.e. # define A */
   if (*cur == LF)  /* if no preprocessor command, we're done */
      return;
   tokit();  /* get token */
   if (heir != RESERVED && heir != UNDEF && heir != OPERAND)
      error_l("bad control");  /* not a preprocessor command oops! */
   else
      switch (string[0])
         {
         case 'a':  /* #asm */
            if (match("asm"))
               if (xkwd)  /* recognize extended keywords */
                  return cur = tokat;  /* backup so statement will find ASM */
               else
                  error("#asm option not on");
            break;
         case 'd':  /* #define */
            if (match("define"))
               return adddef();  /* add defined, return new position */
            break;
         case 'e':  /* #elif, #else, #endif, or #error */
            if (match("elif") || match("else"))
                   /* if the "#if..." was not skipped via skipsome(),
                    * so the #else (#elif) clause must be skipped.
                    */
               return skipsome(0);
            else
               if (match("endif"))
                     /* the "#if..." was not skipped via skipsome(), so we can
                      * ignore the #endif
                      */
ppexit:
                  return SKIPL();
               else
                  if (match("error"))  /* emit error message and continue */
                     {
                     char *cp;

                     cp = string;
                     while (cur < lineEnd)
                        *cp++ = *cur++;
                     while (--cp >= string && *cp <= 0x20)
                        ;
                     *(cp+1) = '\0';
                     e_msg(-1,string);
                     goto ppexit;
                     }
            break;
         case 'i':  /* #include, #if, #ifdef, #ifndef */
            if (match("include"))
               return doinc();  /* find & start parsing include file */
            else
               if (match("if"))
                  {
                  long cexpr;
ifproc:
                  mactokn = 1;  /* flag: processing a "#if" */
                  tokit();
                  cexpr = constexp();  /* evaluate expression */
                  mactokn = 0;
                  if (cexpr == 0)  /* if logically false */
                     if (skipsome(1))  /* #elif, see if THIS is true... */
                        goto ifproc;
                  goto ppexit;
                  }
               else
                  if (match("ifdef") || match("ifndef"))
                     {
                     want = match("ifdef");
                     if (ltype[*cur++] != LETTER)  /* A .. Z, _, a .. z */
                        error_l("invalid identifier in #ifdef/#ifndef");
                     else
                        {
                        find(1);  /* find(macOK == TRUE) */
                        if ((heir == DEFINED && want) ||
                                    (heir != DEFINED && want == 0))
                           goto ppexit;  /* #ifdef / #ifndef TRUE, continue processing */
                        if (skipsome(1))  /* if #elif, see if THIS is true */
                           goto ifproc;
                        goto ppexit;
                        }
                     }
            break;
         case 'l':  /* #line */
            if (match("line"))
               {
               tokit();
               if (heir != CONSTANT)
                  error_l("invalid digit-sequence in #line");
               else
                  {
                  cline = wvalue;  /* set new line number */
                  }
               whitesp();
               if (*cur == LF)
                  return;
               tokit();
               if (heir != STRNG)
                  error_l("invalid string-literal in #line");
               else
                  strcpy(name, string);  /* set new filename */
               goto ppexit;
               }
            break;
         case 'p':  /* #pragma */
            if (match("pragma"))
               {
               switch (toupper(*cur++))
                  {
                  case 'T':
                     pswitch(&trigraphs, 0);  /* process trigraphs */
                     break;
                  case 'W':
                     pswitch(&wopt, 0);  /* show warnings */
                     break;
                  case 'E':
                     if (toupper(*cur) == 'X')
                        {
                        pswitch(&xkwd, 0);  /* recognize extended keywords */
                        __stdc = !xkwd;  /* standard C = !recognize extended keywords */
                        }
                     break;
                  }
               goto ppexit;
               }
            break;
         case 'u':  /* #undef */
            if (match("undef"))
               {
               if (ltype[*cur++] != LETTER)  /* A .. Z, _, a .. z */
                  error_l("invalid identifier in #undef");
               else
                  {
                  find(1);
                  if (heir == RESERVED && bvalue >= R_LINE && bvalue <= R_STDC)
                     error_l("cannot #undef predefined names");
                  else
                     if (heir != DEFINED)
                        error_l("#undef identifier not defined");
                     else
                        (nameat-2)->byte = ' ';
                           /* we can't unlink this from the symbol table, so
                            * instead change the last char of the symbol to
                            * a space, which is never valid.
                            */
                  }
               goto ppexit;
               }
            break;
         }
   error_l("bad control");
   }

/* skip lines until the matching #endif / #elif */
skipsome(char elifOK)
   {
   char ifnest, eltype;

   ifnest = 1;  /* level of #if ... #endif nesting */
   while (ifnest)
      {
      SKIPL();  /* skip entire line */
      dolf(1);  /* skipit == 1, DO NOT skip leading spaces & DO NOT do preprocessor */
      while (ltype[*cur] == SPACE)  /* ^I, ^L, ^M, ' ' */
         cur++;  /* skip leading spaces */
      if (*cur == CONTZ)  /* EOF, we're done */
         return 0;  /* BUG12292005:  This did not return anything.  It must
                     *   return 0 else the 1-line program
                     *      #ifdef A
                     *   can cause an infinite loop, since the return was
                     *   (accidentally) non-zero, proproc() would try to
                     *   evaluate the EOF as an #elif
                     */
      if (*cur == '#' && *(cur+1) != '\r' && *(cur+1) != '\n')
                /* BUG: this does not handle a line like:
                 *         #<space>CRLF
                 *      Thus, in the 4-line program
                 *         #ifdef A
                 *         #<space>
                 *         #endif
                 *         int i;
                 *      will not be properly parsed -- the #endif will be
                 *      missed.  In fact, this would cause C88 to crash, see
                 *      BUG12292005.
                 */
         {
         eltype = 0;
         tokit();  /* skip '#' */
         tokit();  /* get token */
         if (heir == RESERVED || heir == UNDEF || heir == OPERAND)
            {
            if (match("if") || match("ifdef") || match("ifndef"))
               ifnest++;  /* skip this level of #if, too */
            else
               if (match("endif"))  /* endif, "pop" an #if level */
                  ifnest--;
               else
                  if (elifOK && match("else") && ifnest == 1)
                     ifnest = 0;  /* #else only applies to "root" #if */
                  else
                     if (elifOK && match("elif") && ifnest == 1)
                       return 1;  /* #elif of "root" #if, go process it */
            }
         }
      }
   SKIPL();  /* skip this line, too */
   return 0;  /* and we're done */
   }

match(mat)
   char *mat;
   {
   char *str;

   str = string;
   do
      if (*str != *mat++)
         return 0;
   while (*str++);
   return 1;
   }

/* process general #define, possibly with arguments.  i.e.
 *    #define MIN(a,b) (((a) < (b)) ? (a) : (b))
 */
adddef()
   {
   char *newdef,*olddef,*defat,i,lastdef, wsp = 0;

   olddef = 0;

/* do not do normal tokit as may be redefining a define */
   tokat = cur;
   curch = *cur++;
   if (ltype[curch] == LETTER)  /* A .. Z, _, a .. z */
      {
      find(1);  /* get entire symbol, see if we already know it */
      whitesp();
      if (strncmp(tokat, "defined", 7) == 0 ||
                 /* NOTE: this gives an error for
                  *          #define defined A
                  *       is that suposed to be legal/illegal?
                  */
            (heir == RESERVED && bvalue >= R_LINE && bvalue <= R_STDC))
         {
         error("Can't redefine");
         return;
         }
      if (heir == DEFINED)  /* kill old define */
         {
         *(nameat-2) = '0';  /* with zero as first letter */
                             /* NOTE: this actually zeros the LAST letter */
         olddef = nameat;
         search(1);
         }
      }
   else
      {
      cur--;  /* back to actual token start */
      tokit();  /* let regular logic handle it */
         /* NOTE: This allows something like
          *          #define 1234 5678
          *       to be legal, although it does not seem to matter.
          *       i.e.
          *          #define 1234 5678
          *          main() { printf("%d\n",1234); }
          *       produces the expected value.
          */
      }

   wp = mfree;
   *wp = machash[hashno];
   machash[hashno] = wp;  /* link in new #define symbol */
   j = 0;
   nameat = wp + 1;
   newdef = string;
   while (*nameat++ = *newdef++)
      j++;
   defat = nameat;

   defat->defcl = DEFINED;
   defat->dnlen = j + 1;
   if (blevel)  /* NOTE: I'm not sure what this is supposed to do... */
      {
      defat->dchain = hash[32];
      hash[32] = defat;
      }
   defat->dargs = 255;  /* default: no parameters in #define */
   newdef = &defat->dval;
   if (*cur == '(' && *(cur-1) != ' ' && *(cur-1) != '\11')
             /* if '(' IMMEDIATELY follows the defined symbol.  i.e.
              *    #define A(b)
              * but not
              *    #define A (b)
              * we have arguments to the #define
              */
      {
      defat->dargs = 0;  /* reset to no args */
      cur++;  /* skip the '(' */
      whitesp();  /* skip any whitespace */
      if (*cur == ')')  /* if blank argument list -- i.e. #define A() B */
         {
         cur++;  /* skip closing ')' */
         whitesp();  /* skip any whitespace */
         goto no_args;  /* and skip arg processing */
         }
      if (maxmem-mfree < 1200)
         {
         error("out of memory");
         real_exit(2);
         }
      mfree += 1000;  /* give us room for expansion */
      do
         {
         whitesp();  /* skip whitespace */
         tokat = cur;  /* save ptr to argument */
         curch = *cur++;  /* get start of argument */
         find(1);  /* see if we know this symbol already */
         whitesp();
         if (heir && heir > 4 && heir != OPERAND && heir != UNDEF)
                /* NOTE: I'm not sure what would trigger this */
            {
            error("invalid parameter");
            SKIPL();
            goto cleanup;
            }
         newname();  /* make a table entry */
         nameat->defcl = DEFPARM;  /* this is a #define argument */
         nameat->dpnum = ++defat->dargs;  /* index number of argument */
         mfree = nameat + 2;  /* set new allocation pointer */
         } while (*cur++ == ',');
      if (*(cur-1) != ')')
         {
         error("need closing parenthesis");
         SKIPL();
         goto cleanup;
         }
      }

no_args:
   lastdef = 0;
   whitesp();
   while (*cur != LF)
      {
      wsp = 0;  /* no whitespace */
      tokat = cur;
      if (*cur == CR)  /* if CR/LF, skip the CR */
         {
         cur++;
         continue;
         }
      if (*cur == '#')  /* stringification or concatination? */
         {
         if (*++cur == '#')  /* concatination */
            {
            if (!lastdef)
               {
               error("## can't be first");
               SKIPL();
               goto cleanup;
               }
            if (lastdef == DEFPARM)  /* #define A(x,y) x##y
                                      * lastdef refers to the 'x'
                                      */
               {
               *newdef = *(newdef-1);
               *(newdef-1) = DEFCPY;  /* copy literal of macro value */
               newdef++;
               }
            *newdef++ =
            lastdef = DEFPAS;  /* concatinate macro strings */
            cur++;
            whitesp();
            continue;
            }
         *newdef++ =
         lastdef = DEFSFY;  /* stringify this macro */
         whitesp();
         continue;
         }
      switch (ltype[curch = *cur++])
         {
         case LETTER:  /* A .. Z, _, a .. z */
            find(1);  /* do we already know this symbol? */
            if (heir == DEFPARM)  /* if one of our parameters */
               {
               *newdef++ = nameat->dpnum;  /* replace "name" with index number */
               lastdef = DEFPARM;  /* and this is one of our parameters */
               continue;
               }
            else
               if (heir == DEFINED && nameat == defat)
                    /* if we have our own name in this macro expansion, i.e.
                     *   #define A A
                     * treat it as a "do not macro expand this string"
                     */
                  {
                  if (lastdef == DEFSTR)
                     newdef--;  /* over-write trailing LF */
                  else
                     *newdef++ = DEFSTR;  /* else start new string expansion */
                  lastdef = DEFSTR; 
                  *newdef++ = '$';  /* flag: do not macro expand this token */
                                    /* (no recursive expansions) */
                  while (tokat < cur)  /* copy token into expansion */
                     *newdef++ = *tokat++;
                  *newdef++ = LF;  /* default: end of expansion */
                  continue;
                  }
               else
                  {
defstring:
                  if (lastdef == DEFSFY)  /* stringify */
                     {
                     error("parameter must follow #");
                     newdef--;
                     lastdef = 0;
                     goto cleanup;
                     }
                  if (lastdef == DEFSTR)
                     newdef--;  /* over-write trailing LF */
                  else
                     *newdef++ = DEFSTR;  /* else start new string expansion */
                  lastdef = DEFSTR;
                  if (wsp && *(newdef-1) != ' ')
                       /* if this is whitespace and last part does not end in whitespace */
                     *newdef++ = ' ';  /* put whitespace seperator */
                  else
                     while (tokat < cur)  /* must be normal token, add it */
                        *newdef++ = *tokat++;
                  *newdef++ = LF;
                  continue;
                  }
         case SPACE:  /* ^I, ^L, ^M, ' ' */
defspace:
            whitesp();
            if (lastdef == DEFPARM && *cur == '#' && *(cur+1) == '#')
               continue;  /* do not put in seperator if string concatination */
            wsp++;
            goto defstring;
         case 5:  /* " */
         case 8:  /* ' */
            while (*cur++ != curch)  /* find matching end quote */
               if (*(cur-1) == '\\')  /* escaped quotes do not count */
                  cur++;
            goto defstring;
         default:
            if (curch == '/' && *cur == '*')  /* if a comment */
               {
               cur--;  /* un-do parsing */
               goto defspace;  /* skip whitespace */
               }
            /* digits, symbols, etc */
            while ((curch = ltype[*cur]) != LETTER && curch != SPACE
                     && *cur != '"' && *cur != '\'' && *cur != LF && *cur != '/')
               cur++;  /* get entire batch */
            goto defstring;
         }

      }
cleanup:
   if (lastdef == DEFPAS || lastdef == DEFSFY)
      {
      error("#(#) can't be last");
      newdef--;
      }
   *newdef = DEFEND;  /* terminate #define expansion */
   mfree = newdef + 1;  /* set new mfree */
   for (i = 0; i < 32; i++)
      while (hash[i] > mfree)
         hash[i] = hash[i]->word;  /* remove "temporary" values i.e. local
                                    * placeholder names for macro values
                                    */
   }



/* add_define  --  add a 'n' option from command tail- a define. */

add_define()
   {
   char *defat,*newdef;

   tokit();  /* get name of variable */
             /* NOTE: if we already have a -n<var> of this variable name,
              *       tokit() will return the macro expansion, so we will
              *       usually not be able to re-define a symbol that way.
              *       However, if we do something like:
              *          c88 name -nA=A -nA=2
              *       the macro expansion of A will be A (and recursive
              *       to boot!), and the second -n will do undefined things
              *       due to the "infinite" macro expansion (recursive).
              */
   newname();  /* make a symbol table entry for this name */
   defat = nameat;
   defat->defcl = DEFINED;
   defat->dargs = 255;  /* no #define arguments */
   newdef = &defat->dval;
   *newdef++ = DEFSTR;  /* literal replacement */
   if (*cur == 0)  /* if no value, assume 1 */
      {
      *newdef++ = '1';
      }
   else
      if (*cur == '=')
         while (*++cur)
            *newdef++ = *cur;
               /* NOTE: this does not take "recursive definitions" into
                *       account, like adddef() does.  i.e.
                *          C88 ATEST -nA=A
                *       can do bad things...
                */
      else
         {
         os("illegal N option ");
         real_exit(2);
         }

   *newdef++ = LF;  /* EOS for expansion */
   *newdef = DEFEND;
   mfree = newdef + 1;
   }

/* set DOS allocated memory block to program size */
xsetblock()
   {
#asm
   dseg
   public _pcb_:word
oldsp rw 1
oldbp rw 1
   cseg
   mov  oldsp,sp
   mov  oldbp,bp
   mov  ax,_pcb_
   cli
   mov  ss,ax
   mov  sp,100H
   sti
   mov  bx,ds
   add  bx,1000H
   mov  es,ax
   sub  bx,es:[2CH]
   mov  ah,4AH
   int  21H
   mov  ax,ds
   cli
   mov  ss,ax
   mov  sp,oldsp
   sti
   mov  bp,oldbp
#endasm
   }


xalloc(int size)
   {
#asm
   mov  ah,48H
   mov  bx,#size
   add  bx,15
   mov  cl,4
   shr  bx,cl
   int  21H
   jnc  ok
   xor  ax,ax
ok:
#endasm
   }

doinc()
   {
   int  i, ifd;
   char inc_search;
   char inctemp[66];

   if (incnext == MAXINC)
      {
      error_l("include nesting too deep");
      return;
      }
/* accept <> around an include name */

   if (*cur != '<' && *cur != '"')
      {
      tokit();  /* expand macro */
      if (curch == '"')
         {
         inc_search = 0;  /* quote, do not search include path(s) */
         goto inc_nest;  /* name already in string[], don't parse it again */
         }
      if (curch == '<')
         {
         inc_search = 1;  /* brace, search include path(s) */
         goto inc_scan;
         }
      goto bad_inc;
      }
   if (*cur == '<')  /* brace, search include path(s) */
      inc_search = 1;
   else
      if (*cur == '"')  /* quote, do not search include path(s) */
         inc_search = 0;
      else
bad_inc:
         error_l("bad include");

   cur++;  /* skip '<' or '"' */
inc_scan:
   i = 0;
   while (*cur != '"' && *cur != '>' && i < 200)
               /* NOTE: should really search for "matching" char */
               /* NOTE: max size is 200, causes a bug below */
      {
      if (*cur == '\n' || *cur == CONTZ)  /* if EOL or EOF */
         {
         error_l("unmatched \"");
         return;
         }
      string[i++] = *cur++;
      }
   if (i == 200)
      {
      error_l("string too long");
      return;
      }
   cur++;
   string[i] = 0;
inc_nest:
   SKIPL();

/* supply a drive name of i option was specified */

   if (istring[0])  /* include path is not NULL */
      {
      if (string[1] != ':')
         strcpy(&string[100],string);
      else
         strcpy(&string[100],&string[2]);
      strcpy(string,istring);  /* copy in -i prefix */
      strcat(string,&string[100]);  /* add name to end */
         /* BUG: up to 200 chars is allowed, but we assume that we are
          *      safe to use buffer at 100.  Oops!
          */
      }

/* if MS-DOS V2.0 and use include= to find file */
   extern char _msdos2;
   if (inc_search && _msdos2)
      {
      findfile(string,inctemp);
      strcpy(string,inctemp);
      }

   if ((ifd = open(string,0)) == -1)  /* if we can't open file! */
      {
      os("cannot open ");
      os(string);
      ocrlf();
      real_exit(2);
      }
   if (incPara[incnext] == 0 && (incPara[incnext] = xalloc(SAVSIZE)) == 0)
      error_l("not enough #include buffer space");
   else
      {
      if (incnext == 0)
         lastline = cline;
      _lmove(SAVSIZE, name, _showds(), 0, incPara[incnext++]);
      eofptr =
      cline = 0;
      file = ifd;
      strcpy(name,string);
      bufEnd = savEnd;
      macname[nested] = 0;
      cur =
      nestfrom[nested] = bufEnd - 1;
      lineinc = 1;
      *cur = LF;
      dolf(0);  /* skipit == 0, skip leading spaces & do preprocessor */
      }
   }

/* handle data declaration */
/* returns node of init value (if any), otherwise 0 */
doinit()
   {
   char isstatic,*varis,*bptr;
   int  value;

   value = 0;  /* default: no value == 0 */
   if (addat->nstor == STYPEDEF || original)
              /* I don't know how original would ever be set here */
      {
      original = 0;
      return 0;
      }
   isstatic = addat->nstor <= SEXTERN;  /* SEXTER, SSTATIC, or SEXTONLY */
   if (isstatic)
      {
      ctlb(1);  /* control structure to CTEMP1:  EXTERN */
      if (is_big)  /* large memory model */
         {
         if ((was_ext || addat->nstor == SEXTERN || addat->nstor == SEXTONLY) &&
                    (addat->ntype[0] == CSTRUCT || addat->ntype[0] == ARRAY))
            ctlb(addat->nstor + 128);  /* visibility, ES seg */
         else
            ctlb(addat->nstor);  /* visibility */
         }
      else
         ctlb(addat->nstor);  /* visibility */
      ctlw(addat->noff);  /* symbol table index number */

/* make the names of statics within functions distinct */
      if (addat->nstor == SSTATIC && funname)
         {
         bptr = funname;
         while (*bptr)
            ctlb(*bptr++);  /* add function name to start of variable name */
         ctlb('_');  /* and an underscore to separate them */
         }
      ctls(addat - addat->nlen);  /* add name */
      }
   varis = &addat->ntype[0];  /* pointer to type bytes */
   locplus = 0;  /* local offset from locoff */
   if (ifch('='))  /* if we are assigning a value */
      {
      if (*varis == CSTRUCT && curch != '{')  /* } */
         {
         error("need '{' for STRUCT initilization");  /* } */
         return 0;
         }
      if (addat->nstor == SEXTONLY)
         {
         error("cannot initilize EXTERN");
         return 0;
         }
      value = initsome(varis,1,isstatic,0);
      if (isstatic)
         ctlb(INITEND);
      }
   else
      {
      if (isstatic && addat->nstor != SEXTONLY)
         {
         if (addat->ntype[0] != FUNCTION)
            {
            ctlb(INITRB);  /* uninitlized variable, just allocate space */
            ctlw(dsize(varis));
            }
         else
            ctlb(INITFUN);  /* tell GEN we have a function tree to parse */
         }
      }
   return value;  /* return init node, if any */
   }

/* varis    = ptr to type byte(s)
 * num      = number of elements to init (-1 == unspecified dimension)
 * isstatic = flag: global or local
 * initnode = node number in tree
 */
/* WARNING: recursive subroutine */
initsome(varis,num,isstatic,initnode)
   char *varis;
   int  num,isstatic,initnode;
   {
   int  asize,nodeo[4],inode,lasttree,swant;
   char *memat,sis,at,sinit,itype;

   do
      {
      if (*varis == ARRAY)
         {
         asize = (varis+1)->word;  /* get array size */
         if (asize < 0 && num < 0)
              /* multiple unsized dimensions i.e int a[][] = { { 1 }, { 2 } }; */
              /* NOTE: this prevents multiple unsigned dimensions in a row.
               *       However, something like this is legal:
               *          int a[][2][][2][] = { { 1 }, { 2 } };
               *       is this correct?
               */
            {
            error("missing dimension");
            return 0;
            }
         if (ifch('{'))  /* } */  /* array value of init */
            {
            /* allow the improper braces if string init of char */
            if (heir == STRNG && ((varis+3)->byte == CCHAR || (varis+3)->byte == CSCHAR))
               initsome(varis,1,isstatic,initnode);  /* num == 1, treat string as 1 element */
            else
               initnode = initsome(varis+3,asize,isstatic,initnode);
/* { */     notch('}');  /* make sure braces balance */
            }
         else
            if (heir == STRNG && ((varis+3)->byte == CCHAR || (varis+3)->byte == CSCHAR))
               {
               if (isstatic == 0)
                  {
                  error("sorry, no string initilization of AUTO");
                  return 0;
                  }
               if (asize < 0)
                  (varis+1)->word = 0;  /* unknown dimension, we'll calculate it */
concatstring:
               ctlb(INITSTR);
               sis = 0;
               while (string[sis])  /* sis = strlen(string); */
                  sis++;
               if (asize < 0)  /* if unsized array i.e. char a[] = "abc"; */
                  {
                  (varis+1)->word += sis;  /* update size */
                  swant = sis;
                  }
               else
                  swant = asize - 1;  /* want array size - EOS */
               if (sis > swant)  /* if string too long, only use allocated space */
                  sis = swant;
               if (swant > 255)
                  {
                  error("string too long");
                  swant = 255;
                  }
               ctlb(swant);  /* emit size */
               i = 0;
               while (i < sis)  /* put string in CTEMP1 */
                  ctlb(string[i++]);
               while (sis++ < swant)  /* put padding in CTEMP1 */
                  ctlb(0);
               tokit();  /* get next string */
               if (heir == STRNG)  /* if another adjecent string */
                  goto concatstring;  /* add it, too */
                     /* BUG: this does not work quite right in the case
                      *      of
                      *         char a[4] = "ab""cd";
                      *      what winds up being generated is
                      *         'ab',0,'cd',0,0
                      *      This is because asize is not -1 (size defined),
                      *      and so the size wanted is applied to each
                      *      individually quoted string, not the sum of
                      *      all the concatinated strings.
                      */
               ctlb(0xFF);  /* put EOS marker in CTEMP1 */
               if (asize < 0)  /* unknown array length */
                  (varis+1)->word++; /* trailing zero */
               }
            else  /* an array without an opening brace and not a quoted string
                   * i.e. int a[10] = 1;
                   */
               {
/* { */        if (curch == ',' || curch == '}' || curch == ';')
                        /* NOTE: there is no check for any elements actually
                         *       defined.  i.e.
                         *          int a[10] = ;
                         *       is legal.
                         */
                  {
                  if (num < 0)  /* if unspecified dimension */
                     {
                     bptr = varis - 3;
                     (bptr+1)->word = -1 - num;  /* set actual used size */
                     }
                  else
                     {
                     asize = dsize(varis) * num;
                     if (isstatic)
                        {
                        ctlb(INITDB);
                        ctlw(asize);
                        }
                     else
                        locplus += asize;
                     }
                  return initnode;
                  }
               initnode = initsome(varis+3,asize,isstatic,initnode);
               }
         }
      else  /* not an ARRAY */
         if (*varis == CSTRUCT)
            {
            memat = (varis+1)->word;
            memat = memat->schain;
            sinit = ifch('{');  /* } */  /* structure within structure */
/* { */     if ((curch == '}' || curch == ';') && num < 0)
               {
               bptr = varis-3;
               (bptr+1)->word = -1 - num;
               return initnode;
               }
            while (memat)
               {
               initnode = initsome(&memat->ntype[0],1,isstatic,initnode);
               memat = memat->nchain;
               ifch(',');
               }
            if (sinit)
/* { */        notch('}');
            }
         else  /* not ARRAY, not CSTRUCT */
            {
/* { */     if (curch == ',' || curch == '}' || curch == ';')  /* end of init data */
               {
               if (num < 0)  /* unspecified dimension */
                  {
                  bptr = varis-3;
                  while (*bptr != ARRAY)  /* work backword in types to find array */
                     bptr--;
                  (bptr+1)->word = -1 - num;  /* set real size */
                  }
               else
                  {
                  asize = dsize(varis) * num;
                    /* size of data element * number remaining elements */
                  if (isstatic)  /* global var */
                     {
                     ctlb(INITDB);  /* fill with 0 */
                     ctlw(asize);
                     }
                  else
                     locplus += asize;  /* update BP relative address */
                  }
               return initnode;
               }
            if (ifch('{'))  /* } */
               {
               initnode = initsome(varis,num,isstatic,initnode);
/* { */        notch('}');
               return initnode;
               }
            itype = *varis;
            if (itype == PTRTO && !is_big)  /* !large memory model */
               itype = CUNSG;
            ininit = 1;   /* needed so new variable is not allowed in init list */
            if (isstatic)
               {
               lasttree = ntree;  /* re-use space for each expression */
               inode = exprnc();  /* no commas allowed in this expression */
               ctlb(INITVAL);
               ctlb(itype);
               ctlw(lasttree-1);  /* supply place to chop off tree */
               ctlw(ntree-1);
               ctlw(inode);
               ntree = lasttree;  /* chop of the tree */
               }
            else
               {
               nodeo[0] = typeof(varis);   /* setup variable reference */
               nodeo[1] = 0;   /* local variable */
               nodeo[2] = locoff + locplus;  /* BP offset */
               locplus += dsize(varis);
               nodeo[3] = 0;
               nodeo[1] = tree4(nodeo);   /* setup tree of variable reference */
               nodeo[0] = ASGN;   /* and assign */
               nodeo[2] = exprnc();  /* the expression */
               if (initnode)  /* if already nodes created */
                  {
                  nodeo[2] = tree3(nodeo);
                  nodeo[1] = initnode;
                  nodeo[0] = LST + 2;  /* put them in node list, too */
                  }
               initnode = tree3(nodeo);  /* tree of local init (ASGN statement) */
               }
            ininit = 0;  /* no doing init anymore, new variables are OK */
            }
      if (num != 1)  /* if not last element */
         ifch(',');  /* eat comma if it exists */
                     /* NOTE: this makes a comma optional, so that
                      *           int a[2] = 1 2;
                      *       is legal.  Is this correct?
                      */
      } while (--num);  /* while elements to process */
   return initnode;
   }



/* FINDFILE.C */
/*
   This file contains the routine to locate a file, utilizing the INCLUDE
   environment variable for the directories to search.

   Interface:

      findfile(filename, target_buf)

      where filename is the name of the file to be searched for,
            target_buf is the buffer to place the path name if found.

      if the file is found, findfile return 1 and the pathname,
      otherwise it returns 0.

   This program uses the environ routines to access the PATH variable.

   Stack requirements:  ~300 bytes

*/

findfile(filename, target_buf)
   char *filename, *target_buf;
   {
   int fid;
   char paths[256], *p_ptr, *t_ptr;

   /* first check in the local directory */
   strcpy(target_buf, filename);
   fid = open(target_buf, 0);
   if (fid >= 0)  /* got it */
      {
      close(fid);
      return (1);
      }
   fid = _environ("DSINC", paths, 256);
      /* NOTE: _environ() at least clears the first byte of paths */
   if (fid == -1)
      fid = _environ("INCLUDE", paths, 256);
   p_ptr = paths;

   while (*p_ptr != 0)  /* copy the directory name */
      {
      t_ptr = target_buf;
      while (*p_ptr != ';' && *p_ptr != 0)
         {
         *t_ptr++ = *p_ptr++;
         }
      if (*(t_ptr-1) != '/' && *(t_ptr-1) != '\\')
            /* if path does not end in '/' or '\' */
         *t_ptr++ = '\\';  /* add path seperator */
      *t_ptr = 0;
      if (*p_ptr)  /* beyond the ';' */
         p_ptr++;
      strcat(target_buf, filename);
      fid = open(target_buf, 0);
      if (fid >= 0)  /* got it */
         {
         close(fid);
         return (1);
         }
      }
   strcpy(target_buf, filename);
   return (0);  /* can't find one */
   }



/* ENVIRON.C */

/*
   This file contains the routine for searching the environment
   area for a given string.

   Interface:

      _environ(search_str, buffer, buf_len)

      where search_str is the actual string being searched for,
            buffer is the location to store the string if found,
            buf_len is the maximum number of characters to store
                    into the buffer (includes terminator).

        -1 is returned if the environment variable isn't found,
        otherwise the length of the string copied is returned.
*/

extern unsigned _pcb;  /* c88 variable with the PSP segment */

int _environ(search_str, buffer, buf_len)
   char *search_str, *buffer; int buf_len;
   {
#asm
   mov      ax, _pcb_;            ; get the env segment into ES
   mov      es, ax                ; 
   mov      ax, es:[2CH]
   mov      es,ax

   mov      bx, [6][bp]           ; buffer
   mov      byte[bx],0
   mov      si, 0                 ; offset into the environment
   jmp      outer_begin

outer_loop:
   inc      si
outer_begin:
   mov      di, [4][bp]           ; search_str
   cmp      byte es:[si],0
   jne      cmp_loop
                                  ; not found
   mov      ax,-1
   pop      bp
   ret

cmp_loop:
   mov      al, byte es:[si]      ; pick up the search character
   cmp      al,'='
   je       end_str
   cmp      al, byte[di]
   jne      outer_next
   inc      si
   inc      di
   jmp      cmp_loop

end_str:
   cmp      byte[di], 0
   jne      outer_next
                                  ; got it, copy them bytes!
   mov      cx, [8][bp]           ; buf_len
   inc      si                    ; beyond the '='

move_loop:
   mov      al, byte es:[si]
   or       al,al
   jz       now_return
   dec      cx
   jz       now_return
   mov      byte[bx],al
   inc      si
   inc      bx
   jmp      move_loop

now_return:
   mov      byte[bx],0
   mov      ax, [8][bp]
   sub      ax, cx
   pop      bp
   ret

outer_next:
   cmp      byte es:[si],0        ; skip to the next entry
   je       outer_loop
   inc      si
   jmp      outer_next
#endasm
}
