/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  DeSmet C is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; either version 2 of the License, or any
 *  later version.
 *
 *  DeSmet C is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/* c3.c         third pard of medium c */
/*
 *  Modification history:
 *    William Couture    10/11/05   Reformat/cleanup/comments
 *                                  changed 'heir' magic numbers to names
 */

#include "pass1.h"
#include "nodes.h"
#if CHECK
#include "OBJ.H"
#endif


/* recursive via compound() */
/* returns a node number */
statement()
   {
   int value,stmtnum,nodeo[3];

   laststmt = 0;
   value =
   stmtnum = 0;
   if (specs(SAUTO))  /* if this is a variable definition */
      value = allocloc();
             /* Note:   Hmmm... this allows statement sequences like:
              *             main(int i){ if (i) int j; j = 1; }
              *         to be legal.  Is that right?
              */
   else
      {
      /* remember the statement number */
      if (copt && newline)  /* add checkout info to object file */
         stmtnum = cline;
      newline = 0;

      if (heir == RESERVED)  /* reserved keyword, probably a C keyword */
         switch (bvalue)
            {
            case RBREAK:
               value = dobreak();
               break;
            case RCONTINUE:
               value = docont();
               break;
            case RDO:
               value = dodow();
               laststmt = 0;
               break;
            case RFOR:
               value = dofor();
               laststmt = 0;
               break;
            case RGOTO:
               value = dogoto();
               break;
            case RIF:
               value = doif();
               laststmt = 0;
               break;
            case RRETURN:
               value = doreturn();
               laststmt = RET;
               break;
            case RSWITCH:
               value = doswitch();
               laststmt = 0;
               break;
            case RWHILE:
               value = dowhile();
               laststmt = 0;
               break;
            case RCASE:
               value = docase();
               laststmt = 0;
               break;
            case RDEFAULT:
               value = dodflt();
               break;
            case RASM:
               value = doasm();
               break;
            default:  /* sizeof, __LINE__, etc */
               error("bad statement");
            }
      else
         if (ifch(';'))  /* null statement */
            stmtnum = 0;
         else
            if (ifch('{'))  /* } */  /* start of compound statement */
               {
               value = compound(mfree,0);  /* parse { <statment; ...> } */
               laststmt = 0;
               }
            else
               if (*cur == ':')  /* a label */
                  value = dolabel();
               else
                  {
                  value = expr();  /* must be an expression */
                  notch(';');
                  }
      }

/* output a chaining STMT number node */

#if CHECK
   if (stmtnum)
      {
      nodeo[0] = STMT;
      nodeo[1] = value;
      nodeo[2] = stmtnum;
      value = tree3(nodeo);
      }
#endif
   return value;
   }

/* returns a node number */
/* do <statement> while (expr); */
dodow()
   {
   int nodeo[3],value;

   nodeo[0] = DOW;  /* build dow_node structure (see nodes.h) */
   tokit();
   nodeo[1] = statement();
   if (heir != RESERVED || bvalue != RWHILE)
      {
      error("missing while");
      return 0;
      }
   nodeo[2] = pexpr();  /* expression must be in parenthesis */
   notch(';');  /* and followed by a ';' */
   value = tree3(nodeo);
   tree3z();  /* and follow with 3 0's */
   return value;
   }

/* returns a node number */
/* goto <label>; */
dogoto()
   {
   int nodeo[4];

   tokit();  /* get target label */
   deflab();  /* get ptr to previous label info, or make a new one */
   if (nameat->ntype[0] != CLABEL)  /* if symbol is not a label */
      {
      error("need label");
      return 0;
      }
         /* build opnd_node struct (see nodes.h) */
   nodeo[0] = CLABEL << 8;  /* ntyp == OPND, dtyp == CLABEL */
   nodeo[1] = nameat->noff;  /* symbol table entry number */
   nodeo[2] =
   nodeo[3] = 0;   /* not in structure/union */
         /* build gotos_node struct (see nodes.h) */
   nodeo[1] = tree4(nodeo);
   nodeo[0] = GOTOS;
   tokit();  /* skip the label */
   notch(';');  /* make sure this is ';' */
   return tree2(nodeo);
   }

/* returns a node number */
/* if <expression> <statement> [else <statement>] */
doif()
   {
   int nodeo[4];

   nodeo[0] = IFE;  /* build an ife_node struct (see nodes.h) */
   nodeo[1] = pexpr();  /* make sure expression is in parenthesis */
   nodeo[2] = statement();  /* parse the statement */
   if (heir == RESERVED && bvalue == RELSE)  /* if statement followed by 'else' */
      {
      tokit();  /* skip the "else" */
      nodeo[3] = statement();  /* set else node in ife_node struct */
      }
   else
      nodeo[3] = 0;  /* no else, use null node */
   return tree4(nodeo);
   }

/* returns a node number */
/* return <expression> ; */
doreturn()
   {
   int nodeo[2], rnode[4];

   tokit();
   if (ifch(';'))  /* if <expresion> is NULL */
      {
      if (fun_ret_len)  /* if function returns a structure, this is its size */
         error("must return structure");
      else
         if (funtype >= CFLOAT)  /* float or double */
            error("must return float");
      nodeo[1] = 0;  /* set NULL return value */
      }
   else
      {
      heir11(nodeo);  /* get tree of expression */
      if (nodeo[0] == 0)  /* return (); or some such */
         error("return lacks argument");
      notch(';');  /* make sure <expression> is followed by ';' */
      if (fun_ret_len)  /* if structure return */
         {
         if (wopt)  /* show warnings */
            warning("returns structure");
         /* construct move to return structure */

         struct_type = nodeo[1];
         if (struct_type->sttype != CSTRUCT ||
                          struct_type->ststagat->staglen != fun_ret_len)
            error("must return structure");
               /* build opnd_node (see nodes.h) for passed return address */
         rnode[0] = OPND + (is_big ? (PTRTO << 8) : (CUNSG << 8));  /* large memory model */
         rnode[1] =
         rnode[3] = 0;  /* not in structure/union */
         rnode[2] = is_big ? 6: 4;  /* put structure at [bp+4] or [bp+6] for large memory model */
                    /* now output the move */
         rnode[2] = tree4(rnode);
            /* build math_node for move operand (see nodes.h) */
         nodeo[1] = nodeo[0];  /* take address of source */
         nodeo[0] = TA + (is_big ? (PTRTO << 8) : (CUNSG << 8));  /* large memory model */
                     /* TA == Take Address (use address of, not reference to) */
         rnode[1] = tree2(nodeo);
         rnode[3] = fun_ret_len;
         rnode[0] = MOVE;  /* build a move_node struct */
         nodeo[1] = tree4(rnode);
         }
      else
         {
         nodeo[1] = nodeo[0];
         nodeo[0] = CAST + (funtype << 8);
         nodeo[1] = tree2(nodeo);
         }
      }
   nodeo[0] = RET;
   return tree2(nodeo);
   }

/* returns a node number */
/* while (expression) <statement> */
dowhile()
   {
   int nodeo[3], value;

   nodeo[0] = WHIL;  /* build whil_node structure (see nodes.h) */
   nodeo[1] = pexpr();  /* expression must be inparenthesis */
   nodeo[2] = statement();
   value = tree3(nodeo);
   tree3z();
   return value;
   }

/* returns a node number */
/* for (exp1; exp2; exp3) <statement> */
dofor()
   {
   int nodeo[5], value;

   nodeo[0] = FORS;  /* build fors_node structure (see nodes.h) */
   tokit();
   if (notch('('))  /* must have opening parenthesis */
      return 0;
   nodeo[1] = expr();  /* expr1 */
   if (notch(';'))  /* must have separating semi-colon */
      return 0;
   nodeo[2] = expr();  /* expr2 */
   if (notch(';'))  /* must have separating semi-colon */
      return 0;
   nodeo[3] = expr();  /* expr3 */
   if (notch(')'))  /* must have closing parenthesis */
      return 0;
   nodeo[4] = statement();  /* and statement */
   value = treev(5,nodeo);
   tree3z();
   return value;
   }

/* returns a node number */
/* label : <statement> */
dolabel()
   {
   int nodeo[3];

   deflab();  /* get ptr to previous label info, or make a new one */
   if (nameat->nstor != SUNSTOR)  /* if not previously undefined */
      {
      error("duplicate label");  /* oops!  already exists! */
      return 0;
      }
   nodeo[0] = LAB;  /* build lab_node structure (see nodes.h) */
   nameat->nstor = SSTATIC;  /* change storage type */
   nodeo[2] = nameat->noff;  /* index into symbol table */
   tokit();  /* skip label */
   tokit();  /* skip ':' */
   nodeo[1] = statement();  /* label must refer to a statement,
                             * even if it is NULL i.e. just a ';'
                             */
   return tree3(nodeo);
   }

/* get ptr to previous label info, or make a new one */
deflab()
   {
   char *bptr;

   for (wp = labhash[hashno]; wp; wp = *wp)
      {
      char ch;

      nameat = wp + 1;
      bptr = string;
      while ((ch = *nameat++) && ch == *bptr++)  /* while not EOS & string matches */
         ;
      if (ch == 0 && *bptr == 0)  /* both at EOS? */
         return;  /* yes, we have our pointer */
      }
   wp = mfree;  /* time to build new label */
   *wp = labhash[hashno];
   labhash[hashno] = wp;
   nameat = wp + 1;
   bptr = string;
   while (*nameat++ = *bptr++)
      ;
   nameat->nchain = hash[32];  /* keep labels */
   hash[32] = nameat;
   nameat->opcl = OPERAND;
   nameat->nlen = bptr - string;
   nameat->nstor = SUNSTOR;   /* flag: have not defined this label yet */
   nameat->noff = ++ordinal;
   nameat->ntype[0] = CLABEL;
   ctlb(1);  /* a label needs it's own symbol table entry, emit a EXTERN to CTEMP1 */
   ctlb(SSTATIC);  /* visibility = STATIC */
   ctlw(ordinal);  /* symbol table index number */
   /* add function name to label name */
   bptr = funname;
   while (*bptr)
      ctlb(*bptr++);
   ctlb('_');
   ctls(mfree+2);
   ctlb(INITFUN);
   mfree = &nameat->ntype[1];
   }

/* returns a node number */
/* NOTE: This code (& GEN) will compile the following code:
 *            int i,j; main() { switch (i) case 0: j = 2; }
 *       is this correct?
 */
/* switch (expression) <statement(s)> */
doswitch()
   {
   int nodeo[3];

   nodeo[0] = SWIT;  /* build a swit_node structure (see nodes.h) */
   nodeo[1] = pexpr();  /* expression must be in parenthesis */
   nodeo[2] = statement();  /* statement(s), hopefully with case statements */
      /* NOTE: GEN notes the lack of CASE statements, not C88 */
   return tree3(nodeo);
   }

/* returns a node number */
/* break ; */
dobreak()
   {
   int value,nodeo[1];

   nodeo[0] = BRK;  /* build brk_node structure (see nodes.h) */
   value = tree1(nodeo);
   tokit();  /* skip "break" */
   notch(';');  /* must be followed by ';' */
   return value;
   }

/* returns a node number */
/* continue ; */
docont()
   {
   int value, nodeo[1];

   nodeo[0] = CONT;  /* build cont_node structure (see nodes.h) */
   value = tree1(nodeo);
   tokit();  /* skip "continue" */
   notch(';');  /* must be followed by ';' */
   return value;
   }

/* returns a node number */
/* case <value> : */
/* NOTE: GEN checks to make sure this is in a SWITCH statement, not C88 */
docase()
   {
   int nodeo[3], minus, beg, end;

   tokit();  /* skip "case" */
   minus = 0;  /* negative flag = FALSE */
   if (heir == 22 && curch =='-')  /* if leading minus sign */
      {
      minus = 1;  /* remember this is negative */
      tokit();  /* skip the '-' */
      }
   if (heir != CONSTANT && !(heir == RESERVED && bvalue == RSIZEOF) && curch != '(')
       {
       error("need constant");
       return 0;
       }
   beg = constexp();  /* get value of constant expression */
   if (minus)  /* remember minus sign, if any */
      beg = -beg;
   if (ifch(RDOTDOT))  /* if "case x..y" */
      {
      minus = 0;  /* minus sign for 2nd value */
      if (heir == 22 && curch =='-')
         {
         minus = 1;
         tokit();
         }
      if (heir != CONSTANT && curch != '(')
          {
          error("need constant");
          return 0;
          }
      end = constexp();
      if (minus)
         end = -end;
      if (beg > end)  /* if "out of order" */
         {
         minus = end;  /* re-order them */
         end = beg;
         beg = minus;
         }
      }
   else
      end = beg;
   if (notch(':'))  /* must be followed by ':' */
      return 0;
   nodeo[0] = CAS;  /* build cas_node structure (see nodes.h) *
   nodeo[1] = statement();  /* must be followed by statement */
   while (beg < end)  /* if case range */
      {
      nodeo[2] = beg++;  /* multiple CASE statements reference this node */
      nodeo[1] = tree3(nodeo);
      }
   nodeo[2] = end;
   return tree3(nodeo);
   }

/* returns a node number */
/* default : <statement> */
/* NOTE: GEN checks to make sure this is in a SWITCH statement, not C88 */
dodflt()
   {
   int nodeo[2];

   nodeo[0] = DFLT;  /* build dflt_node structure (see nodes.h) */
   tokit();  /* skip "default" */
   if (notch(':'))  /* must be followed by ':' */
      return 0;
   nodeo[1] = statement();  /* and by a statement */
   return tree2(nodeo);
   }

/* output some assembler */
/* returns a node number */
doasm()
   {
   int nodeat;

   have_asm = 1;  /* flag GEN: need to invoke ASM88 to handle assembly,
                   *           do not generate .O file directly
                   */
   nodeat = OUTASM;
   nodeat = tree1(&nodeat);  /* node for embedded assembler */
           /* NOTE: tree1() expects a POINTER, not a value */
   SKIPL();  /* skip rest of #asm line */
   while (1)
      {
      dolf(1);  /* skipit == 1, DO NOT skip leading spaces & DO NOT do preprocessor */
      if (*cur == '#' || *cur == CONTZ)  /* time to end assembly! */
         {
         if (*cur == CONTZ)  /* EOF, that's a no-no */
            error("missing end of #asm");
         treew(-1);  /* write ntyp == 255 and dtyp == 255 to flag end-of-assembly */
         SKIPL();  /* skip rest of line (so that, for example, #endasm works
                    * as well as just #
                    */
         tokit();  /* re-prime parse pump */
         return nodeat;
         }
      while (*cur != LF && *(cur+1) != LF && *cur != CONTZ &&
                      *(cur+1) != CONTZ)
         {
         if (*cur != '#' && *(cur+1) != '#')  /* this word will not contain a '#' */
            {
putwrd:     wp = cur;
            treew(*wp);
            cur += 2;
            }
         else  /* a '#' is a reference to a variable, put in it's address */
            {
            char locname[64], *ep = cur;
            int n = 0;

            if (*cur != '#')  /* if 1st char is not '#', just pass it through */
               locname[n++] = *cur++;
            if (ltype[*(cur+1)] != LETTER)  /* A .. Z, _, a .. z */
                     /* if not a symbol, pass it through */
               {
               cur = ep;
               goto putwrd;
               }
            cur += 2;  /* update pointer */
            find(1);  /* if a symbol */
            if (heir == DEFINED && nameat->dargs == 255)
                    /* #define with no args, i.e.
                     *    #define A 1
                     * not
                     *    #define MIN(a,b) ((a) < (b) ? (a) : (b))
                     */
               {
               char *dp = &nameat->dval+1, suffix = 0;

               while (*dp != LF)
                  {
                  char c;

                  if ((c = ltype[*dp]) == DIGIT || c == LETTER)  /* A .. Z, _, a .. z */
                     {
                     if (*dp == '0')  /* if this starts with '0' */
                        {
                        locname[n++] = '0';
                        if (toupper(*++dp) == 'X')  /* if this is a hex number */
                           {
                           dp++;  /* skip the 'x' */
                           suffix = 'H';  /* add 'H' to represent hex */
                           while (ltype[c = *dp] == DIGIT
                                    || ((c=toupper(c)) >= 'A' && c <= 'F'))
                              locname[n++] = *dp++;  /* copy hex digits */
                           }
                        else
                           if (ltype[*dp] == DIGIT)  /* if this is octal */
                              {
                              suffix = 'O';  /* add 'O' to represent octal */
                              while (ltype[*dp] == DIGIT)  /* copy octal digits */
                                   /* BUG:  should not copy digits > '7' */
                              locname[n++] = *dp++;
                              }
                           else
                              continue;
                        if (suffix)  /* if we need to add a char */
                           {
                           locname[n++] = suffix;
                           suffix = 0;
                           }
                        }
                     else
                        while ((c = ltype[*dp]) == DIGIT || c == LETTER)  /* A .. Z, _, a .. z */
                           locname[n++] = *dp++;  /* copy string */
                     }
                  else
                     locname[n++] = *dp++;  /* copy this char */
                  }
               goto putloc;  /* finish, make sure number of chars is EVEN */
               }
            if (heir != OPERAND && nameat->nstor != SAUTO)
                   /* global symbol, pass it through */
               {
               cur = ep;
               goto putwrd;
               }
            locname[n++] = '[';  /* local symbols are BP relative */
            locname[n++] = 'B';
            locname[n++] = 'P';
            if (nameat->noff >= 0)  /* if passed on stack */
               locname[n++] = '+';  /* location is BP+xxxx */
                  /* else itoa() will automatically add the negative sign */
            itoa(nameat->noff, &locname[n], 10);
            while (locname[++n])  /* skip name we just added */
               ;
            locname[n++] = ']';  /* finish BP relative addressing */
putloc:     if (n & 1)  /* if odd number of chars (not a full WORD) */
               locname[n++] = ' ';  /* pad it with a space */
            wp = locname;  /* write it out as WORDs */
            while (n)
               {
               treew(*wp++);
               n -= 2;
               }
            }
         }
      if (*cur == LF)  /* only LF left (1st char)? */
         treew(LF);  /* write LF, <null> */
      else
         {
         wp = cur;  /* write chars */
         treew(*wp);
         if (*cur != CONTZ)  /* if not Ctrl-Z, skip char */
            cur++;
         }
      }
   }

static char undef_msg[40] = "undefined - ";
   /* NOTE: is 40 enough?  What is the maximum label length? */

/* parse { <statment; ...> } */
/* recursive via statement() */
/* returns a node number */
compound(oldfree,isproc)
   char *oldfree,isproc;
   {
   int list[7], value, *zero;
   char n, *lastfun, *lowfun, *last_move, **mpp, *mp;

   blevel++;  /* count opening curly brace */
   n = 0;
   list[1] = 0;
       /* { -- for SEE's curly brace matching */
   while (curch != '}' && curch != CONTZ)
      {
      if (value = statement())
         {
         list[++n] = value;
         if (n == 6)
            {
            list[0] = LST+6;
            list[1] = treev(7,list);
            n = 1;
            }
         }
      }
   if (n <= 1)
      value = list[1];
   else
      {
      list[0] = LST + n;
      value = treev(n+1,list);
      }
   if (mfree > oldfree)  /* if locals allocated, forget them */
      {
      if (mfree >= maxmem)
         {
         error("out of memory");
         real_exit(2);
         }
      if (mfree > hiwater)
         hiwater = mfree;
      for (n = 0; n < 32; n++)
         {
         for (wp = hash[n]; wp >= oldfree; wp = *wp)
            ;
         hash[n] = wp;
         for (wp = labhash[n]; wp >= oldfree; wp = *wp)
            ;
         labhash[n] = wp;
         for (wp = machash[n]; wp >= oldfree; wp = *wp)
            ;
         machash[n] = wp;
         }
      lastfun = hash[32];
      /* clean off the function chain */
      while (hash[32] > oldfree)
         hash[32] = hash[32]->nchain;
      while (lastfun > oldfree)
         {
         wp = &lastfun;
         while (((char *)*wp)->nchain > oldfree)
            wp = &((char *)*wp)->nchain;
         if ((lowfun = *wp) == 0)
            break;
         *wp = 0;
         /* abandon labels if exiting a procedure */
         if (isproc && lowfun->ntype[0] == CLABEL)
            {
            if (lowfun->nstor == SUNSTOR)  /* if referenced label was never defined */
               {
               strcpy(&undef_msg[12],lowfun-lowfun->nlen);
               error(undef_msg);
               }
            continue;
            }
         i = lowfun->nlen;
         wp = oldfree + 2 + i;  /* where OPERAND will go */
         hashno = i - 1;
         bptr = lowfun - 1;
         while (--i)
            hashno += *--bptr;
         hashno &= 31;
         if (lowfun->opcl == DEFINED)
            {
            oldfree->word = machash[hashno];
            machash[hashno] = oldfree;
            }
         else
            if (lowfun->ntype[0] == CLABEL)
               {
               oldfree->word = labhash[hashno];  /* rebuild label chain */
               labhash[hashno] = oldfree;
               }
            else
               {
               oldfree->word = hash[hashno];  /* rebuild hash chain */
               hash[hashno] = oldfree;
               }
         oldfree += 2;
         last_move = &lowfun->ntype[0];
         if (lowfun->opcl == DEFINED)
            while (*last_move != DEFEND)
               last_move++;
         else
            while (*last_move >= FUNCTION)
               {
               if (*last_move == ARRAY || *last_move == CSTRUCT)
                  last_move += 3;
               else
                  last_move++;
               }
         while (bptr <= last_move)
            *oldfree++ = *bptr++;
         ((void *)wp)->nchain = hash[32];  /* rebuild function hash chain */
         hash[32] = wp;
         }
      mfree = oldfree;
      }
   retnum = cline;
       /* { -- for SEE's curly brace matching */
   notch('}');
   blevel--;  /* count closing curly brace */
   return value;
   }

/* allocate local variable */
/* returns a node number -- init value node */
allocloc()
   {
   char islocal, realstor;
   int value, newvalue, nodeo[3];

   value = 0;
   islocal = addstor == SAUTO;  /* local variable? */
   if (ifch(';'))  /* null definition.  i.e. int ; */
      return 0;
   if (locoff & 1)  /* make sure variables are allocated on WORD boundaries */
      locoff--;
   findover = mfree - 1;
       /* allow local variable names to over-ride more global variable names */
   do
      {
      if (addvar(0) == 0)  /* if invalid definition */
         {
         error("bad declaration");
         findover = 0;
         return 0;
         }
      addproto =   /* forget function prototypes / parameters */
      addparm = 0;
      if (addat->ntype[0] == FUNCTION)  /* declaring a function.  i.e. long lseek(); */
         {
         realstor = addstor;  /* remember real storage but fix function */
         if (addstor != SSTATIC)
            addstor = addat->nstor = SEXTERN;
         atype(addtype);
         addstor = realstor;
#if CHECK
         if (copt)  /* add checkout info to object file */
            objtype(OPTYPE,addat->noff);
#endif
         }
      else
         {
         was_ext = 0;
         atype(addtype);  /* add type bytes to symbol table */
         if (islocal)
            {
            locoff -= dsize(&addat->ntype[0]);  /* adjust local stack used */
            addat->noff = locoff;  /* stack-based location for local */
            }
#if CHECK
         if (copt)  /* add checkout info to object file */
            {
            if (addstor == SEXTERN || addstor == SSTATIC)
               objtype(OPTYPE,addat->noff);
            else
               if (islocal)
                  {
                  objtype(OLTYPE,locoff);
                  }
            }
#endif
         if (newvalue = doinit())  /* any variable initalizaton? */
            {
            if (value)  /* if we already have an initalization */
               {
               nodeo[0] = LST+2;  /* setup init tree of previous initalization */
               nodeo[1] = value;
               nodeo[2] = newvalue;
               value = tree3(nodeo);  /* remember this tree branch */
               }
            else
               value = newvalue;  /* for later init */
            }
         }
      } while (ifch(','));
   findover = 0;  /* end of locals definitions */
   notch(';');  /* syntax check */
   return value;  /* return init node, if any */
   }

/* parse an expression and build a tree.
 * This is the "root function" for an expression
 */
expr()
   {
   int enode[2];
      /* NOTE: enode[0] == root node of expression
       *       enode[1] == ptr to type of expression.  Note that this
       *                   assumes sizeof(int) == sizeof(ptr)
       */

   heir11(enode);  /* check for comma in expression */
   return enode[0];
   }

exprnc()  /* 'nc' -- no commas allowed (no compound statements) */
   {
   int enode[2];
      /* NOTE: enode[0] == root node of expression
       *       enode[1] == ptr to type of expression.  Note that this
       *                   assumes sizeof(int) == sizeof(ptr)
       */

   heir12(enode);  /* skip comma heirarchy level, start with opeq (+=, etc) */
   if (enode[0] == 0)  /* if no expression */
      error("missing expression");
   return enode[0];
   }

/* parenthesized expression */
pexpr()
   {
   int  value;

   tokit();  /* get this token */
   if (ifch('('))  /* if a parenthesis */
      {
      value = expr();  /* normal expression parsing */
      if (ifch(')'))  /* if ends with a ')' */
         return value;  /* all is OK */
      }
   error("need ()");
   return 0;
   }

ifch(ch)
   char ch;
   {
   if (curch == ch)  /* if current char is correct */
      {
      tokit();  /* get next token */
      return 1;  /* and report OK */
      }
   return 0;  /* else report not our char */
   }

notch(ch)
   char ch;
   {
   char *msg;

   if (curch == ch)  /* if our char */
      {
      tokit();  /* get next token */
      return 0;  /* report no error */
      }
   msg = "missing  ";  /* oops! */
   *(msg + 8) = ch;
   error(msg);
   return 1;  /* report missing char */
   }
