/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  DeSmet C is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; either version 2 of the License, or any
 *  later version.
 *
 *  DeSmet C is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/* C4.C               PART4 OF MEDIUM C COMPILER */
/*
 *  Modification history:
 *    William Couture    10/11/05   Reformat/cleanup/comments
 *                                  changed 'heir' magic numbers to names
 */

#include "pass1.h"
#include "nodes.h"

/* get a constant expression */

/* uses toknlf instead if tokit. this does not allow a constant
   expression to span a line. */

toknlf()
   {
   do
      {
      if (*cur == LF && cur < savEnd)
         {
         heir = 0;
         curch = LF;
         break;
         }
      } while (tokone());
   }

/* recursive via constexp() call */
long heir25c()  /* must have constant or left parenthesis */
   {
   long value, constexp();  /* forward reference */
   char needrp, savestor, savenest;
   unsigned backup, saveat, saveloc, savetype;
   int temptree[30], ointree, ontree, omfree, node[2];

   if (heir == RESERVED && bvalue == RSIZEOF)  /* sizeof() operator? */
      {
      if (mactokn)  /* not for #if */
         {
         error_l("sizeof operator not allowed in #if/#elif");
         while (curch != ')')
            toknlf();
         return 0;
         }
      tokit();
      needrp = 0;
      backup = cur;
      savenest = nested;
      /* save add information in case dimentioning a variable */
      savestor = addstor;
      savetype = addtype;
      saveat = addat;
      saveloc = addloc;
      if (ifch('('))  /* if a parenthesis */
         {
         if (specs(SUNSTOR))  /* if a data type, not variable name */
                /* NOTE: this is using the "typecast" logic to implement
                 *       sizeof(datatype)
                 */
            {
            omfree = mfree;
            mfree += 20;
            value = dsize(abstract(0));  /* get size of abstract data type */
            mfree = omfree;
            addstor = savestor;
            addtype = savetype;
            addat = saveat;
            addloc = saveloc;
            return value;
            }
         else
            {
            cur = backup;
            nested = savenest;
            tokit();  /* get variable name */
            needrp = 1;  /* we have to take care of tailing parenthesis */
            }
         }
/* if SIZEOF, do not allow a logical expression. */
/* do not output any to tree if SIZEOF */
      ointree = intree;
      intree = temptree;
      ontree = ntree;
      heir24(node);  /* do sizeof processing from expression expansion */
      addstor = savestor;  /* restore processing */
      addtype = savetype;
      addat = saveat;
      addloc = saveloc;
      intree = ointree;
      ntree = ontree;
      value = dsize(node[1]);  /* get size of type */
      if (needrp)
         {
         if (curch == ')')  /* if right character, parse without skipping LF */
            toknlf();
         else
            notch(')');  /* can't use this for "true", it will skip LF */
         }
      return value;
      }

/* look for a parened constant expression */

   if (ifch('('))  /* if a parenthesis */
      {
      value = constexp();  /* re-parse everything in this parenthesis */
      if (curch == ')')  /* if right character, parse without skipping LF */
         toknlf();
      else
         notch(')');  /* can't use this for "true", it will skip LF */
      return value;
      }

   if (mactokn && strcmp(string, "defined") == 0)  /* #if can have defined check */
      {
      if (needrp = (curch = *cur++) == '(')  /* if opening paren, need closing too */
         {
         whitesp();
         curch = *cur++;
         }
      if (ltype[curch] == LETTER)  /* A .. Z, _, a .. z */  /* could be a symbol */
         {
         find(1);  /* get symbol, see if we know it */
         whitesp();
         value = (heir == DEFINED || (heir == RESERVED && bvalue >= R_LINE
                                    && bvalue <= R_STDC));
         heir = LCONSTANT;
         if (needrp)
           {
            toknlf();
            if (curch == ')')  /* if right character, parse without skipping LF */
               toknlf();
            else
               notch(')');  /* can't use this for "true", it will skip LF */
            }
         }
      else
         {
         value = 0;
         error_l("invalid identifier in defined() expression");
         }
      return value;
      }
   value = dvalue;  /* dvalue from number() via tokit() */
                    /* BUG: was
                     *         value = wvalue;
                     *      which limited the result to 0 .. 65535, despite
                     *      the fact that all the constexp() functions are
                     *      longs.  For example, the code
                     *          main()
                     *             {
                     *          #if 65536 > 0
                     *             printf("Correct: 65536 > 0\n");
                     *          #else
                     *             printf("Wrong: 65536 <= 0\n");
                     *          #endif
                     *             }
                     *      would print "Wrong: 65536 <= 0".
                     */
   if (heir != CONSTANT && heir != LCONSTANT)
      if (mactokn)  /* it's "ok" to have a bad expression in an #if */
         {
         toknlf();
         return 0;
         }
      else
         error("must have constant");
   else
      toknlf();
   return value;
   }


/* recursive via constexp() call in heir25c() */
long heir24c()  /* operator is prefix - + ! */
   {
   if (heir == 22 && curch == '-')  /* leading '-', negative number */
      {
      tokit();
      return -heir24c();  /* Hmmm... this allows "-+-5" to equal 5 */
      }
   else
      if (heir == 22 && curch == '+')  /* leading '+', positive number */
         {
         tokit();
         return heir24c();
         }
      else
         if (heir == 24)  /* leading '!', logical NOT */
            {
            tokit();
            return !heir24c();
            }
   return heir25c();  /* BUG: did not specify "return", that it worked was
                       *      accidental.
                       */
   }


/* recursive via constexp() call in heir25c() */
long heir23c()  /* operator is * and / and % */
   {
   long value;
   char op;

   value = heir24c();  /* get LEFT term */
   while (heir == 23)  /* while MUL or DIV or MOD */
      {
      op = bvalue;  /* remember which operator */
      toknlf();
      if (op == MUL)  /* do approprate math */
         value *= heir24c();
      else
         if (op == DIV)
            value /= heir24c();
         else
            value %= heir24c();
      }
   return value;
   }


/* recursive via constexp() call in heir25c() */
long heir22c()  /* operator is + and - */
   {
   long value;
   char op;

   value = heir23c();  /* get LEFT term */
   while (heir == 22)  /* while + or - */
      {
      op = bvalue;  /* remember which operator */
      toknlf();
      if (op == ADD)  /* do ADD / SUB */
         value += heir23c();
      else
         value -= heir23c();
      }
   return value;
   }

/* recursive via constexp() call in heir25c() */
long heir21c()   /* operator is << and >> */
   {
   long value;
   char op;

   value = heir22c();  /* get LEFT term */
   while (heir == 21)  /* while a SHIFT */
      {
      op = bvalue;  /* remember which shift */
      toknlf();
      if (op == SHL)  /* do appropriate shift */
         value <<= heir22c();
      else
         value >>= heir22c();
      }
   return value;
   }

/* recursive via constexp() call in heir25c() */
long heir20c()  /* operator is relational */
   {
   long value;
   char op;

   value = heir21c();  /* get LEFT term */
   while (heir == 20)  /* while realtional operator */
      {
      op = bvalue;  /* remember which operator */
      toknlf();
      switch (op)  /* do appropriate update */
         {
         case LT:
            value = value < heir21c();
            break;
         case LE:
            value = value <= heir21c();
            break;
         case GE:
            value = value >= heir21c();
            break;
         case GT:
            value = value > heir21c();
         }
      }
   return value;
   }

/* recursive via constexp() call in heir25c() */
long heir19c()   /* operator is == and != */
   {
   long value;
   char op;

   value = heir20c();  /* get LEFT term */
   while (heir == 19)  /* == or != */
      {
      op = bvalue;   /* remember if this is == or != */
      toknlf();
      value = value != heir20c();  /* new value, assume operator was != */
      if (op == EQ)  /* if opeator was == */
         value = !value;  /* logically NOT the results */
      }
   return value;
   }

/* recursive via constexp() call in heir25c() */
long heir18c()  /* operator is binary AND (&) */
   {
   long value;

   value = heir19c();  /* get LEFT term */
   if (heir == 18)  /* opeator is binary AND? */
      {
      toknlf();
      value &= heir18c();  /* new value */
      }
   return value;
   }

/* recursive via constexp() call in heir25c() */
long heir17c()  /* operator is binary XOR (^) */
   {
   long value;

   value = heir18c();  /* get LEFT term */
   if (heir == 17)  /* opeator is binary XOR? */
      {
      toknlf();
      value = value ^ heir17c();  /* new value */
      }
   return value;
   }

/* recursive via constexp() call in heir25c() */
long heir16c()  /* operator is binary OR (|) */
   {
   long value;

   value = heir17c();  /* get LEFT term */
   if (heir == 16)  /* operator is logical OR? */
      {
      toknlf();
      value |= heir16c();  /* new value */
      }
   return value;
   }

/* recursive via constexp() call in heir25c() */
long heir15c()  /* operator is logical AND (&&) */
   {
   long value, valr;

   value = heir16c();  /* get LEFT term */
   if (heir == 15)  /* operator is logical AND? */
      {
      toknlf();
      valr = heir15c();  /* get RIGHT term */
      value = value && valr;  /* new value */
      }
   return value;
   }

/* recursive via constexp() call in heir25c() */
long heir14c()  /* operator is logical OR (||) */
   {
   long value, valr;

   value = heir15c();  /* get LEFT term */
   if (heir == 14)  /* operator is logical OR? */
      {
      toknlf();
      valr = heir14c();  /* get RIGHT term */
      value = value || valr;  /* new value */
      }
   return value;
   }

/* recursive via constexp() call in heir25c() */
long constexp()  /* operator is ? : */
   {
   long value, vall, valr;

   value = heir14c();  /* get LEFT */
   if (ifch('?'))  /* if conditional */
      {
      vall = constexp();  /* get TRUE value */
      if (notch(':'))
         return 0;
      valr = constexp();  /* get FALSE value */
      if (value)  /* new value */
         value = vall;
      else
         value = valr;
      }
   return value;
   }

/* expression evaluator */

/* recursive via heir11() call in heir25(), heir26() */
heir11(node)  /* operator is , */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int list[7];  /* max LST node size */
   char n, lval;

   n = 0;  /* number of nodes == 0 */
   do
      {
      lval = heir12(node);  /* parse individual element */
      if (n && node[0] == 0)
         nooper();  /* missing operand */
      list[++n] = node[0];  /* add nodes to list */
      if (n == 6)  /* if max LST size */
         {
         list[0] = LST + 6;
         list[1] = treev(7,list);  /* emit LST node */
         n = 1;
         }
      } while (ifch(','));  /* while comma-seperated expressions */
   if (n > 1)  /* if more nodes */
      {
      list[0] = LST + n;
      node[0] = treev(n+1,list);  /* emit LST node */
      lval = 0;  /* since we already emited a LST node, no need to do it again */
      }
   return lval;  /* return value to add to LST node */
   }

/* recursive via heir11() call in heir25(), heir26() */
heir12(node)  /* operater is (op)assignment */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int nodeo[4], node2[2], tanode[2];
   char lval, addop, ltype, tp;

   lval = heir13(node);  /* element can be any higher level operator */
   if (heir == 12)  /* if (op)assignment (=, +=, -=, *=, etc) */
      {
      struct_type = node[1];      /* see if structure assignment */
      if (struct_type->sttype == CSTRUCT && bvalue == ASGN)  /* just assignment */
         {
         nodeo[0] = MOVE;  /* build move_node struct */
         tanode[0] = TA + (is_big ? (PTRTO << 8) : (CUNSG << 8));  /* large memory model */
                           /* take-address node, get ptr to structure */
         tanode[1] = node[0];
         nodeo[2] = tree2(tanode);  /* set LEFT node */
         nodeo[3] = struct_type->ststagat->staglen;  /* sizeof(structure) */
         tokit();  /* skip '=' */
         heir12(node2);  /* get right-hand size of assignment */
         tanode[1] = node2[0];  /* get node of right-hand tree */
         nodeo[1] = tree2(tanode);  /* set RIGHT node */
         struct_type = node2[1];
         if (struct_type->sttype != CSTRUCT ||
                         struct_type->ststagat->staglen != nodeo[3])
            error("illegal structure assignment");
         node[0] = tree4(nodeo);  /* emit MOVE node */
         return 0;  /* tree for expression already emitted */
         }
      if (lval == 0)  /* if LEFT is not a variable (1 = a; = a, etc) */
         {
         error("need lval");
         return 0;
         }
      nodeo[0] = bvalue;  /* assignment type =, +=, etc */
      addop = 0;
      ltype = ((char *)node[1])->byte;
      if (ltype == PTRTO && ((bvalue == AADD) || (bvalue == ASUB)))
         addop = 1;  /* pointer addition/subtraction */
      nodeo[1] = node[0];  /* set left */
      tokit();
      heir12(node2);  /* get right node */
      nodeo[2] = node2[0];  /* set right node */
      if (nodeo[1] == 0 || nodeo[2] == 0)  /* either node is NULL */
         nooper();  /* missing operand */
      struct_type = node2[1];
      if (struct_type->sttype == CSTRUCT)  /* assigning a struct to a simple var */
         error("illegal structure assignment");
      if ((ltype >= CFLOAT) && (ltype < FUNCTION) && (nodeo[0] != ASGN))
                 /* check for special (op)assignment cases */
         {
         if (nodeo[0] == AMOD)  /* if MOD, LEFT cannot be a float/double */
            nofloat(node[1],node[1]);
         if (ltype >= BITS && ltype < FUNCTION)
               /* i.e. struct { int i:1; } x,y; main() { x.i += y.i; }
                *      cannot be done "in one piece", break it into an OP
                *      and an ASSIGN.
                */
            {
            nodeo[0] -= AADD - ADD;
            nodeo[2] = tree3(nodeo);
            nodeo[0] = ASGN;
            }
         addop = 0;
         }
      if (addop)  /* pointer arithmetic */
         addptr(nodeo[0],node,node2);  /* adjust for sizeof(target data type) */
      else
         node[0] = tree3(nodeo);  /* just output the tree for this (op)assignment */
      return 0;  /* tree has already been accounted for */
      }
   return lval;  /* return tree root */
   }

/* recursive via heir11() call in heir25(), heir26() */
heir13(node)  /* operater is conditional ( ? : ) */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int nodeo[4], node2[2], node3[2];
   char lval, ntype;
   int typ;

   lval = heir14(node);  /* get value of LEFT in conditional */
   if (ifch('?'))  /* if this is a conditional */
      {
      nodeo[0] = TEST;  /* build test_node struct */
      nodeo[1] = node[0];
      heir12(node2);  /* get tree for TRUE code */
 /* NOTE:  It may be a bug, but I think that the C standard says that
  *        the following code is legal:
  *           int i,j,k,l; main() { i = j ? k += l, l += 1 : l += k, k += 1; }
  *        i.e. that comma operators are legal in the conditional code.  If
  *        that is correct, then this should be a call to heir11(node2), not
  *        heir12(node2).
  */
      nodeo[2] = node2[0];
      if (notch(':'))
         return 0;
      heir12(node3);  /* get tree for FALSE code */
 /* NOTE:  It may be a bug, but I think that the C standard says that
  *        the following code is legal:
  *           int i,j,k,l; main() { i = j ? k += l, l += 1 : l += k, k += 1; }
  *        i.e. that comma operators are legal in the conditional code.  If
  *        that is correct, then this should be a call to heir11(node3), not
  *        heir12(node3).
  */
      nodeo[3] = node3[0];
      if (nodeo[1] == 0 || nodeo[2] == 0 || nodeo[3] == 0)
         nooper();  /* missing operand */
      node[1] = maxtype(node2[1],node3[1]);  /* get type of resulting node */
      if (*node[1] >= BITS && *node[1] < FUNCTION)
         node[1] = u_type;
      typ = typeof(node[1]);
      if (typ == (ARRAY << 8) || typ == (CSTRUCT << 8))
         typ = is_big ? (PTRTO << 8) : (CUNSG << 8);  /* large memory model */
      nodeo[0] += typ;
      node[0] = tree4(nodeo);
      return 0;  /* tree has already been accounted for */
      }
   return lval;  /* return root of tree */
   }

/* recursive via heir11() call in heir25(), heir26() */
heir14(node)  /* operater is logical OR (||) */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int nodeo[3];
   char lval;

   lval = heir15(node);  /* get LEFT term */
   if (heir == 14)  /* operator is logical OR? */
      {
      nodeo[0] = LOR;  /* build math_node struct */
      nodeo[1] = node[0];
      tokit();  /* skip "||" */
      heir14(node);  /* get RIGHT term */
      nodeo[2] = node[0];
      if (nodeo[1] == 0 || nodeo[2] == 0)
         nooper();  /* missing operand */
      node[0] = tree3(nodeo);  /* emit LOR node */
      node[1] = int_type;
      return 0;
      }
   return lval;
   }


/* recursive via heir11() call in heir25(), heir26() */
heir15(node)  /* operater is logical AND (&&) */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int nodeo[3];
   char lval;

   lval = heir16(node);  /* get LEFT term */
   if (heir == 15)  /* operator is logical AND? */
      {
      nodeo[0] = LAND;  /* build math_node struct */
      nodeo[1] = node[0];
      tokit();
      heir15(node);  /* get RIGHT term */
      nodeo[2] = node[0];
      if (nodeo[1] == 0 || nodeo[2] == 0)
         nooper();  /* missing operand */
      node[0] = tree3(nodeo);  /* emit LAND node */
      node[1] = int_type;
      return 0;
      }
   return lval;
   }


/* recursive via heir11() call in heir25(), heir26() */
heir16(node)  /* operater is binary OR (|) */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int nodeo[3], node2[2];
   char lval;

   lval = heir17(node);  /* get LEFT term */
   if (heir == 16)  /* operator is binary OR? */
      {
      nodeo[0] = OR;  /* build math_node struct */
      nodeo[1] = node[0];
      tokit();
      heir16(node2);  /* get RIGHT term */
      nodeo[2] = node2[0];
      if (nodeo[1] == 0 || nodeo[2] == 0)
         nooper();  /* missing operand */
      node[0] = tree3(nodeo);  /* emit OR node */
      node[1] = nofloat(node[1],node2[1]);  /* neither LEFT nor RIGHT may be float/double */
      return 0;
      }
   return lval;
   }

/* recursive via heir11() call in heir25(), heir26() */
heir17(node)  /* operater is binary XOR (^) */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int nodeo[3], node2[2];
   char lval;

   lval = heir18(node);  /* get LEFT term */
   if (heir == 17)  /* operator is binary XOR? */
      {
      nodeo[0] = XOR;  /* build math_node struct */
      nodeo[1] = node[0];
      tokit();
      heir17(node2);  /* get RIGHT term */
      nodeo[2] = node2[0];
      if (nodeo[1] == 0 || nodeo[2] == 0)
         nooper();  /* missing operand */
      node[0] = tree3(nodeo);  /* emit XOR node */
      node[1] = nofloat(node[1],node2[1]);  /* neither LEFT nor RIGHT may be float/double */
      return 0;
      }
   return lval;
   }

/* recursive via heir11() call in heir25(), heir26() */
heir18(node)  /* operater is binary AND (&) */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int nodeo[3], node2[2];
   char lval;

   lval = heir19(node);  /* get LEFT term */
   if (heir == 18)  /* operator is binary AND? */
      {
      nodeo[0] = AND;  /* build math_node struct */
      nodeo[1] = node[0];
      tokit();
      heir18(node2);  /* get RIGHT term */
      nodeo[2] = node2[0];
      if (nodeo[1] == 0 || nodeo[2] == 0)
         nooper();  /* missing operand */
      node[0] = tree3(nodeo);  /* emit AND node */
      node[1] = nofloat(node[1],node2[1]);  /* neither LEFT nor RIGHT may be float/double */
      return 0;
      }
   return lval;
   }

/* recursive via heir11() call in heir25(), heir26() */
heir19(node)  /* operater is equality (== and !=) */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int nodeo[3];
   char lval;

   lval = heir20(node);  /* get LEFT term */
   while (heir == 19)  /* EQUAL / NOT_EQUAL? */
      {
      nodeo[0] = bvalue;  /* build math_node struct */
      nodeo[1] = node[0];
      tokit();
      heir20(node);  /* get RIGHT Term */
      nodeo[2] = node[0];
      if (nodeo[1] == 0 || nodeo[2] == 0)
         nooper();  /* missing operand */
      node[0] = tree3(nodeo);  /* emith EQUAL / NOT_EQUAL node */
      node[1] = int_type;  /* result is an interger */
      lval = 0;
      }
   return lval;
   }

/* recursive via heir11() call in heir25(), heir26() */
heir20(node)  /* operater is relational (< <= etc.) */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int nodeo[3];
   char lval;

   lval = heir21(node);  /* get LEFT term */
   while (heir == 20)  /* while <, <=, <>, etc */
      {
      nodeo[0] = bvalue;  /* build math_node struct */
      nodeo[1] = node[0];
      tokit();
      heir21(node);  /* get RIGHT term */
      nodeo[2] = node[0];
      if (nodeo[1] == 0 || nodeo[2] == 0)
         nooper();  /* missing operand */
      node[0] = tree3(nodeo);  /* emit relational node */
      node[1] = int_type;  /* result is an INT */
      lval = 0;
      }
   return lval;
   }


/* recursive via heir11() call in heir25(), heir26() */
heir21(node)  /* operater is shift: << or >> */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int nodeo[3], node2[2];
   char lval;

   lval = heir22(node);  /* get LEFT node */
   while (heir == 21)  /* while a SHIFT type */
      {
      nodeo[0] = bvalue;  /* build math_node struct */
      nodeo[1] = node[0];
      tokit();
      heir22(node2);  /* get RIGHT node */
      nodeo[2] = node2[0];
      if (nodeo[1] == 0 || nodeo[2] == 0)
         nooper();  /* missing operand */
      node[0] = tree3(nodeo);  /* emit SHIFT node */
      node[1] = nofloat(node[1],node2[1]);  /* neither LEFT nor RIGHT may be float/double */
      lval = 0;
      }
   return lval;
   }


/* recursive via heir11() call in heir25(), heir26() */
heir22(node)  /* operater is add sub (+ -) */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
{
   int noder[2];
   char lval, addop;

   lval = heir23(node);  /* get LEFT term */
   while (heir == 22)  /* while ADD or SUB */
      {
      addop = bvalue;  /* save ADD / SUB */
      lval = 0;
      tokit();
      noder[0] = 0;  /* init right node struct */
      heir23(noder);  /* get RIGHT term */
      if (node[0] == 0 || noder[0] == 0)
         nooper();  /* missing operand */
      else
         addptr(addop,node,noder);  /* do ADD / SUB, allowing for pointers */
      }
   return lval;
   }

addptr(addop,node,noder)
   char addop;
   int node[],noder[];
   {
   int nodeo[3], sizel, sizer, val;

   sizel = scaled(node[1]);
   sizer = scaled(noder[1]);
   if (sizel == 1 && sizer > 1)  /* LEFT not scaled, but RIGHT is
                                  * i.e. int *a, b; main() { a = b + a; }
                                  */
      {
      scale(node,sizer);  /* scale LEFT by size (in bytes) of RIGHT */
      node[1] = noder[1];
      }
   else
      if (sizer == 1 && sizel > 1)  /* LEFT scaled, but not RIGHT
                                     * i.e. int *a, b; main() { a = a + b; }
                                     */
         scale(noder,sizel);  /* scale RIGHT by size (in bytes) of LEFT */
   /* if subtracting pointers, change to offset only */
   if (is_big && *noder[1] >= FUNCTION &&           /* large memory model */
                          (addop == SUB || addop == ASUB))
      {
      if (*node[1] < FUNCTION )
         error("illegal arithmetic");
      /* cast both to unsigned */
      node[1] = node[0];
      node[0] = CAST + (CUNSG << 8);
      node[0] = tree2(node);
      node[1] = u_type;
      noder[1] = noder[0];
      noder[0] = CAST + (CUNSG << 8);
      noder[0] = tree2(noder);
      noder[1] = u_type;
      }      
   nodeo[0] = addop;  /* build math_node struct for scaled ADD/SUB */
   nodeo[1] = node[0];
   nodeo[2] = noder[0];
   node[0] = tree3(nodeo);
   node[1] = maxtype(node[1],noder[1]);  /* result is "bigger" of two terms */
   if (sizer > 1 && sizer == sizel && (addop == SUB || addop == ASUB))
             /* if scaled SUB / ASUB (-=), divide by size to get index */
      {
      nodeo[2] = oconst(sizer);  /* size of data element (bytes) */
      nodeo[0] = addop + DIV - SUB;  /* DIV or ADIV */
      nodeo[1] = node[0];
      node[0] = tree3(nodeo);  /* get element index */
      node[1] = int_type;  /* result is an integer */
      }
   }

scale(node,times)
   int node[], times;
   {
   int nodeo[3];
   char typ;

   nodeo[2] = node[0];  /* start building math_node struct */
   typ = *((char *)(node[1]));
   if (typ == CCHAR || typ == CSCHAR || typ == PTRTO)  /* char, unsigned char, or pointer */
      {
      node[1] = node[0];
      node[0] = CAST + (CINT << 8);  /* cast this to an INT */
      nodeo[2] = tree2(node);
      }
   nodeo[1] = oconst(times);  /* get size in bytes */
   nodeo[0] = MUL;
   node[0] = tree3(nodeo);
   node[1] = int_type;  /* result is an integer */
   }

/* return a pointer to the maximum types of the nodes pointer to. report
   an error if a type is floating point. */

nofloat(typea,typeb)
   char *typea, *typeb;
   {
   char *typ;

   typ = maxtype(typea,typeb);
   if (*typ == CFLOAT || *typ == CDOUBLE)
      {
      error("illegal use of FLOAT");
      typ = int_type;
      }
   return typ;
   }


/* return the type that will be produced by arighmetic on the two
   arguments. */

maxtype(typea,typeb)
   char *typea, *typeb;
   {
   return *typea > *typeb ? typea: typeb;
   }
