/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  DeSmet C is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; either version 2 of the License, or any
 *  later version.
 *
 *  DeSmet C is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/* c5.C               PART 5 OF C88 COMPILER */
/*
 *  Modification history:
 *    William Couture    10/12/05   Reformat/cleanup/comments
 *                                  changed 'heir' magic numbers to names
 */

#include "pass1.h"
#include "nodes.h"
#if CHECK
#include "OBJ.H"
#endif

/* recursive via heir11() call in heir25(), heir26() */
heir23(node)  /* operater is MUL and DIV and MOD (* and / and %) */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int nodeo[3], node2[2];
   char lval, bv;

   lval = heir24(node);  /* get LEFT term */
   while (heir == 23)  /* while MUL/DIV/MOD */
      {
      nodeo[0] =
      bv = bvalue;  /* remember which op */
      nodeo[1] = node[0];
      tokit();
      heir24(node2);  /* get RIGHT term */
      nodeo[2] = node2[0];
      if (nodeo[1] == 0 || nodeo[2] == 0)
         nooper();  /* missing operand */
      node[0] = tree3(nodeo);
      if (bv != MOD)
         node[1] = maxtype(node[1],node2[1]);  /* type promotion */
      else
         node[1] = nofloat(node[1],node2[1]);  /* MOD doesn't work for FLOAT/DOUBLE */
      lval = 0;
      }
   return lval;
   }


/* recursive function */
/* also recursive via heir11() call in heir25(), heir26() */
heir24(node)  /* operater is prefix (* & - ! ~ ++ --) */
              /* or SIZEOF */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int nodeo[3], backup;
   char lval, needrp, *bptr, savenest;
   char savestor;
   unsigned saveat, saveloc, savetype;

   if (heir == RESERVED && bvalue == RSIZEOF)  /* if sizeof operator */
      {
      tokit();
      needrp = 0;
      backup = cur;
      savenest = nested;
      /* save add information in case initilizing a variable */
      savestor = addstor;
      savetype = addtype;
      saveat = addat;
      saveloc = addloc;
      if (ifch('('))
         {
         if (specs(SUNSTOR))  /* if a data type, not variable name */
                /* NOTE: this is using the "typecast" logic to implement
                 *       sizeof(datatype)
                 */
            {
            node[0] = oconst(dsize(abstract(0)));  /* implicit small memory model */
            node[1] = int_type;
            addstor = savestor;
            addtype = savetype;
            addat = saveat;
            addloc = saveloc;
            return 0;
            }
         else
            {
            addstor = savestor;
            cur = backup;
            nested = savenest;
            tokit();  /* get variable name */
            needrp = 1;  /* we have to take care of tailing parenthesis */
            }
         }
/* if SIZEOF, do not allow a logical expression. */
      heir24(node);
      addstor = savestor;
      addtype = savetype;
      addat = saveat;
      addloc = saveloc;
      node[0] = oconst(dsize(node[1]));  /* implicit small memory model */
      node[1] = int_type;
      if (needrp)
         notch(')');
      return 0;
      }

   switch(curch)  /* handle prefixes (* (ptr), & (addr of), +, -, ++, --, !, ~) */
      {
      case '*':  /* leading '*' is a pointer, not a mul */
         if (heir == 23)
            {
            tokit();
            heir24(node);  /* get variable pointed to */
            bptr = node[1];
            if (*bptr == PTRTO)  /* de-referenceing a pointer? */
               bptr++;
            else
               if (*bptr == ARRAY)  /* de-referencing an array used as a pointer? */
                  bptr += 3;
               else
                  {
                  if (is_big || (*bptr != CINT && *bptr != CUNSG))  /* large memory model */
                     error("illegal indirection");
                  bptr = c_type;
                  }
            node[1] = bptr;
            if (*bptr == ARRAY)  /* if array, we are already doing indirect */
               return 0;
            nodeo[0] = IND + typeof(bptr);  /* build ind_node structure */
            nodeo[1] = node[0];
            if (nodeo[1] == 0)
               nooper();  /* missing operand */
            node[0] = tree2(nodeo);
            return 1;
            }
         break;
      case '&':  /* address-of */
         if (heir == 18)  /* bit-and, but leading '&' is a ADDRESS OF */
            {
            tokit();
            lval = heir24(node);  /* get variable pointed to */
            bptr = node[1];
            if (*bptr == ARRAY || *bptr == FUNCTION)  /* array or fuction is already an address */
               return 0;
            if (lval == 0 && *bptr != CLABEL && *bptr != CSTRUCT)  /* not a variable */
               {
               error("illegal address");
               return 0;
               }
            nodeo[0] = TA;  /* build ta_node struct */
            nodeo[1] = node[0];
            if (nodeo[1] == 0)
               nooper();  /* missing operand */
            node[0] = tree2(nodeo);
            node[1] = abstract(node[1]);
            return 0;
            }
         break;
      case '-':  /* leading '-', negative number */
         if (heir == 22)
            {
            tokit();
            if (heir == CONSTANT)
               {
               wvalue = -wvalue;
               heir24(node);
               return 0;
               }
            else
               if (heir == LCONSTANT)
                  {
                  dvalue = -dvalue;
                  heir24(node);
                  return 0;
                  }
               else
                  if (heir == FFCONSTANT)
                     {
                     fvalue[1] ^= 0x8000;
                     heir24(node);
                     return 0;
                     }
                  else
                     if (heir == FDCONSTANT)
                        {
                        fvalue[3] ^= 0x8000;
                        heir24(node);
                        return 0;
                        }
            heir24(node);  /* get value */
            nodeo[0] = NEG;  /* build unary_node struct */
            nodeo[1] = node[0];
            if (nodeo[1] == 0)
               nooper();  /* missing operand */
            node[0] = tree2(nodeo);
            return 0;
            }
         /* NOTE: no break, let pre-dec (heir == 24) fall through into
          *       pre-inc code
          */
      case '+':
         if (heir == 22)  /* leading '+', positive number */
            {
            tokit();
            heir24(node);  /* is basically a NOP, get value */
            return 0;
            }
         if (heir == 24)  /* leading '++' or '--' */
            {
            nodeo[0] = bvalue;  /* remember PREI or PRED */
            tokit();
            lval = heir24(node);   /* get variable */
            if (lval == 0)  /* must be a variable */
               {
               error("need lval");
               return 0;
               }
            nodeo[1] = node[0];
            if (nodeo[1] == 0)
               nooper();  /* missing operand */
            nodeo[2] = scaled(node[1]);  /* if pointer, scale appropriately */
            node[0] = tree3(nodeo);
            return 0;
            }
         break;
      case '!':
         if (heir == 24)  /* logical NOT */
            {
            if (heir == CONSTANT)  /* BUG: since we know heir == 24, it can
                                    *      NEVER be == CONSTANT.  Is there
                                    *      a tokit(); missing before this if?
                                    *      Note that this would be an
                                    *      efficiency thing, not correctness
                                    */
               {
               wvalue = !wvalue;  /* and this should be wvalue = !dvalue,
                                   * and the type should be forced to INT
                                   */
               heir24(node);  /* parse variable */
               return 0;
               }
            nodeo[0] = NOT;  /* build unary_node struct */
            tokit();
            lval = heir24(node);
            nodeo[1] = node[0];
            if (nodeo[1] == 0)
               nooper();  /* missing operand */
            node[0] = tree2(nodeo);
            node[1] = int_type;  /* result is always an INT */
            return 0;
            }
         break;
      case '\176':  /* '~' */
            nodeo[0] = COMP;  /* bitwise complement, build unary_node struct */
            tokit();
            lval = heir24(node);
            nodeo[1] = node[0];
            if (nodeo[1] == 0)
               nooper();  /* missing operand */
            node[0] = tree2(nodeo);
            node[1] = nofloat(node[1],node[1]);
            return 0;
      default:  /* default case isn't really needed */
         ;
      }
   lval = heir25(node);  /* parse term */
   if (heir == 24 && (curch == '+' || curch == '-'))  /* post-inc or post-dec */
      {
      nodeo[0] = bvalue + 2;  /* change PREI and PRED to POSTI and POSTD */
      tokit();
      if (lval == 0)
         {
         error("need lval");
         return 0;
         };  /* BUG: trivial one, the semi-colon doesn't do anything */
      nodeo[1] = node[0];
      if (nodeo[1] == 0)
         nooper();  /* missing operand */
      nodeo[2] = scaled(node[1]);  /* if pointer, scale appropriately */
      node[0] = tree3(nodeo);
      lval = 0;
      }
   return lval;
   }

/* recursive via heir11() call in heir25(), heir26() */
heir25(node)  /* look for a primary */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   char lval, *bptr, really_stru;
   int node2[2], nodeo[3], scal, struct_len;

   lval = heir26(node);  /* get term */
   if (heir == 25 && *node[1] == CSTRUCT)  /* need offset of struct */
      {
      node2[1] = node[0];
      node2[0] = TA + (is_big ? (PTRTO << 8) : (CUNSG << 8));  /* large memory model */
      node[0] = tree2(node2);
      }

   really_stru = 0;
   while (1)
      {
      if (ifch('['))  /* if array index */
         {
         heir11(node2);  /* get index */
         if (node2[0] == 0)
            nooper();  /* missing operand */
         if (is_big && *node2[1] == PTRTO)  /* large memory model */
            error("illegal index");
         if (notch(']'))  /* end of array index? */
            return 0;
         bptr = node[1];
         if ((scal = scaled(bptr)) != 1)  /* if not pointer to BYTE */
            scale(node2,scal);  /* scale index to element size */
         nodeo[0] = ADD;  /* build math_node struct */
         nodeo[1] = node[0];  /* add scaled index to array base */
         nodeo[2] = node2[0];
         node[0] = tree3(nodeo);
         if (*bptr == ARRAY)
            bptr += 3;
         else
            if (*bptr == PTRTO)
               bptr++;
            else
               {
               if (is_big)  /* large memory model */
                  error("illegal indirection");
               bptr = c_type;
               }
         node[1] = bptr;
         lval = 0;
         really_stru = 0;
         if (*bptr != ARRAY && *bptr != CSTRUCT)
            {
            nodeo[0] = IND + typeof(bptr);  /* build ind_node struct */
            nodeo[1] = node[0];
            node[0] = tree2(nodeo);
            lval = 1;
            }
         else
            if (*bptr == CSTRUCT)
               really_stru = 1;
         }
      else
         if (heir == 25)  /* structure reference: -> or . */
            {
            tokit();
            if (is_big && *node[1] != PTRTO && *node[1] != CSTRUCT)  /* large memory model */
               error("illegal indirection");
            if (!find_member(node[1]))
               return 0;
            if (heir != OPERAND || nameat->nstor != SMEMBER)
               {
               error("need member");
               return 0;
               }
            if (nameat->noff)  /* if not at offset 0 from base address */
               {
               nodeo[1] = oconst(nameat->noff);  /* add offset to base address */
                                          /* implicit small memory model */
               nodeo[0] = ADD;  /* build math_node struct */
               nodeo[2] = node[0];
               node[0] = tree3(nodeo);
               }
            lval = 0;
            really_stru = 0;
            if (nameat->ntype[0] != ARRAY && nameat->ntype[0] != CSTRUCT)
               {
               nodeo[0] = IND + typeof(&nameat->ntype[0]);  /* i.e. int i,*j; *j=i; */
               nodeo[1] = node[0];
               node[0] = tree2(nodeo);
               lval = 1;
               }
            else
               if (nameat->ntype[0] == CSTRUCT)
                  really_stru = 1;
            node[1] = &nameat->ntype[0];
            tokit();
            }
         else
            if (curch == '(')  /* function call */
               {
               bptr = node[1];
               if (*bptr == FUNCTION)
                  bptr++;
               else
                  if (*bptr == PTRTO && *(bptr+1) == FUNCTION)
                     bptr += 2;
                  else
                     {
                     if (is_big)  /* large memory model */
                        error("illegal indirection");
                     else
                        warning("indirect call");
                     }
               nodeo[0] = CALL + typeof(bptr);  /* setup function call */
               nodeo[1] = node[0];
               for (wp = protohash[nameat & 31]; wp; wp = *wp)
                             /* see if we can find a function prototype */
                  if (*(wp+1) == nameat)  /* if we have a prototype */
                     {
                     nodeo[2] = args(*(wp+2));  /* parse args and compare to prototype */
                     goto pfmt;
                     }
               nodeo[2] = args(-1);  /* parse args, no prototype to check */
                      /* NOTE: pass -1 to distinguish this from 0 (== NULL),
                       *       end-of-parameters
                       */
pfmt:          struct_type = bptr;
               nodeo[3] = struct_type->sttype == CSTRUCT ?
                                       struct_type->ststagat->staglen : 0;
                          /* if function returns a structure, emit structure len else 0 */
               node[0] = tree4(nodeo);
               if (bptr != node[1])  /* node type is not current type */
                  node[1] = bptr;
               else
                  node[1] = int_type;  /* default to INT */
               really_stru = 0;
               lval = 0;
               }
            else
               break;
      }

   /* if found a structure as an element of another structure or array,
      need to take contents of. */
   if (really_stru)
      {
      nodeo[1] = node[0];
      nodeo[0] = IND + (CSTRUCT << 8);
      node[0] = tree2(nodeo);
      }
   return lval;
   }

/* parse function arguments */
/* recursive via heir12() call */
args(struct pargs *ap)
   {
   int n;
   int list[7], node[2], nodeo[4], val;
   char *ntype;

   tokit();  /* skip '(' */
   n = 7;
   if (ifch(')'))  /* empty argument list */
      {
      if (ap != 0 && ap != -1)
         error("not enough arguments");
      return 0;  /* NULL, no args */
      }
   do
      {
      if (ap == 0)  /* if end-of-chain on arguments */
         {
         error("too many arguments");
         return 0;
         }
      val = heir12(node);  /* get tree for this argument */
      if (node[0] == 0)
         {
         error("missing argument");
         return 0;
         }
      if (ap != -1)  /* if we have a type for this argument */
         {
         if (typeof(ap->ptype) != typeof(node[1])  /* different types */
                        && ap->ptype[0] < ARRAY && *((char*)node[1]) < ARRAY)
                                   /* and arguments are not pointers */
            {
            if (wopt)  /* show warnings */
               warning("argument type conversion");
            ntype = node[1];
            node[1] = node[0];
            node[0] = CAST + typeof(ap->ptype);  /* build cast_node struct */
            if (is_big && val)  /* large memory model */
               {
               if (node[0] == CAST+(PTRTO << 8) && *ntype != PTRTO &&
                    *ntype != CSTRUCT && *ntype != ARRAY && *ntype != FUNCTION)
                  error("illegal indirection");
               }
            node[0] = tree2(node);
            node[1] = ap->ptype;
            }
         ap = ap->plink;  /* next argument type */
            /* BUG: I have to check, but I think there needs to be a check
             *      for RDOTDOT argument (i.e. int printf(char *,...);),
             *      beyond which arguments are not checked.  However, since
             *      RDOTDOT didn't work, lack of a check here doesn't hurt
             *      anything.
             */
         }
      ntype = node[1];
      if (*ntype == CSTRUCT)
         {
         if (wopt)  /* show warnings */
            warning("structure argument");
          /* need address of struct */
         nodeo[0] = TA + (is_big ? (PTRTO << 8) : (CUNSG << 8));  /* large memory model */
                       /* build ta_node struct */
         nodeo[1] = node[0];
         nodeo[1] = tree2(nodeo);
         nodeo[0] = MOVE;  /* build move_node struct */
         nodeo[2] = 0;  /* RIGHT == NULL (no dest), do a PUSH */
         struct_type = ntype;
         nodeo[3] = struct_type->ststagat->staglen;
         node[0] = tree4(nodeo);
         }
      list[--n] = node[0];  /* add node of parameter, right-to-left */
      if (n == 1)  /* if FULL (6 nodes) */
         {
         list[0] = LST + 6;  /* emit a LST node */
         list[6] = treev(7,list);
         n = 6;  /* we always have one "active" node, the previous LST */
         }
      } while (ifch(','));
   if (notch(')'))  /* syntax check */
      return 0;
   if (ap != 0 && ap != -1)  /* if not end-of-list, and have a prototype */
      {
      error("not enough arguments");
      return 0;
      }
   if (n == 6)  /* if no nodes beyond the previous LST */
      val = list[6];  /* return LST node to be placed in CALL node */
   else
      {
      list[0] = LST + 7 - n;  /* write LST for current nodes */
      val = tree1(list);  /* want to return node number of LST node */
      treev(7-n,&list[n]);  /* write LST node values */
      }
   return val;  /* return node of tree to "push" */
   }

/* type of a string. needed by odd casts. */
struct
   {
   char ary;
   int length;
   char tp;
   } string_type = { ARRAY, 0, CCHAR };


/* recursive via heir11() call */
heir26(node)  /* operand or constant or string */
   int node[];
      /* NOTE: node[0] == root node of expression
       *       node[1] == ptr to type of expression.  Note that this
       *                  assumes sizeof(int) == sizeof(ptr)
       */
   {
   int nodeo[5], save, dim;
   char lval, *typad, slen, savenest, oldstor, toked, tp, *oldtype;

   if (ifch('('))  /* term starts with '(', could be typecast or parenthesised expression */
      {
      oldstor = addstor;
      oldtype = addtype;
      if (specs(SUNSTOR))  /* check for a cast */
         {
         save = abstract(0);
/* dont allow a cast of a logical expression */
         lval = heir24(node);
         if (node[0] == 0)
            nooper();  /* missing operand */
         tp = *node[1];
         node[1] = node[0];
         node[0] = CAST + typeof(save);
         if (is_big && lval)  /* large memory model */
            {
            if (node[0] == CAST+(PTRTO << 8) && tp != PTRTO &&
                             tp != CSTRUCT && tp != ARRAY && tp != FUNCTION)
               error("illegal indirection");
            }
         node[0] = tree2(node);
         node[1] = save;
         }
      else
         {
         addtype = oldtype;
         lval = heir11(node);  /* parse parenthesised expression */
         notch(')');
         }
      addstor = oldstor;
      return lval;
      }

   toked = 0;
   if (heir == UNDEF)  /* if new symbol */
      {
      newname();  /* add it to symbol table */
      toked = 1;
      tokit();
      if (ininit)  /* initing variables, source must be defined */
         {
         toked = 0;
         wp = mfree;  /* remove the hash pointer for this guy */
         hash[hashno] = *wp;
         error("undefined variable");
         }
      else
         if (curch == '(')  /* new symbol is a function */
            {
            addstor =
            nameat->nstor = SEXTERN; /* must be globally visible */
            addat = nameat;
            addat->ntype[0] = FUNCTION;
            addtype = CINT;
            addloc = &nameat->ntype[1];
            atype(addtype);
#if CHECK
            if (copt)  /* add checkout info to object file */
               objtype(OPTYPE,addat->noff);
#endif
            }
         else  /* assume it is a local integer */
            {
            warning("undefined variable"); 
            nameat->nstor = SAUTO;
            locoff -= 2;
            nameat->noff = locoff;
            nameat->ntype[0] = CINT;
            mfree = &nameat->ntype[1];
            }
      goto nowop;  /* now, go use our new symbol */
      }
   if (heir == OPERAND && nameat->nstor != STYPEDEF)
      {
nowop:
      nodeo[1] =
      nodeo[2] =
      nodeo[3] = 0;
      if (nameat->nstor < SAUTO)  /* global, noff is symbol table index */
         nodeo[1] = nameat->noff;
      else
         nodeo[2] = nameat->noff;  /* local, noff is BP offset */
      typad = &nameat->ntype[0];
      if (toked == 0)  /* if we have not parsed this yet */
         tokit();
      while (curch == '.' || curch == '[')  /* if structure element or array element */
         {
         if (curch == '.')  /* structure element */
            {
            if (*typad != CSTRUCT)  /* but this is not a structure */
               {
               if (*typad != PTRTO)  /* if it is not even a pointer! */
                  {
                  error("need structure");
                  return 0;
                  }
               break;  /* pointer, "DeSmet-ism" that we can reference an
                        * element from any structure via it.  i.e.
                        *    struct { char byte0; char byte1; char byte2; };
                        *    char *p;
                        *    p->byte2;
                        */
               }
            tokit();  /* skip '.' */
            find_member(typad);  /* find structure element */
            if (heir != OPERAND || nameat->nstor != SMEMBER)
                   /* not a member of ANY structure (DeSmet-ism) */
               {
               error("need member");
               return 0;
               }
            /* if a zero offset, set flag for could be member of union */
            if (nameat->noff)
               nodeo[2] += nameat->noff;
            else
               nodeo[3] = 1;
            typad = &nameat->ntype[0];
            tokit();
            }
         else  /* must be '[' */
            {
            if (*typad != ARRAY)  /* if not an ARRAY, skip all this */
                        /* BUG: If not an array or pointer, this should be an
                         *      error.  i.e.
                         *         struct { char x; int y;} a;
                         *         main() { a[1] = 1; }
                         *      should be an error.  Instead the value in the
                         *      first word of a is treated as a char *
                         *      The LARGE memory model, at least, treats
                         *      this as an error.
                         */
               break;
            while (*cur == LF && cur > savEnd)  /* parsing */
               {
               cur = nestfrom[--nested];
               whitesp();
               }
            save = cur;
            savenest = nested;
            tokit();
            if (heir == CONSTANT)  /* constant value for array index */
                      /* efficiency -- if this is a constant offset,
                       * put the effective address in directly instead of
                       * computing it at runtime.
                       */
               {
               dim = wvalue;  /* NOTE: implicit small memory model! */
               tokit();
               if (ifch(']'))  /* just a simple constant index */
                  {
                  typad += 3;  /* up to size of array element */
                  dim = wvalue;  /* NOTE: implicit small memory model */
                  nodeo[2] += dsize(typad) * dim;  /* get byte offset from base address */
                  continue;
                  }
               }

            cur = save;  /* un-parse everything */
            nested = savenest;
            curch = '[';  /* back to '[' for more parsing */
            break;
            }
         }  /* while (curch == '.' || curch == '[') */
      nodeo[0] = OPND + typeof(typad);  /* variable reference, get it from table */
      node[0] = tree4(nodeo);
      node[1] = typad;
      if (*typad < BITS)  /* "normal" types */
         {
         if (*typad < CLABEL)  /* CCHAR .. CULONG */
            return 1;
         }
      else
         if (*typad < FUNCTION || *typad > ARRAY)  /* BITS, FPTR2, HPTR2, PTRTO */
            return 1;
      return 0;  /* CSTRUCT, FUCTION, ARRAY */  /* I think... */
      }
   if (heir == CONSTANT)
      {
      node[0] = oconst(wvalue);  /* appropriate NODE type for constant BYTE/WORD constant */
                 /* implict small memory model */
      node[1] = wvalue > 255 ? int_type : c_type;  /* set type */
      tokit();
      return 0;
      }
   if (heir == LCONSTANT)  /* long constant */
      {
      nodeo[0] = CONST + (CLONG << 8);  /* build const_lnode struct */
      nodeo[1] = dvalue;
      nodeo[2] = ((void *)&dvalue)->hiword;
      node[0] = tree3(nodeo);
      node[1] = l_type;  /* set type */
      tokit();
      return 0;
      }
   if (heir == FFCONSTANT)
      {
      nodeo[0] = CONST + (CFLOAT << 8);  /* build const_lnode struct */
      nodeo[1] = fvalue[0];
      nodeo[2] = fvalue[1];
      node[0] = tree3(nodeo);
      node[1] = ff_type;
      tokit();
      return 0;
      }
   if (heir == FDCONSTANT)
      {
      nodeo[0] = CONST + (CDOUBLE << 8);  /* build const_dnode struct */
      nodeo[1] = fvalue[0];
      nodeo[2] = fvalue[1];
      nodeo[3] = fvalue[2];
      nodeo[4] = fvalue[3];
      node[0] = tree5(nodeo);
      node[1] = fd_type;
      tokit();
      return 0;
      }
   if (heir == STRNG)
              /* char *p = "abc"; */
      {
      nodeo[0] = OPND + (ARRAY << 8);  /* build opnd_node struct */
      nodeo[1] = ++ordinal;  /* index of new string in symbol table */
      nodeo[2] =
      nodeo[3] = 0;  /* not in structure/union */
      node[0] = tree4(nodeo);
      node[1] = &string_type;  /* type is a literal string */
      ctlb(1);    /* a string needs it's own symbol table entry, emit EXTERN to CTEMP1 */
      ctlb(SSTATIC);  /* visibility = STATIC */
      ctlw(ordinal);  /* symbol table index number */
      ctlb(0);  /* name is null */
concatstring:
      ctlb(INITSTR);  /* start string in table *I/
      slen = 0;
      while (string[slen] || string[slen+1] != 0xff)  /* find EOS */
         slen++;
      string_type.length = slen;   /* set length */
      ctlb(slen);  /* and put in CTEMP1 */
      i = 0;
      while (i < slen)
         ctlb(string[i++]);  /* put string itself in CTEMP1 */
      tokit();
      if (heir == STRNG)  /* if another STRING */
         goto concatstring;  /* add it to init */
      ctlb(0xFF);
      ctlb(INITEND);  /* EOS in CTEMP1 */
      return 0;
      }
   node[0] = 0;  /* NULL, nothing to parse */
   return 0;
   }

/* create an abstract data type. if argument is not zero result type
   must go in table and argument points to thing address taken of. */

abstract(oldtype) 
   char *oldtype;
   {
   char oldstor, *oldloc, *oldt, *oldat, *absat;

   oldstor = addstor;  /* remember current type info */
   oldloc = addloc;
   oldt = addtype;
   oldat = addat;
   mfree->word = 0;  /* null link */
   (mfree+2)->byte = 0;  /* null name */
   addat = mfree + 3;  /* ptr to build area for data storage */
   addat->opcl = OPERAND;
   addat->nchain = 0;
   addat->nlen = 1;
   addat->nstor =
   addstor = SUNSTOR;  /* default: unknown type */
   addat->noff = 0;
   addloc = &addat->ntype[0];  /* ptr to type info */
   if (oldtype)  /* oldtype only set on "take-address-of" in heir24() */
      {
      *addloc++ = PTRTO;  /* make this a pointer */
      addtype = oldtype;  /* to previous type */
      }
   else
      {
      getab(0);  /* get data type specification */
      notch(')');  /* syntax check, closing parenthesis on typecast */
      }
   atype(addtype);  /* build type byte(s) */
   addstor = oldstor;  /* restore type informatoin */
   addloc = oldloc;
   addtype = oldt;
   absat = addat;
   addat = oldat;
   return (&absat->ntype[0]);  /* return abstract data type */
   }

/* Warning: recursive routine */
getab(char lvl)
   {
   char *otype;

   while (1)
      {
      if (ifch('*'))  /* pointer to data */
         {
         getab(1);
         *addloc++ = PTRTO;
         return;
         }
      otype = addtype;
      if (lvl)
         while ((heir == RESERVED && bvalue >= RCDECL && bvalue <= RPASCAL)
                                      || (heir == UNDEF) || (heir == OPERAND))
            tokit();  /* skip "nop" keywords, variable names */
      if (lvl)
         while (specs(0))  /* if a data type */
            ifch(',');  /* skip a comma */
      addtype = otype;
      if (heir == RELIPSE)
         tokit();
      if (ifch('('))
         {
         if (curch != ')')
            {
            getab(1);
            notch(')');
            continue;
            }
         tokit();
         *addloc++ = FUNCTION;
         }
      else
         if (ifch('['))
            {
            if (ifch(']'))
               {
               *addloc++ = PTRTO;
               return;
               }
            if (heir != CONSTANT)
               {
               error("missing dimension");
               return;
               }
            tokit();
            notch(']');
            *addloc++ = ARRAY;
            addloc->word = wvalue;
            addloc += 2;
            }
         else
            return;
      }
   }

oconst(con)  /* build a BYTE/WORD constant node (implicit small memory model) */
   char *con;
   {
   int nodeo[2];

   if (con > 255)
      nodeo[0] = CONST + (CINT << 8);
   else
      nodeo[0] = CONST + (CCHAR << 8);
   nodeo[1] = con;
   return tree2(nodeo);
   }


lookup_member(char *mptr)
   {
   while (mptr)  /* until end of chain */
      {
      char *nptr = mptr - mptr->nlen;  /* get ptr to name */
      if (strcmp(nptr, string) == 0)  /* correct name? */
         return heir = *(nameat = mptr);  /* yes, return heirarch */
      mptr = mptr->schain;  /* try next element */
      }
   return 0;  /* not found, return 0 */
   }

/* FIND_MEMBER -- find a member name. use the stag's chain list if possible,
                  otherwise look for any member name.   */

find_member(typeat)
   char *typeat;
   {
   char *name_at, *string_at, *typeis, *memat;

   typeis = typeat;  /* remember default type */
   if (heir == OPERAND || heir == UNDEF)  /* variable name or new name */
      {
      while (*typeis == PTRTO)
         typeis++;
      if (*typeis == CSTRUCT)
         {
         struct_type = typeis;
         memat = struct_type->ststagat;  /* address of stag */
         memat = memat->schain;  /* address of first member */
         if (lookup_member(memat))
            return 1;
         }
      else
         if (*typeis == CCHAR)
   /* NOTE: This implements the "DeSmet-ism" that a char * can reference *ANY*
    *       struct/union member without typecasting to the correct pointer
    *       type.  i.e.
    *          struct { char b; int w; }; char *p; main() { p->w = 2; }
    *       will put a WORD of 2 at offset 1 from the address pointed to
    *       by p.
    *
    *       This is used by the compilier, but I don't know if "general"
    *       code uses it.  The compiler should be "fixed" to not need this
    *       and this code removed.
    *
    *       Note that if multiple structructures/unions have an element with
    *       the same name, the "last" definition will be used.  i.e.
    *          struct { int a; char b; int w; };
    *          struct { long a; char b; int w; };
    *          char *p; main() { p->w = 2; }
    *       will set the word at (p+5) to 2, but
    *          struct { long a; char b; int w; };
    *          struct { int a; char b; int w; };
    *          char *p; main() { p->w = 2; }
    *       will set the word at (p+3) to 2.
    *
    *       Also, if the last referenced member is a member of a union,
    *       the offset will be 0, only the type will be used.
    */
            {
            wp = anon;  /* get ptr to list of all structures/unions */
            while (wp)
               {
               struct_type = wp + 1;  /* get ptr to structure type */
               memat = struct_type->ststagat;  /* address of stag */
               memat = memat->schain;  /* address of first member */
               if (lookup_member(memat))
                  return 1;
               wp = *wp;
               }
            }
         else
            {
            error("Need structure");
            return 0;
            }
      }
   error("member not in structure");
   return 0;
   }
