/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  DeSmet C is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; either version 2 of the License, or any
 *  later version.
 *
 *  DeSmet C is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/* c6.c               part 6 of medium c compiler */
/*
 *  Modification history:
 *    William Couture    10/12/05   Reformat/cleanup/comments
 *                                  changed 'heir' magic numbers to names
 */

#include "PASS1.H"
#include "NODES.H"

/* the "workhorse" token parser */
tokit()
   {
   do 
      ;
   while (tokone());  /* parse until valid token (not start of #define
                       * expansion or end of #include)
                       */
   }

/* build macro expansion string (i.e. __LINE__, __FILE__, etc, value) */
strngcpy(char *cp, char qs)
   {
   whitesp();
   nestfrom[nested++] = cur;
   cur =
   macfrom[nested] = macxp;
   if (qs)  /* put string in quotes */
      *macxp++ = '"';
   while (*macxp++ = *cp++)
      ;
   if (qs)  /* ending quote */
      *(macxp-1) = '"';
   else
      macxp--;
   *macxp++ = LF;
   }

/* handle ANSI-C built-in #define values */
builtinMac()
   {
   char xpnd[32];

   switch (bvalue)
      {
      case R_LINE:
         itoa(cline, xpnd, 10);
         strngcpy(xpnd,0);
         break;
      case R_FILE:
         strngcpy(name,1);
         break;
      case R_DATE:
         strngcpy(mdate,1);
         break;
      case R_TIME:
         strngcpy(mtime,1);
         break;
      case R_STDC:
         itoa(__stdc, xpnd, 10);  /* only standard C */
         strngcpy(xpnd,0);
         break;
      }
   }

/* parse token from input stream.  returns 1 (TRUE) if DEFINE expansion or
 * end of include file (we changed contents of cur, and so do not have a
 * valid token), else 0
 *
 * NOTE:  on return i is the length of the string found (via find())
 */
tokone()
   {
   char macOK;

   tokat = cur;
   if (!(macOK = ((curch = *cur++) != '$')))  /* do not macro expand a symbol preceeded by '$' */
      curch = *(tokat = cur++);
   heir = 0;

   switch (ltype[curch])
      {
      case LETTER:  /* A .. Z, _, a .. z */
         if (cur < savEnd)
            lastname = tokat;
         find(macOK);  /* parse token, see if we already know it */
         if (heir == DEFINED)  /* #define'ed symbol */
            {
            addnest();  /* build expanded string and start parsing it */
            whitesp();
            return 1;
            }
         if (heir == RESERVED)  /* reserved keyword (if, while, etc) */
            {
            if (bvalue >= R_LINE && bvalue <= R_STDC)
               {
               builtinMac();
               return 1;
               }
            if (!xkwd && bvalue >= RASM)  /* recognize extended keywords */
               heir = UNDEF;
            }
         break;
      case DIGIT:  /* 0 .. 9 */
         number();  /* it seems to be a number */
         break;
      case 4:  /* '!' */
         if (*cur == '=')
            {
            cur++;
            heir = 19;  /* HEIR_EQUALITY */
            bvalue = NE;
            break;
            }
         heir = 24;  /* HEIR_PREFIX */
         bvalue = NOT;
         break;
      case 5:  /* '"' */
         isstring();
         break;
      case 6:  /* '%' */
         heir = 23;  /* HEIR_MULDIV */
         opeq(MOD);
         break;
      case 7:  /* '&' */
         if (*cur == '&')
            {
            cur++;
            heir = 15;  /* HEIR_LOG_AND */
            break;
            }
         heir = 18;  /* HEIR_BIT_AND */
         opeq(AND);
         break;
      case 8:  /* ''' */
         charac();  /* single quote, this is a character */
         break;
      case 9:  /* '*' */
         heir = 23;  /* HEIR_MULDIV */
         opeq(MUL);
         break;
      case 10:  /* '+' */
         if (*cur == '+')
            {
            cur++;
            heir = 24;  /* HEIR_PREFIX */
            bvalue = PREI;
            break;
            }
         heir = 22;  /*  HEIR_ADDSUB */
         opeq(ADD);
         break;
      case 11:  /* ')' and ',' */
         break;
      case 12:  /* '-' */
         if (*cur == '-')
            {
            cur++;
            heir = 24;  /* HEIR_PREFIX */
            bvalue = PRED;
            break;
            }
         if (*cur == '>')
            {
            cur++;
            heir = 25;  /* HEIR_STRUCTREF */
            break;
            }
         heir = 22;  /* HEIR_ADDSUB */
         opeq(SUB);
         break;
      case 13:  /* '.' */
         if (*cur == '.')
            {
            heir = RESERVED;
            if (*++cur == '.')
               {
               cur++;
               curch =
               bvalue = RELIPSE;
               }
            else
               if (xkwd)  /* recognize extended keywords */
                  curch = bvalue = RDOTDOT;
               else
                  {
                  error("case range option not on");
                  heir = 0;
                  }
            break;
            }
         if (ltype[*cur] == DIGIT)
            number();  /* floating-point number */
         else
            heir = 25;  /* HEIR_STRUCTREF */
         break;
      case 14:  /* '/' */
         heir = 23;  /* HEIR_MULDIV */
         opeq(DIV);
         break;
      case 15:  /* '<' */
         if (*cur == '<')
            {
            cur++;
            heir = 21;  /* HEIR_SHIFT */
            opeq(SHL);
            break;
            }
         if (*cur == '=')
            {
            cur++;
            bvalue = LE;
            }
         else
            bvalue = LT;
         heir = 20;  /* HEIR_COMPARE */
         break;
      case 16:  /* '=' */
         if (*cur == '=')
            {
            cur++;
            heir = 19;  /* HEIR_EQUALITY */
            bvalue = EQ;
            break;
            }
         heir = 12;  /* HEIR_OPEQ */
         bvalue = ASGN;
         break;
      case 17:  /* '>' */
         if (*cur == '>')
            {
            cur++;
            heir = 21;  /* HEIR_SHIFT */
            opeq(SHR);
            break;
            }
         if (*cur == '=')
            {
            cur++;
            bvalue = GE;
            }
         else
            bvalue = GT;
         heir = 20;  /* HEIR_COMPARE */
         break;
      case 18:  /* '^' */
         heir = 17;  /* HEIR_BIT_XOR */
         opeq(XOR);
         break;
      case 19:  /* '|' */
         if (*cur == '|')
            {
            cur++;
            heir = 14;  /* HEIR_LOG_OR */
                /* BUG: need a 'break;' here, else ||= is a legal statement */
            }
         else  /* BUG: with the break above, this else is not needed */
            heir = 16;  /* HEIR_BIT_OR */
         opeq(OR);
         break;
      case 20:  /* ^Z */
         if (incnext == 0)  /* no more #include's to exit from */
            {
            cur--;  /* forever CONTZ */
            break;
            }
         incnext--;  /* end of this include */
         close(file);
         _lmove(SAVSIZE, 0, incPara[incnext], name, _showds());
         SKIPL();
         return 1;
      case 21:  /* ^J */
         cur--;
         dolf(0);  /* skipit == 0, skip leading spaces & do preprocessor */
         return 1;
      default:
         ;
      }
   if (eopt)  /* tokens in pass 1, debug info in pass 2 */
      {
      obnum(heir);
      oc(' ');
      oc(curch);
      oc(' ');
      obnum(bvalue);
      oc(' ');
      ohw(nameat);
      oc(' ');
      ohw(cur);
      oc(' ');
      obnum(nested);
      ocrlf();
      }
   whitesp();
   return 0;
   }

/* if name not found, set nameat to structure with type OPERAND,
   nstor of SUNSTOR and nchain to zero. set heir to UNDEF. */
/* if name found, set heir to type and nameat to address of type. */
/* if reserved, set bvalue to rvalue. */
/* i is the length of the symbol, used by search() and newname() */

find(char macOK)
   {
   cur--;
   i =
   hashno = 0;
   while ((ltype[*cur] <= DIGIT) && i < 31)
        /* LETTER or DIGIT, we know 1st char was a LETTER */
      {
      hashno += *cur;
      string[i++] = *cur++;
      }
   hashno += i;
   if (i == 31)  /* ignore characters past 31 */
      while (ltype[*cur] <= DIGIT)
         cur++;
   string[i] = 0;  /* set EOS */
   if (eopt)  /* tokens in pass 1, debug info in pass 2 */
      {
      os(string);
      oc(' ');
      }
   hashno &= 31;  /* get index into hash table(s) */
   search(macOK);
   }

/* i is the length of the symbol, set by find() */
search(char macOK)
   {
   if (macOK)  /* OK to expand a #define'd symbol */
      {
      bptr = machash[hashno];
      while (bptr)
         {
         j = -1;
         wp = bptr;
         bptr += 2;
         while (*bptr++ == string[++j])
            ;
         if (j > i)  /* if we have found this symbol */
            {
            int ndx = nested;

            nameat = wp;
            nameat += i + 3;
            while (ndx)  /* backtrack through currently active macro expansions */
               if (macname[ndx--] == nameat)  /* if already in this expansion */
                  goto recur;  /* don't allow recursive expansions */
            heir = *nameat;  /* get heirarchy (DEFINED) */
            return;
            }
         bptr = *wp;  /* try next entry */
         }
      }
recur:
   bptr = hash[hashno];
   while (bptr)  /* for this chain */
      {
      j = -1;
      wp = bptr;
      bptr += 2;  /* sizeof(char *) */
      while (*bptr++ == string[++j])
         ;
      if (j > i)  /* up to & including EOS of bptr string */
         {
         nameat = wp;
         nameat += i + 3;  /* start + strlen(string) + sizeof(char *) + EOS */
         heir = *nameat;  /* get opcl from structure */
         if (heir == RESERVED)
            bvalue = nameat->rvalue;
         return;
         }
      bptr = *wp;  /* try next chain element */
      }
   heir = UNDEF;  /* can't find it, must be new symbol */
   }


/* i is the length of the symbol, set by find() */
/* memory format of symbol table:
 *   xxyy         hash link
 *   'abc...'00   symbol (0 terminated, length varies)
 *   cc           class (== heir) RESERVED (==1), STAG(==2), DEFINED(==3),
 *                       DEFPARM(==4) OPERAND(==26)
 *   ....         varies depending on class, see PASS1.H
 */
/* at exit, nameat points to the class byte in the new symbol table entry */
newname()
   {
   wp = mfree;
   if (in_stru == 0)  /* if not defining structure elements */
      {
      *wp = hash[hashno];  /* add hash link */
      hash[hashno] = wp;
      }
   j = 0;
   nameat = wp + 1;
   do
      *nameat++ = string[j++];
   while (j <= i);
   heir = UNDEF;
   nameat->opcl = OPERAND;
   nameat->nchain = 0;
   nameat->nlen = i + 1;
   nameat->ntype[0] = CINT;
   }

/* distinguish "op-assignment" (i.e. +=, etc) from operator (i.e. +, etc) */
opeq(val)
   int val;
   {
   if (*cur == '=')  /* if operator assignment (i.e. +=, -= /=, etc) */
      {
      cur++;  /* skip this char */
      heir = 12;  /* HEIR_OPEQ */
      bvalue = val + AADD - ADD;
      }
   else
      bvalue = val;
   }

/* skip parenthesized arguments */
/* warning: recursive function */
skipa()
   {
   char inquote = 0, ch;

   while ((ch = *cur++) != ')' || inquote) 
      switch (ch)
         {
         case '\'':
         case '"':
            if (inquote == 0)
               inquote = ch;
            else
               if (inquote == ch && *(cur-2) != '\\')  /* matching, un-escaped quote */
                  inquote = 0;
            continue;
         case '(':  /* nested parenthesis, skip this before finishing our search */
            skipa();
            continue;
         case LF:
            dolf(1);  /* skipit == 1, DO NOT skip leading spaces & DO NOT do preprocessor */
         }
   }


static char *macmove(char *beg, char *end, char **free)
   {
   int len = end - beg;

   *free -= len;
   _move(len, beg, *free);
   return *free;
   }

/* build #define expansion to be parsed into program */
/* Warning: recursive function */
addnest()
   {
   char *defat, *defstr, *mname,
        *nptr = wp + 1,
            /* NOTE: wp is still set from find(), points to start of #define'ed symbol entry */
        nlen = cur - tokat;
   char narg, inquote, plevel;
   char *argfrom[MAXNEST],
        argstk[1024],
        *ap, *sp;

   ap = &argstk[1024];  /* end of argstack */
   sp = argstk;  /* init stack pointer */
      /* NOTE: sp starts at the "bottom" of our local stack and grows
       *       upwards.  It is reset for every parameter, this makes
       *       the bottom of the stack temp storage for freshly parsed
       *       macro arguments.
       *
       *       ap starts at the top of our local stack, and grows downwards
       *       for every macro argment "pushed" on the stack.
       *
       *       Hopefully, they never meet in the middle...
       */
   mname =
   defat = nameat;
   narg = 
   inquote =
   plevel = 0;

/* dargs of 255 means none. 0 means need empty parens. more is count. */
   if (defat->dargs != 255)  /* if #define args */
      {
      whitesp();
      if (cur > argstk)  /* if we are parsing our local argstk[] */
         {
         curch = *cur++;  /* get char from our local argstk[] */
         if (curch == LF)  /* don't skip past LF */
            cur--;
         }
      else
         tokit();  /* use regular token parser */
      if (curch != '(')  /* I don't know how to get here without curch == '(' */
         {
         if (nested+1 >= MAXNEST)
            {
            error("define too deep");
            return;
            }
         nestfrom[nested] = cur;  /* remember pointer */
         nestfrom[++nested] = macxp;
                /* remember ptr into macro expansion buf maxExp[] */
         macfrom[nested] = macxp;
#ifdef MacHiWater
         if (nested > maxNest)
            maxNest = nested;
#endif
/*         *macxp++ = '$'; */
         while (*nptr)  /* copy name */
            *macxp++ = *nptr++;
         *macxp++ = LF;
         cur = nestfrom[nested];
#ifdef MacHiWater
         if (macxp > MacHiWaterP)
            MacHiWaterP = macxp;
#endif
         return;
         }  /* if (curch != '(') */
      if (*cur == ')')  /* empty arg case */
         {
         tokit();
         if (defat->dval == DEFEND)  /* and no parameters expected */
            return;
         }
      else
         if (defat->dval == DEFEND)  /* no parameters expected, skip them */
            {
            skipa();
            return;
            }
         else
            {
            goto setarg;
               /* NOTE: this goto is used to jump into the middle
                *       of the following loop for the first pass.
                *
                *       It could be replaced with
                *          if (++narg == MAXNEST)
                *             {
                *             error("macro buffer overflow");
                *             return;
                *             }
                *          sp = argstk;
                */
            while (1)
               {
               whitesp();  /* skip whitespace */
               tokat = cur;
               if ((curch = *cur++) == LF)  /* EOL? */
                  {
                  if (cur > argstk)  /* parsing local expansion */
                     {
                     error("Unexpected EOF in macro argument");
                     real_exit(2);
                     }
                  dolf(1);  /* skipit == 1, DO NOT skip leading spaces & DO NOT do preprocessor */
                  continue;
                  }
               if (ltype[curch] == LETTER)  /* A .. Z, _, a .. z */
                  while (ltype[*cur] == LETTER)  /* A .. Z, _, a .. z */
                     cur++;
               if (curch == ')' && plevel-- == 0)  /* end of macro parameters */
                  {
                  if (*(sp-1) != ' ')
                     *sp++ = ' ';
                  *sp++ = LF;  /* end this macro expansion piece */
                  argfrom[narg] = macmove(argstk, sp, &ap);
                  break;  /* exit macro expansion loop */
                  }
               if (curch == ',' && plevel == 0)  /* end of this macro argument */
                  {
                  if (*(sp-1) != ' ')
                     *sp++ = ' ';
                  *sp++ = LF;  /* end this macro expansion piece */
                  argfrom[narg] = macmove(argstk, sp, &ap);
setarg:           if (++narg == MAXNEST)  /* too many arguments? */
                             /* NOTE: the first argument is narg == 1
                              *       this matches the macro parsing that the
                              *       first argument is 1, etc.
                              */
                     {
                     error("macro buffer overflow");
                     return;
                     }
                  sp = argstk;  /* use our local stack */
                  continue;  /* expand next macro argument */
                  }
               if (curch == '(')  /* make sure parenthesis balance */
                  plevel++;
               while (tokat < cur)  /* get this token */
                  {
                  char ch;

                  if (ltype[(ch = *tokat)] == SPACE)  /* ^I, ^L, ^M, ' ' */
                     {
                     tokat++;
                     continue;  /* collapse whitespace to single char at most */
                     }
                  if (ch == '"' || ch == '\'')  /* quoted char/string, copy entire char/string */
                     {
                     *sp++ = *tokat++;
                     while (*tokat != ch || *(tokat-1) == '\\')  /* escaped quote doesn't count */
                        *sp++ = *tokat++;
                     cur = tokat + 1;
                     }
                  *sp++ = *tokat++;  /* copy single char */
                  }
               if (ltype[*cur] == SPACE)  /* ^I, ^L, ^M, ' ' */
                  *sp++ = ' ';   /* (multiple) whitespace, just one use char */
               }  /* while (1) */
            }  /* if (defat->dval == DEFEND) ... else */
xpand:
      if (narg != defat->dargs)
         {
         error("wrong number of arguments");
         return;
         }
      }
   if (defat->dval == DEFEND)
      return;
   _setsp(ap - 1);  /* set SP register! */
      /* NOTE: this actually sets the SP register.   Brrrr.....
       *       it will be restored at the end of the function, where
       *          MOV SP,BP
       *       is part of the function exit code.
       *
       *       Note that BP is unchanged, so our locals are still addressable.
       *
       *       Why is this here?  To save stack space (small memory model).
       *       Since this function is recursive, and taking up a large
       *       amount of stack ( > 1K),  this reclaims unused stack space
       *       from argstk[1024].  Note that char *ap and char *sp are
       *       "thrown away", since they are not used after this point.
       */
   nestfrom[nested] = cur;  /* remember last piece */
   if (nested+1 >= MAXNEST)
      {
      error("define too deep");
      return;
      }
   nestfrom[++nested] = macxp;  /* remember ptr into expansion */
   macfrom[nested] = macxp;
#ifdef MacHiWater
   if (nested > maxNest)
      maxNest = nested;
#endif
   defstr = &defat->dval;  /* get ptr to expansion */
   *macxp++ = LF;  /* set EOS for expansion */
   macname[nested] = mname;  /* #define name */
   while ((curch = *defstr) != DEFEND)  /* until end of expansion */
      {
      char *bp;

      macxp--;  /* remove trailing LF */
      if (curch == DEFSTR)  /* copy #define'ed value i.e. the 'B' of #define A B */
         {
         defstr++;  /* skip DEFSTR */
         while ((*macxp++ = *defstr++) != LF)  /* copy string to EOS */
            ;
         macOp = DEFSTR;  /* remember: we just parsed a DEFSTR */
         }
      else
         if (curch == DEFPAS)  /* concatination  i.e. value of 'c' in #define A(b,c) b##c */
            {
            while (ltype[*(macxp-1)] == SPACE)  /* ^I, ^L, ^M, ' ' */
               macxp--;  /* skip whitespace */
            if (*++defstr == DEFSTR)  /* if 1st arg is a string */
               {
               while (ltype[*defstr] == SPACE)  /* ^I, ^L, ^M, ' ' */
                  defstr++;  /* skip whitespace */
               if (*defstr == '$')  /* skip "no macro expansion" flag */
                  defstr++;
               while ((*macxp++ = *defstr++) != LF)  /* copy literal */
                  ;
               }
            else  /* copy macro argument */
               {
               bp = argfrom[*defstr++];  /* get ptr to correct argument */
               while (ltype[*bp] == SPACE)  /* ^I, ^L, ^M, ' ' */
                  bp++;  /* skip whitespace */
               if (*bp == '$')  /* skip "no macro expansion" flag */
                  bp++;
               while ((*macxp++ = *bp++) != LF)  /* copy string */
                  ;
               }
            macxp--;  /* backup over LF */
            while (ltype[*(macxp-1)] == SPACE)  /* ^I, ^L, ^M, ' ' */
               macxp--;  /* remove trailing whitespace */
            *macxp++ = LF;  /* add EOS */
            macOp = DEFPAS;  /* remember: we just parsed a DEFPAS */
            }  /* if (curch == DEFPAS) */
         else
            if (curch == DEFSFY)  /* stringify i.e. #define A(b) printf( #b) */
                                  /* Note the space after "printf(" in the
                                   * above line.  See the bug list at the start
                                   * of C88.C
                                   */
               {
               bp = argfrom[*++defstr];  /* get ptr to correct argument */
               *macxp++ = '"';  /* add leading quote */
               while (*bp != LF && ltype[*bp] == SPACE)  /* ^I, ^L, ^M, ' ' */
                  bp++;  /* skip leading whitespace */
               while (*bp != LF)  /* copy to EOS */
                  {
                  if (*bp == '$' && ltype[*(bp+1)] == LETTER)  /* A .. Z, _, a .. z */
                     bp++;  /* skip "do not expand" flag */
                  if (*bp == '"' || *bp == '\\')  /* if quote, escape it */
                     *macxp++ = '\\';
                  *macxp++ = *bp++;
                  }
               while (ltype[*(macxp-1)] == SPACE)  /* ^I, ^L, ^M, ' ' */
                  macxp--;  /* trim trailing whitespace */
               *macxp++ = '"';  /* add ending quote */
               *macxp++ = LF;  /* and EOS */
               defstr++;  /* skip argument number */
               macOp = DEFSFY;  /* remember: we just parsed a DEFSFY */
               }
            else
               if (curch == DEFCPY)  /* copy literal i.e. value of 'b' in #define A(b,c) b##c */
                  {
                  bp = argfrom[*++defstr];  /* get ptr to correct argument */
                  while (*bp != LF)  /* until EOS */
                     {
                     if (*bp == '$' && ltype[*(bp+1)] == LETTER)  /* A .. Z, _, a .. z */
                        bp++;  /* skip "do not expand" flag */
                     *macxp++ = *bp++;  /* copy argument */
                     }
                  defstr++;  /* skip argument number */
                  macOp = DEFCPY;  /* remember: we just parsed a DEFSFY */
                  }
               else  /* a macro parameter, copy it */
                  {
                  char oldnest = nested;

                  cur = argfrom[*defstr++];  /* get ptr to argument value */
                  while (1)
                     {
                     tokat = cur;  /* remember start of token */
                     if ((curch = *cur++) == LF)  /* EOS, we're done */
                        break;
                     if (curch == '$')  /* if "do not macro expand" flag */
                        {
                        while (ltype[*cur] == LETTER)  /* A .. Z, _, a .. z */
                           *macxp++ = *cur++;  /* copy string */
                        continue;
                        }
                     else
                        if (ltype[curch] == LETTER)  /* A .. Z, _, a .. z */
                           {
                           if (strncmp(nptr, tokat, nlen) == 0)  /* current macro again */
                              {
                              wp = nptr - 2;  /* set pointers */
                              cur = tokat + nlen;
                              nameat = defat;
                              heir = DEFINED;
                              }
                           else
                              find(1);  /* try to find this symbol */
                           if (heir == DEFINED
                                 || (heir == RESERVED && bvalue >= R_LINE && bvalue <= R_STDC))
                                     /* #define or built-in #define */
                              {
                              if (heir == DEFINED)  /* #define, expand it */
                                 addnest();
                              else
                                 builtinMac();  /* parse built-in macro like __LINE__ */
                              if (nested != oldnest)  /* if "pushed" info */
                                 {
                                 cur = nestfrom[--nested];  /* "pop" current parsing info */
                                 nestfrom[nested] = macfrom[nested];
                                 }
                              macxp--;  /* backup over '\0' EOS */
                              if (macOp == DEFPAS)  /* if we just did a concatination */
                                 whitesp();  /* skip trailing whitespace */
                              continue;  /* continue expansion */
                              }
                           while (tokat < cur)  /* just copy token */
                              *macxp++ = *tokat++;
                           }
                        else
                           if (curch == '"' || curch == '\'')  /* quoted string/char */
                              {
                              *macxp++ = *tokat++;  /* copy quote */
                              while (*tokat != curch || *(tokat-1) == '\\')
                                 *macxp++ = *tokat++;  /* copy string including escpaed quotes */
                              *macxp++ = curch;  /* add ending quote */
                              cur = tokat + 1;  /* update parsing */
                              continue;
                              }
                           else
                              *macxp++ = curch;  /* just copy the single char */
                     }
                  *macxp++ = LF;  /* set EOS in expansion string */
                  macOp = DEFPARM;  /* remember: we just parsed a DEFPARM (!= DEFPAS) */
                  }
      }
   cur = nestfrom[nested];  /* start parsing #define expansion */
#ifdef MacHiWater
   if (macxp > MacHiWaterP)
      MacHiWaterP = macxp;
#endif
   }
   
/* skip a parenthesized expression. */
/* NOTE: I don't think this is called anywhere */
skipp()
   {
   while (*++cur != ')')
      if (*cur == '(')
         skipp();
   }

/* this seems to be a number, parse it */
number()
   {
   char ch, n;
   double strtod();

   heir = CONSTANT;
   dvalue = 0;
   if (curch == '0' && (ch = toupper(*cur)) != 'L' && ch != 'U' && ch != '.')
          /* leading 0, check next character for special cases */
      {
      if (ch == 'X')  /* starts with '0x', a hex number */
         {
         n = 0;  /* character count */
         while (1)
            {
            ch = *++cur;
            if (ch >= 'a')  /* lower case */
               ch -= 32;  /* force upper case */
            if (ch >='A' && ch <= 'F')  /* upper-case HEX digit? */
               ch -= '7';  /* yes, change to hex nibble */
            else
               if (ltype[ch] == DIGIT)  /* numeric digit? */
                  ch -= '0';  /* yes, change to hex nibble */
               else
                  break;  /* else end of number */
            dvalue = (dvalue << 4) + ch;
            if (n++ == 7)  /* if a LONG value is full */
               *((long *)(&fvalue)+1) = dvalue;  /* save in "high dword" of fvalue */
                   /*  NOTE: since fvalue is
                    *           unsigned fvalue[4];
                    *        this is implicitly little-endian
                    */
            }
         if (n > 8 && n != 16)  /* if more than a LONG but not a DOUBLE */
            error("illegal double constant");
         else
            if (n == 16)  /* if a DOUBLE */
               {
               *((long *)(&fvalue)) = dvalue;  /* save "low dword" of fvalue */
               heir = FDCONSTANT;
                  /* NOTE: this allows a float to be specified as a 16-digit
                   *       hex number for absolute accuracy (no rounding/
                   *       truncation problems convering the ASCII floating
                   *       pointer number to it's internal binary
                   *       representation).
                   */
               return;
               }
         }
      else
         {
         while ((ch = *cur-'0') < 10)
                /* BUG: leading 0 means octal constant. */
            {
            cur++;
            /* BUG: from above, need something like:
             *         if (ch > 7)
             *            {
             *            error("Invalid octal digit");
             *            break;
             *            }
             *         here.
             */
            dvalue = (dvalue << 3) + ch;
            }
         }
      if (dvalue > 65535 || dvalue < 0)
         heir = LCONSTANT;
      }
   else
      {
      if (*tokat == '.')  /* floating point number */
         cur = tokat;
      dvalue = curch - '0';  /* make sure we get leading digit */
      while ((ch = *cur-'0') <= 9)  /* floats do not enter this code */
         {
         dvalue = dvalue * 10 + ch;
         cur++;
         }
      if ((*cur == '.' && *(cur+1) != '.') || *cur == 'e' || *cur == 'E')
                /* if float & not a range, or 'E' notation number */
         {
/*         cur = tokat + _finput(tokat,&fvalue,100); */
         *((double*)(&fvalue)) = strtod(tokat, &cur);
         heir = FDCONSTANT;
         while ((ch = toupper(*cur)) == 'L' || ch == 'F')
            {
            if (ch == 'F')
               {
               heir = FFCONSTANT;
               *((float*)(&fvalue)) = *((double*)(&fvalue));
               }
            cur++;
            }
         return;
         }
      if (dvalue > 32767)
         heir = LCONSTANT;
      }
   while ((ch = toupper(*cur)) == 'L' || ch == 'U')
      {
      if (ch == 'L')
         heir = LCONSTANT;
      cur++;
      }
   wvalue = dvalue;
   }

/* token starts with a double quote, parse a string */
isstring()
   {
   i = 0;
   heir = STRNG;
   while (*cur != '"' && i < 255)  /* until matching quote or too big */
      {
      if (*cur == '\n' || *cur == CONTZ)
         {
         error("unmatched \"");
         cur--;  /* remain at EOL / Ctrl-Z */
         break;
         }
      string[i++] = getach();  /* allow for escaped chars */
      }
   if (i == 255)  /* too big, report error */
      {
      error("string too long");
      return;
      }
   cur++;  /* skip end quote */
   string[i] = 0;  /* set EOS for C */
   string[i+1] = 0xff;  /* set EOS for symbol table */
   }

/* single quote, this is a character */
charac()
   {
   heir = CONSTANT;  /* a character is a CONSTANT */
   if (*cur == '\'')  /* blank char: '' */
      {
      wvalue = 0;
      }
   else
      {
      wvalue = getach();  /* allow for escaped characters */
      if (*cur != '\'' && *cur != LF)  /* can pack two chars in a WORD value */
         {
         wvalue <<= 8;
         wvalue += getach();
         }
      }
   if (*cur++ != '\'')  /* if not matching end single quote */
      error("missing '");
   }

/* get character constant, process escape sequences */
getach()
   {
   char ch, och, max, xd;

getagain:
   ch = *cur++;  /* get character */
   if (ch == '\\')  /* escaped? */
      {
      och =
      ch = *cur++;
      if (ch >= 'a')  /* if lower case */
         ch -= 32;  /* force upper case */
      switch (ch)
         {
         case '\r':
            cur++;  /* '\r' is not EOL, skip it */
            /* NO BREAK:  fall through */
         case '\n':
            if (addparm == 0 && addproto == 0)
               lineinc++;
            goto getagain;
            /* NOTE: I have no idea how to get '\r' or '\n' after a
             *       '\' past the dolf() parsing.
             */
         case 'A':  /* '\a' */
            ch = 7;
            break;
         case 'B':  /* '\b' */
            ch = 8;
            break;
         case 'F':  /* '\f' */
            ch = 12;
            break;
         case 'N':  /* '\n' */
            ch = LF;
            break;
         case 'R':  /* '\r' */
            ch = 13;
            break;
         case 'T':  /* '\t' */
            ch = 9;
            break;
         case 'V':  /* '\v' */
            ch = 11;
            break;
         case 'X':  /* '\xddd' */
            max = 3;
            ch = 0;
            while (max--)
               {
               xd = *cur++;
               if (xd >='a')
                  xd -= 32;
               if (xd >='A' && xd <= 'F')
                  xd -= '7';
               else
                  if (ltype[xd] == DIGIT)
                     xd -= '0';
                  else
                     {
                     cur--;
                     break;
                     }
               ch = (ch << 4) + xd;
               }
            break;
         default:  /* '\ooo' (octal), or just returned escaped char */
            if (ch >= '0' && ch <= '7')
               {
               max = 2;
               ch -= '0';
               while (*cur >= '0' && *cur <= '7' && max--)
                  ch = (ch << 3) + *cur++ - '0';
               }
            else
               ch = och;
         }
      }
   return ch;
   }

/* NOTE: whitesp() may not change the global variable i, which contains
 *       the length of the current symbol.  This is set in find(), and
 *       used by search().
 */
whitesp()
   {
   char ch;
   int bline = cline;
   static char eoferr[] = "EOF within comment beginning on line #    ";
       /* NOTE: the line number will be placed in this string starting at
        *       the '#'.  Is 5 digits enough?
        */

   while (1)
      {
      if (ltype[*cur] == SPACE)  /* ^I, ^L, ^M, ' ' */
         cur++;
      else
         if (*cur == '/' && (cur+1)->byte == '/')  /* ANSI-C '//' line comment */
            {
            cur = lineEnd - 1;
            dolf(1);  /* skipit == 1, DO NOT skip leading spaces & DO NOT do preprocessor */
            }
         else
            if (*cur == '/' && (cur+1)->byte == '*')  /* normal K&R comment start */
               {
               cur += 2;
               while (1)
                  {
                  ch = *cur;
                  while (ch != LF && ch != '*' && ch != 26)
                     ch = *++cur;
                  if (ch == LF)
                     dolf(1);  /* skipit == 1, DO NOT skip leading spaces & DO NOT do preprocessor */
                  else
                     if (ch == 26)  /* Ctrl-Z, EOF */
                        {
                        cur--;
                        ltoa((long)bline, &eoferr[37], 10);
                        error(eoferr);
                        return;
                        }
                     else
                        if (*++cur == '/')
                           {
                           cur++;
                           break;
                           }
                  }
               }
            else
               return;
      }
   }
