/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  DeSmet C is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; either version 2 of the License, or any
 *  later version.
 *
 *  DeSmet C is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/* c7.c               part 7 of medium c compiler */
/*
 *  Modification history:
 *    William Couture    10/12/05   Reformat/cleanup/comments
 *                                  changed 'heir' magic numbers to names
 */

#include "PASS1.H"
#include "NODES.H"
#if CHECK
#include "OBJ.H"
#endif

/* display an error/warning message */
e_msg(reale,str)
   int reale;
   char *str;
   {
   onum(cline);  /* show line number */
   oc(' ');
   if (see_exit == 0 && reale >= 0)  /* not in SEE editor, and not a #error */
      {
      bptr = lineBeg;  /* get ptr to start of this line */
      do
         {
         if (bptr == tokat)  /* if this is the problem token */
            os(" $$ ");  /* flag it */
         if (*bptr == 9)  /* if a TAB */
            os("    ");  /* just use 4 spaces */
         else
            if (*bptr != CR && *bptr != CONTZ)  /* if not EOL */
               oc(*bptr);  /* display character */
         } while ( *bptr != CONTZ && *bptr++ != LF);  /* until EOL */
      os("      ");  /* be neat */
      }
   if (incnext)  /* if we are in an #include file */
      {
      os("file: ");  /* show current file */
      os(name);
      os("    ");
      }
   if (reale > 0)  /* if an actual error */
      {
      os(" error:");
      nerr++;  /* update error count */
      }
   else
      if (reale == 0)  /* just a warning */
         os(" warning:");
      else
         os("#error:");  /* reale == -1, this is a #error */
   os(str);  /* show associated message */
   if (see_exit)  /* if running out the SEE editor */
      real_exit(2);  /* return there */
   ocrlf();  /* else be neat */
   }

error(str)
   int str;
   {
   if (tokat)  /* if we are parsing source code or not multiple error(s) */
      e_msg(1,str);
   if (curch == ';' || curch == CONTZ)  /* logical end of statement or EOF */
      {
      tokat = 0;  /* flag: we just processed an error */
      return;
      }
   if (tokat == 0)  /* if not parsing anything or previous error(s) */
      {
      tokit();
      return;
      }
   while (1)  /* try to re-sync past error */
      {
      while (*cur != LF && *cur != ';' && *cur != CONTZ && *cur != '{'
                                                        && *cur != '}')
         cur++;
      if (*cur == LF)
         dolf(1);  /* skipit == 1, DO NOT skip leading spaces & DO NOT do preprocessor */
      else
         {
         tokit();
         tokat = 0;  /* flag: we just processed an error */
         return;
         }
      }
   }

/* a common error message */
nooper()
   {
   error("missing operand");
   }

/* show error message, skip rest of line instead of re-syncing */
error_l(str)
   int str;
   {
   e_msg(1,str);
   SKIPL();
   }

/* show warning message */
warning(str)
   int str;
   {
   e_msg(0,str);
   nwarn++;  /* update warning count */
   }

typeof(ptr)
   char *ptr;
   {
   int val;

   if (*ptr == PTRTO && is_big == 0)  /* !large memory model */
      val = CUNSG;
   else
      val = *ptr;
   return val << 8;
   }

unsigned dsize(ptr)
   char *ptr;
   {
   int tot;

   if (*ptr == ARRAY)
      {
      wp = ptr + 1;
      tot = *wp;
      if (tot == -1)
         tot = 0;
      return (tot * dsize(ptr+3));
      }
   if (*ptr == CCHAR || *ptr == CSCHAR)
      tot = 1;
   else
      if (*ptr == CLONG || *ptr == CFLOAT)
         tot = 4;
      else
         if (*ptr == CDOUBLE)
            tot = 8;
         else
            if (*ptr == CSTRUCT)
               {
               wp = ptr + 1;
               wp = *wp;
               tot = ((void *)wp)->staglen;
               if (tot == 0)
                  error("undefined structure");
               }
            else
               if (is_big && *ptr == PTRTO)  /* large memory model */
                  tot = 4;
               else
                  if (*ptr == FUNCTION)
                     return dsize(ptr+1);
                  else
                     tot = 2;
   return tot;
   }

/* stuff for SEE interface. */

static unsigned see_off, see_seg, see_ret, see_bp;
static char see_msg[80], see_index;

see_addr(ptr)
   char *ptr;
   {
   see_exit = 1;
   see_seg = see_hex(&ptr);
   see_off = see_hex(&ptr);
   }

see_hex(ptrptr)
   char **ptrptr;
   {
   char *ptr,ch;
   int num;

   ptr = *ptrptr;
   num = 0;
   while (1)
      {
      ch = *ptr++;
      if (ch >= '0' && ch <= '9')
         num = num * 16 + ch - '0';
      else
         if (ch >= 'a' && ch <= 'f')
            num = num * 16 + ch - 'a' + 10;
         else
            if (ch >= 'A' && ch <= 'F')
               num = num * 16 + ch - 'A' + 10;
            else
               {
               *ptrptr = ptr;
               return num;
               }
      }
   }

/* SEE_CALL -- SEE interface routine. */

see_call(p1,p2,p3)
   int p1,p2,p3;
   {
#asm
   pop    word see_bp_
   pop    word see_ret_
   lcall  dword see_off_
   push   word see_ret_
   push   word see_bp_
#endasm
   }

/* REAL_EXIT -- do an exit or a return to SEE. */

real_exit(cc)
   int cc;
   {
   if (see_exit)
      {
      see_call(2);
      if (cc)
         {
         if (copt)  /* add checkout info to object file */
            unlink(objname);
         see_call(-1,see_msg,incnext ? lastline : cline);
         }
      }
   exit(cc);
   }

/* scale value for sizeof pointer target */
scaled(ptr)
   char *ptr;
   {
   int val;

   if (*ptr == ARRAY)
      val = dsize(ptr+3);  /* get size of array element */
   else
      if (*ptr == PTRTO)
         val = dsize(ptr+1);  /* get size of pointer target */
      else
         val = 1;  /* else no scaling */
   return val;
   }

#if CHECK

dumpsym() 
   {
   char ch, *lastt;
   int num;

   bptr = _memory();  /* start of symbol table */
   while (bptr < mfree)
      {
      ohw(bptr);  /* symbol table element address */
      oc(' ');
      ohw(bptr->word);  /* link address */
      oc(' ');
      bptr += 2;
      while (*bptr)  /* symbol name */
         oc(*bptr++ & 0x7f);
      oc(' ');
      lastt = ++bptr;
      switch (*bptr)  /* entry type */
         {
         case RESERVED:  /* reserved keyword (if, while, else...) */
            os("RESERVED ");
            obnum(*++bptr);
            bptr++;
            break;
         case STREF:  /* structure reference, structure not yet defined */
         case STAG:  /* structure definition */
            os("STAG ");
            ohw(bptr->schain);
            oc(' ');
            onum(bptr->staglen);
            bptr += 5;
            break;
         case DEFINED:  /* #define value */
            os("DEFINED ");
            /* bptr += 2; Replaced 12/14/05 to fix DEFINE symbol dump */
            ohw(bptr->dchain);
            bptr += 4;  /* skip defcl, dchain, dnlen */
            oc(' ');
            ohb(*bptr++);  /* show num args */
            oc(' ');
            while (*bptr != DEFEND)
               {
               if (*bptr != DEFSTR)
                  {
                  oc(' ');
                  onum(*bptr++);
                  oc(' ');
                  }
               else
                  {
                  bptr++;
                  do
                     if (*bptr >= ' ')
                        oc(*bptr);
                  while (*bptr++ != LF);
                  }
               }
            bptr++;
            break;
         case OPERAND:  /* variable definition */
            os("OPERAND ");
            ohw(bptr->nchain);
            oc(' ');
            onum(bptr->nlen);
            oc(' ');
            onum(bptr->nstor);
            oc(' ');
            onum(bptr->noff);
            bptr += 7;
            listtype();
            break;
         default:  /* usually a parsing error by this routine, it's
                    * pretty simple-minded.
                    */
            os("Mystery Symbol ");
            ohw(bptr->word);
         }
      ocrlf();
      if (bptr->word == 0xffff)  /* I don't understand this */
         bptr = lastt + 18;
      }

   os("Normal End");
   ocrlf();
   }

listtype() 
   {
   char tp;

   do
      {
      oc(' ');
      tp = *bptr++;
      switch (tp)
         {
         case CCHAR:
            os("CCHAR");
            break;
         case CSCHAR:
            os("CSCHAR");
            break;
         case CINT:
            os("CINT");
            break;
         case CUNSG:
            os("CUNSG");
            break;
         case CLONG:
            os("CLONG");
            break;
         case CFLOAT:
            os("CFLOAT");
            break;
         case CDOUBLE:
            os("CDOUBLE");
            break;
         case CLABEL:
            os("CLABEL");
            break;
         case CSTRUCT:
            os("CSTRUCT ");
            ohw(bptr->word);
            bptr += 2;
            break;
         case FUNCTION:
            os("FUNCTION");
            break;
         case ARRAY:
            os("ARRAY ");
            onum(bptr->word);
            bptr += 2;
            break;
         case PTRTO:
            os("PTRTO");
            break;
         default:
            os("BITS ");
            onum(tp & 15);
            oc(' ');
            onum((tp >> 4)+1);
            break;
         }
      }
   while (tp >= FUNCTION);
   }
#endif

oc(ch)  /* output a char */
   char ch;
   {
   if (see_exit)
      {
      if (see_index < 78 && ch != '\n')
         see_msg[see_index++] = ch;
      }
   else
      putchar(ch);
   }

os(str)  /* output a string */
   char *str;
   {
   while (*str)
      oc(*str++);
   }

/* warning: recursive routine */
onum(num)  /* output a number (base 10) */
   int num;
   {
   if (num < 0)  /* negative? */
      {
      oc('-');  /* leading minus sign */
      num = -num;
      }
   if (num >= 10)
      onum(num / 10);
   oc(num % 10 +'0');
   }

obnum(bnum)  /* output a "byte" number */
   char bnum;
   {
   int num;
   onum(num = bnum);  /* force integer promotion, high byte will be 0 anyway */
   }

ohb(bnum)  /* output a hex byte */
   int bnum;
   {
   char temp;

   temp = bnum;
   temp >>= 4;
   oc(temp >= 10 ? temp+'7' : temp+'0');
   bnum &= 15;
   oc(bnum >= 10 ? bnum+'7' : bnum+'0');
   }

ohw(num)  /* output a hex WORD */
   int num;
   {
   ohb(num >> 8);
   ohb(num);
   }

ocrlf()  /* output LF */
   {
   oc(LF);
   }

ctlb(byt)  /* put a byte in CTEMP1 (control byte) */
   char byt;
   {
   if (!soopt)  /* !syntax check only */
      {
      if (inctl == &ctlbuf[512])  /* buffer full? */
         {
         if (write(control,ctlbuf,512) == -1)  /* write buffer */
            {
            os("cannot write temporary");
            real_exit(2);
            }
         inctl = &ctlbuf;  /* reset buffer */
         }
      *inctl++ = byt;  /* buffer byte for efficiency */
      }
   }

ctlw(wrd)  /* output a WORD to CTEMP1 (control word) */
   int wrd;
   {
   ctlb(wrd);  /* just send the two bytes */
   ctlb(wrd >> 8);
   }

ctls(str)  /* output a string to CTEMP1 (control string) */
   char *str;
   {
   do
      ctlb(*str);
   while (*str++);
   }

treew(wrd)  /* write a WORD to CTEMP2 */
   int wrd;
   {
   if (!soopt)  /* !syntax check only */
      {
      if (intree == &treebuf[512])  /* if buffer is full */
         {
         if (write(tree,treebuf,512) == -1)  /* write buffer */
            {
            os("cannot write temporary");
            real_exit(2);
            }
         intree = treebuf;  /* reset buffer pointer */
         }
      *intree++ = wrd;  /* buffer word for efficiency */
      ntree++;  /* update node count */
      }
   }

/* returns a node number */
tree1(branch)
   int *branch;
   {
   int at;

   at = ntree;  /* remember node number of this word */
   treew(*branch);  /* write single node */
   return at;
   }

/* returns a node number */
tree2(branch)
   int branch[];
   {
   int  at;

   at = ntree;  /* remember node number of this subtree */
   treew(branch[0]);  /* write 2 nodes */
   treew(branch[1]);
   return at;
   }

/* returns a node number */
tree3(branch)
   int branch[];
   {
   int  at;

   at = ntree;  /* remember n ode number of this subtree */
   treew(branch[0]);  /* write 3 nodes */
   treew(branch[1]);
   treew(branch[2]);
   return at;
   }

/* returns a node number */
tree4(branch)
   int branch[];
   {
   int at;

   at = ntree;  /* remember node number of this subtree */
   treew(branch[0]);  /* write 4 nodes */
   treew(branch[1]);
   treew(branch[2]);
   treew(branch[3]);
   return at;
   }

/* returns a node number */
tree5(branch)
   int branch[];
   {
   int at;

   at = ntree;  /* remember node number of this subtree */
   treew(branch[0]);  /* write 5 nodes */
   treew(branch[1]);
   treew(branch[2]);
   treew(branch[3]);
   treew(branch[4]);
   return at;
   }

/* returns a node number */
treev(num,branch)
   int branch[], num;
   {
   int at;
   char i;

   at = ntree;  /* remember node number of this subtree */
   for (i = 0; i < num; i++)  /* write nodes */
      treew(branch[i]);
   return at;
   }

tree3z()
   {
   treew(0);  /* write 3 NULL entries *
   treew(0);  /* Apparently, we do not need node number of this */
   treew(0);
   }

/* debug stuff to the object file */

#if CHECK

objb(byt)
   char byt;
   {
   if (!soopt)  /* !syntax check only */
      {
      if (inobj == &objbuf[512])  /* buffer full? */
         {
         if (write(objnum,objbuf,512) == -1)  /* write buffer */
            {
            os("cannot write temporary");
            real_exit(2);
            }
         inobj = objbuf;  /* reset buffer ptr */
         }
      *inobj++ = byt;  /* buffer byte for efficiency */
      }
   }

objw(wrd)
   int wrd;
   {
   objb(wrd);  /* write bytes of WORD */
   objb(wrd >> 8);
   }

objs(str)
   char *str;
   {
   do
      objb(toupper((*str) & 0x7f));
   while (*str++);  /* write string to object file */
   }


static char last_seg = ODSEG;

objtype(otype,num)
   char otype;
   int num;
   {
   int len;
   char *bp, shorter, ttyp, *sptr, *bptr;

   /* must say if in data or code for statics */
   if (addat->ntype[0] == FUNCTION)
      {
      if (last_seg != OCSEG)
         objb(last_seg = OCSEG);
      }
   else
      if (last_seg != ODSEG)
         objb(last_seg = ODSEG);
   objb(otype);  /* write type byte */
   bp = addat - addat->nlen;  /* get ptr to symbol name */
   if (otype != OPTYPE)  /* if not public */
      objs(bp);  /* write string */
   else
      {
      if (addat->nstor == SSTATIC && funname)  /* static? */
         {
         bptr = funname;  /* prefix symbol name with function name */
         while (*bptr)
            objb(*bptr++);
         objb('_');  /* and an underscore */
         }
      while (*bp)  /* add symbol name, too */
         objb(toupper(*bp++));
      objb('_');  /* add trailing underscore */
      objb(0);  /* EOS */
      }

   objw(num);  /* symbol table entry number */
/* get the length of the type info and output first */

   len = 1;
   bp = addat->ntype;
   shorter = 0;
   while (*bp >= FUNCTION)
      {
      if (len >= 9)
         {
         shorter = 1;
         break;
         }
      if (*bp == ARRAY)
         {
         bp += 2;
         len += 2;
         }
      bp++;
      len++;
      }
   if (*bp == CSTRUCT)
      {
      len += 2;
      }
   objb(len+shorter);
   bp = addat->ntype;
   while (len--)
      {
      objb(ttyp = *bp++);
      if (ttyp == ARRAY)
         {
         objb(*bp++);
         objb(*bp++);
         len -= 2;
         }
      if (ttyp == CSTRUCT)
         {
         sptr = bp->word;
         bp += 2;
         objw(sptr->staglen);
         len -= 2;
         }
      }
   if (shorter)
      objb(CINT);
   }

#endif
