/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  DeSmet C is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; either version 2 of the License, or any
 *  later version.
 *
 *  DeSmet C is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/* C88 COMPILER         C88.C */
/*
 *  Modification history:
 *    William Couture     8/26/05   Reformat/cleanup/comments
 *                                  update ltype to prevent addressing random
 *                                  memory
 *                                  changed 'heir' magic numbers to names
 */
/*
 *  BUG: The following could should be legal, but isn't:
 *          int i; a(int i) { i = 1; }
 *       The 'i' used in the function should be on the stack.  Instead, a
 *       syntax error is generated.   However, the following are correctly
 *       handled:
 *          int i; a(i) int i; { i = 1; }
 *       and
 *          int i; a(int j) { j = 1; }
 *       For some reason, in
 *          int i; a(int i){}
 *       the "int i" in a() is parsed via global(), because of the "addproto"
 *       variable.  The location of the local symbol is less than the
 *       "findover" value, due to the parsing order.  I have to work on that
 *       sometime.
 *       NOTE:  in the control file (CTEMP1), the line
 *                 a(i) int i; { }
 *              generates a function structure with pbytes == 2 (2 bytes
 *              are passed on the stack).  However, the line
 *                 a(int i) { }
 *              generates a function structure with pbytes == 0 (no bytes
 *              passed on the stack), although the generated code is
 *              correct.
 *
 *              This may be related to the above bug.
 *
 *       NOTE (12/22/2005): I think I have found the fix to this bug, but
 *              I have not inserted the code as I have not fully tested it.
 *              See the comments marked BUG12222005
 *
 *
 *  BUG: The following "operator" is legal, but should not be:
 *          ||=
 *       Due to a parsing error, this is taken as '|='
 *
 *  BUG: The following 1-line program will crash C88.EXE:
 *          a() ( }
 *       Note the typo:  '(' instead of '{'.
 *
 *  BUG: The following 1-line program will crash C88.EXE:
 *          a(..) ( }
 *
 *  BUG: Elipses in parameter lists don't seem to work.  i.e.
 *          a(int i, ...) {}
 *       should be legal, but instead generates an error.
 *
 *  BUG: adjecent string initalization doesn't work quite right.
 *          char a[4] = "ab""cd";
 *       give 'ab',0,'cd',0,0 in memory.  However,
 *          char a[] = "ab""cd";
 *       correctly gives 'abcd',0 in memory.  This is in initsome()
 *
 *  BUG: The following 1-line program will crash C88.EXE:
 *          #ifdef A
 *       This has been fixed, see BUG12292005 in C2.C
 *
 *  BUG: While skipping statements in an "#if<xxx>" or "#else" or "#elif",
 *       the statement
 *            #(space><cr><lf>
 *       is not properly parsed, and cause C88 to miss the "#endif".  For
 *       example:
 *          #ifdef A
 *          #<space>
 *          #endif
 *          int i;
 *       never "sees" the "#endif", and will actually crash C88 (see
 *       BUG12292005).
 *
 *  BUG: The preprocessor doesn't quite handle some tokenizing right.
 *       For example, the following does not work:
 *          #define A(b) printf(#b)
 *          main() { A(Hello); }
 *       Although adding a space does work:
 *          #define A(b) printf( #b)
 *          main() { A(Hello); }
 */   

#include "PASS1.H"
#include "NODES.H"
#if CHECK
   #include "OBJ.H"
#endif

char nextpgm[70]="Z:GEN ",
     mmonths[] = "JanFebMarAprMayJunJulAugSepOctNovDec";

char ltype[256] =
/* increased from 128 to 256 so "extended ASCII" in source file will be a
 * valid index into this array.  WAC 11/2/05
 */
   {
    34,    34,    34,    34,    34,    34,    34,    34,  /* NUL ^A ^B ^C ^D ^E ^F ^G */
    34, SPACE,    21,    34, SPACE, SPACE,    34,    34,  /* ^H ^I ^J ^K ^L ^M ^N ^O */  
    34,    34,    34,    34,    34,    34,    34,    34,  /* ^P ^Q ^R ^S ^T ^U ^V ^W */
    34,    34,    20,    34,    34,    34,    34,    34,  /* ^X ^Y ^Z ESC FS GS RS US */
 SPACE,     4,     5,    34,    34,     6,     7,     8,  /* SPACE ! " # $ % & ' */
    34,    11,     9,    10,    11,    12,    13,    14,  /* ( ) * + , - . / */
 DIGIT, DIGIT, DIGIT, DIGIT, DIGIT, DIGIT, DIGIT, DIGIT,  /* 0 1 2 3 4 5 6 7 */
 DIGIT, DIGIT,    34,    34,    15,    16,    17,    34,  /* 8 9 : ; < = > ? */
    34,LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,  /* @ A B C D E F G */
LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,  /* H I J K L M N O */
LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,  /* P Q R S T U V W */
LETTER,LETTER,LETTER,    34,    34,    34,    18,LETTER,  /* X Y Z [ \ ] ^ _ */
    34,LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,  /* ` a b c d e f g */
LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,  /* h i j k l m n o */
LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,LETTER,  /* p q r s t u v w */
LETTER,LETTER,LETTER,    34,    19,    34,    34,    34,  /* x y z { | } ~ DEL */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 128  ...  135 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 136  ...  143 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 144  ...  151 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 152  ...  159 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 160  ...  167 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 168  ...  175 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 176  ...  183 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 184  ...  191 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 192  ...  199 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 200  ...  207 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 208  ...  215 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 216  ...  223 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 224  ...  231 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 232  ...  239 */
    34,    34,    34,    34,    34,    34,    34,    34,  /* 240  ...  247 */
    34,    34,    34,    34,    34,    34,    34,    34   /* 248  ...  255 */
};

static char pNest;

main(argc,argv)
   int argc;
   char *argv[];
   {
   int  i;

#if LIMITED
   os("C88 Compiler    Special Version   V3.01    (c) Mark DeSmet, 1986\n");
#else
   i = 2;
   while (i < argc)
      if (*(argv[i++]+1) == '@')
         see_exit = 1;
#endif
   init(argc,argv);
   while (curch != CONTZ)  /* until EOF */
      {
      ntree = 1;
      global();
      if (ntree > 1)
         {
         ctlb(3);  /* flag flush of tree */
         ntree = 1;
         }
      }
   endit();
   }

static showVersion(void)
   {
   os("C88 Compiler V3.1h1  Copyright Mark DeSmet, 1982-1989\n");
   }

init(argc,argv)
   int  argc;
   char *argv[];
   {
   char *intname, option, cppopt = 0;
   int  i, nopt;

/* options are decoded here. unrecognised ones are passed through
   @    - special SEE option. Followed by hhhh:hhhh.
   A    - produce Assembler file and stop.
   B    - BIG model.
   C    - add checkout information to object.
   Dd   - d is drive to use for overlays.
   E    - write pre=processor output to stdout
   Id   - drive for #include files.
   M    - create .obj instead of .o format output.
   Nname=value - like a define.
   Oname - object name.
   Px   - Set pragma switch x.
   S    - Syntax check only
   Tn   - drive to use for temporaries.
   VE   - print every token in pass 1. print debug information in pass 2.
   VS   - print symbols in pass 1. print the tree in pass 2.
   W    - want warnings on structure passing and return.
   X    - do not invoke next pass and do not delete temporaries.
*/

/* add program name without extension to next invocation. */

   if (argc < 2)
      {
      os("no input file");
      real_exit(2);
      }

   if (argv[1][0] == '-' && toupper(argv[1][1]) == 'V')
      {
      showVersion();
      real_exit(0);
      }

   cur = argv[1];
   nopt = 6;
   i = 0;
   while (*cur && *cur != '.')
      {
      objname[i++] = *cur;
      nextpgm[nopt++] = toupper(*cur++);
      }
/* install the reserved words */

   mfree = _memory();  /* do our own "heap" managment */
   maxmem = _showsp() - 1224;
      /* Note that addres() keywords are in same order as #define Rxxxx constants */
   addres(0,"auto.break.case.char.const.continue.default.do.double.else.");
   addres(10,"enum.extern.float.for.goto.if.int.long.noalias.register.return.short.");
   addres(22,"signed.sizeof.static.struct.switch.typedef.union.unsigned.");
   addres(30,"void.");
   *nameat = RCHAR;  /* make VOID same as CHAR */
   addres(31,"volatile.while.__LINE__.__FILE__.__DATE__.__TIME__.__STDC__.");
   addres(38,"asm.cdecl.far.fortran.huge.interrupt.near.pascal.");

   add_define(cur = "__DESMET__");

   xsetblock();  /* set DOS allocated memory block to program size */
                 /* this allows us to use xalloc() to "push" / "pop"
                  * parse context for #include files.
                  */

   macxp = macExp;

   times(mtime);  /* "hh-mm-ss" */
   mtime[2] =
   mtime[5] = ':';  /* to hh:mm:ss */
   dates(mdate);  /* mm/dd/yy */
                  /* 01234567 */
              /* is transformed into */
                  /* mmm dd yyyy */
                  /*           1 */
                  /* 01234567890 */
   mdate[10] = mdate[7];  /* make space for 4-digit year */
   mdate[9] = mdate[6];
   mdate[8] = '9';        /* add Y2K bug! */
   mdate[7] = '1';
   mdate[5] = mdate[4];
   mdate[4] = mdate[3] == '0' ? ' ' : '1';
   i = mdate[1] - '0';
   if (mdate[0] != ' ')
      i += (mdate[0] - '0') * 10;
   strncpy(mdate, &mmonths[(i-1)*3], 3);
   mdate[6] =
   mdate[3] = ' ';

#if LIMITED == 0
   for (nopt = 2; nopt < argc; nopt++)
      {
      strcat(nextpgm," ");  /* copy command line to next program */
      cur = argv[nopt];
      if (*cur == '-')
         cur++;  /* allow leading - on options */
      option = toupper(*cur);
      if (option != 'I' && option != 'N' && option != 'P')  /* do not pass these parameters */
         strcat(nextpgm,cur);
      do
         {
         cur++;
         switch (option)
            {
            case '@':
               see_addr(cur);
               *cur = 0;
               break;
            case 'A':
               aopt = 1;  /* compile to assembly */
               break;
            case 'B':
               is_big = 1;  /* large memory model */
               /* add LARGE_CASE to symbol table */
               add_define(cur = "LARGE_CASE");
               add_define(cur = "__LARGE__");
               break;
            case 'C':
               copt = 1;  /* add checkout info to object file */
               break;
            case 'D':
               nextpgm[0] = toupper(*cur++);
               break;
            case 'E':
               cppopt = 1;  /* write preprocessor output to stdout */
               break;
            case 'I':
               strcpy(istring,cur);  /* WARNING:  istring is only 40 chars long! */
                                     /* Change this to byte copy later */
               if (istring[1] == 0)
                  {
                  istring[1] = ':';
                  istring[2] = 0;
                  }
               *cur = 0;
               break;
            case 'N':
               add_define();
               goto next_arg;
               break;
            case 'O':
               strcpy(objname,cur);
               *cur = 0;
               break;
            case 'P':
               while (*cur)
                  switch(toupper(*cur++))
                     {
                     case 'T':
                        trigraphs++;  /* process trigraphs */
                        break;
                     case 'W':
                        wopt = 1;  /* show warnings */
                        break;
                     case 'X':
                        xkwd = 1;  /* recognize extended keywords */
                        __stdc = 0;  /* !(only standard C) */
                        break;
                     }
               break;
            case 'S':
               soopt = 1;  /* syntax check only */
               break;
            case 'T':
               topt = toupper(*cur++);  /* temporaries drive */
               break;
            case 'V':
               while (*cur)
                  switch(toupper(*cur++))
                     {
                     case 'E':
                        eopt = 1;  /* tokens in pass 1, debug info in pass 2 */
                        break;
                     case 'S':
                        sopt = 1;  /* symbols in pass 1, tree in pass 2 */
                        break;
                     }
               break;
            case 'X':
               xopt = 1;  /* do not chain, do not delete temporaries */
            }
         } while (*cur);
next_arg:
      ;
      }
#endif

   if (!is_big)  /* !(large memory model) */
      add_define(cur = "__SMALL__");
   i = 0;
   while (objname[i] && objname[i] != '.')
      i++;
   if (objname[i] == 0)
      strcat(objname,".O");
   c_type[0] = CCHAR;
   int_type[0] = CINT;
   u_type[0] = CUNSG;
   l_type[0] = CLONG;
   ff_type[0] = CFLOAT;
   fd_type[0] = CDOUBLE;
   p_type[0] = PTRTO;
   p_type[1] = CCHAR;

/* open files */

   if (*argv[1] == '-')
      {
      if (!isatty(0))
         {
         strcpy(name, "stdin");
         file = 0;
         }
      else
         {
         error("stdin not a device");
         real_exit(2);
         }
      }
   else
      {
      strcpy(name,argv[1]);
      for (i = 0; i < 63; i++)
         {
         if (name[i] == 0)
            {
            strcpy(&name[i],".c");
            break;
            }
         if (name[i] == '.' && name[i+1] != '.')
            break;
         }
      if ((file = open(name,0)) == -1)
         {
         os("cannot open ");
         os(argv[1]);
         real_exit(2);
         }
      }  /* if (*argv[1] == '-') ... else */

   if (cppopt)  /* write preprocessor output to stdout */
      {
      char fname[65], *ebuf, *cbuf, newfile;
      unsigned lineno;

      if (see_exit == 0)  /* if within C, this option does not make sense */
         {
         fname[0] =   /* null initial name */
         lineno = 0;  /* 0 inital line number */
         ebuf =
         cbuf = mfree;  /* init buffers, start with ebuf == cbuf (empty buffer) */
         mfree += 512;
         minmem =
         hiwater = mfree;
         bufEnd = savEnd;
         cur =
         nestfrom[0] = bufEnd - 1;
         *cur = LF;  /* prime pump for dolf() */
         lineinc = 1;
         dolf(0);  /* skipit == 0, skip leading spaces & do preprocessor */
         tokit();  /* grab first token */
         while (curch != CONTZ)  /* until EOF */
            {
            newfile = 0;  /* not a new file */
            if (strcmp(fname,name))  /* unless a new name */
               {
               strcpy(fname,name);
               newfile = 1;
               }
            if (newfile || lineno != cline)  /* new file or new line, display info */
               {
               if (cbuf != ebuf)  /* any un-written info?  write it out */
                  {
                  *cbuf++ = LF;
                  *cbuf = 0;
                  os(ebuf);
                  cbuf = ebuf;
                  lineno++;
                  }
               if (newfile || cline - lineno > 1)  /* new info to display? */
                  {
                  char lbuf[10];

                  os("# line ");
                  itoa(cline, lbuf, 10);
                  os(lbuf);
                  os(" \"");
                  os(name);
                  os("\"\n");
                  lineno = cline;
                  }
               else
                  while (lineno < cline)
                     {
                     os("\n");
                     lineno++;
                     }
                }  /* if (newfile || lineno != cline) */
            if (*tokat == '$')  /* skip "do not macro expand" flag, we already used it */
               tokat++;
            while (tokat < cur)
               {
               char ch;

               if (ltype[(ch = *tokat)] == SPACE)  /* ^I, ^L, ^M, ' ' */
                  break;  /* space in token, we're done */
               if (ch == '"' || ch == '\'')  /* if quoted string/character */
                  {
                  *cbuf++ = *tokat++;  /* copy quote */
                  while (*tokat != ch || *(tokat-1) == '\\')
                     *cbuf++ = *tokat++;  /* copy entire quoted string/character */
                  }
               *cbuf++ = *tokat++;  /* copy token character */
               }
            *cbuf++ = ' ';  /* put a space between tokens */
            tokit();  /* next token */
            }  /* while (curch != CONTZ) */
         if (cbuf != ebuf)  /* if last line in the output buffer */
            {
            *cbuf++ = LF;  /* display output buffer */
            *cbuf = 0;
            os(ebuf);
            cbuf = ebuf;
            }
         }
      else
         error("E option not valid from see");
      real_exit(2);
      }  /* if (cppopt) */

   if (see_exit == 0)
      showVersion();

   if (!soopt)  /* !syntax check only */
      {

#if CHECK
      if (copt)  /* add checkout info to object file */
         {
         if ((objnum = creat(objname)) == -1)
            {
            os("cannot create ");
            os(objname);
            real_exit(2);
            }
         inobj = &objbuf[0];
         if (is_big)  /* large memory model */
            objb(OBIG);
         objb(ONAME);
         objs(name);
         }
#endif

      /* determine drive for temporaries. */

      intname = "A:CTEMP2";
      if (topt)  /* temporaries drive */
         *intname = topt;
      else
         intname += 2;
      if ((tree = creat(intname)) == -1)
         {
         os("cannot create CTEMP2");
         real_exit(2);
         }
      intree = &treebuf;

      intname = "A:CTEMP1";
      if (topt)  /* temporaries drive */
         *intname = topt;
      else
         intname += 2;
      if ((control = creat(intname)) == -1)
         {
         os("cannot create CTEMP1");
         real_exit(2);
         }
      inctl = &ctlbuf;
      }
   minmem =
   hiwater = mfree;
   bufEnd = savEnd;
   cur =
   nestfrom[0] = bufEnd - 1;
   *cur = LF;
   lineinc = 1;
   dolf(0);  /* skipit == 0, skip leading spaces & do preprocessor */
   tokit();  /* get first token for parser */
   }


addres(nres,res)  /* add reserved symbol to table */
   int  res;
   char nres;
   {
   cur = res;
   while (*cur)
      {
      tokit();
      newname();             /* setup symbol table entry */
      *nameat++ = RESERVED;  /* change class to RESERVED */
      *nameat = ++nres;      /* reserved value */
      mfree = nameat + 1;    /* finish memory allocation */
      cur++;  /* skip the '.' used to delimit keyword */
      }
   }

endit()
   {
   unsigned isize;
   int i;

   for (i = 0; i < 32; i++)  /* all hash table entries */
      {
      char *p;

      for (p = hash[i]; p; p = p->word)  /* for each hash chain */
         {
         char *q = p + 2;  /* get ptr to name in symbol table */

         while (*q++)  /* go past name + EOS */
            ;
         if (*q == STREF)  /* structure reference that was never resolved? */
            {
            os("undefined structure ");
            *(p+2) &= 0x7F;  /* back to standard ASCII */
            os(p+2);  /* show structure name */
            ocrlf();
            nerr++;
            }
         }
      }
   if (!soopt)  /* !syntax check only */
      {
#if CHECK
      if (sopt)  /* symbols in pass 1, tree in pass 2 */
         dumpsym();
      if (copt)  /* add checkout info to object file */
         {
         objb(ODSEG);  /* re-establish data default */
         if (inobj-objbuf && write(objnum,objbuf,inobj-objbuf) == -1)
            {
            os("cannot write ");
            os(objname);
            real_exit(2);
            }
         close(objnum);
         if (nerr)
            unlink(objname);
         }
#endif

      if (nerr == 0)  /* no errors, continue to GEN.EXE */
         {
         ctlb(0);  /* mark EOF in CTEMP1 */
         ctlw(nwarn);  /* add number of warnings */
         ctlw((hiwater-minmem)/((maxmem-minmem)/100));  /* usage percentage */
         if (write(control,ctlbuf,inctl-ctlbuf) == -1)  /* write remaining bytes */
            {
            os("cannot write temporary");
            real_exit(2);
            }
         close(control);
         isize = intree;
         if (write(tree,treebuf,isize-treebuf) == -1)  /* write any buffered tree nodes */
            {
            os("cannot write temporary");
            real_exit(2);
            }
         close(tree);

/* if x option is off, use _chain to trigger gen. */

         if (xopt == 0)  /* !(do not chain, do not delete temporaries) */
            {
            if (have_asm == 0 && aopt == 0)  /* aopt == compile to assembly */
               {
               strcat(nextpgm," Z");  /* tell GEN to compile directly to .O */
               }
            if (nextpgm[0] == 'Z')  /* next drive not known */
               _chain(&nextpgm[2]);  /* do not specify drive */
            else
               _chain(nextpgm);
            }
         real_exit(0);  /* otherwise just EXIT */
         }
      else  /* I don't think this needs to be an "else" clause... */
         {
         close(control);  /* errors, we're done */
         close(tree);
         }
      if (nwarn)
         {
         os("Number of Warnings = ");
         onum(nwarn);
         os("    ");
         }
      os("Number of Errors = ");
      onum(nerr);
      real_exit(2);
      }
   else
      real_exit(0);
   }


/* Parse the program.  This is the "root" of the BNF parsing */
global()
   {
   char *oldfree, *oldcur = 0, oldch, *fundef;
   int  funord,oldnerr,funline;
   char gotlname;

   if (curch == ';')  /* null statement */
      {
      tokit();
      return;
      }
   specs(SEXTERN);  /* get data type, default is extern */

   if (addstor > SEXTERN && addstor != STYPEDEF)
      {
      error("only STATIC and EXTERN allowed at this level");
      return;
      }
   if (ifch(';'))  /* NOTE: this allows statements like 'int ;' to be legal */
      return;
   do
      {
      if (addvar(0) == 0)  /* if invalid declaration */
         {
         error("illegal external declaration");
         newfun(mfree,0,0);  /* try to re-sync */
         return;
         }
      was_ext = 0;
      atype(addtype);
#if CHECK
      if (copt && addstor != STYPEDEF)  /* add checkout info to object file */
         objtype(OPTYPE,addat->noff);
#endif
      if (copt)  /* add checkout info to object file */
         funline = cline;
      else
         funline = 0;
      if (addat->ntype[0] == FUNCTION)  /* if this is a function */
         {
         fundef = addat;
         funtype = addat->ntype[1];  /* needed to force ret type */
         fun_ret_len = 0;  /* by default, not returning a structure */
         if (funtype == CSTRUCT)  /* if it is a structure */
            {
            struct_type = &addat->ntype[1];
            fun_ret_len = struct_type->ststagat->staglen;  /* returning a structure of size ... */
            pbytes = is_big ? 4 : sizeof(char *);  /* large memory model */
                 /* already have a parameter:  pointer to structure */
            }
         else
            pbytes = 0;  /* start with no parameters */
         if (is_big && funtype == PTRTO)  /* large memory model */
            ;
         else
            if (funtype > CDOUBLE)  /* CSCHAR, CSHORT, CULONG, CLABEL, CSTRUCT, CBITS */
               funtype = CUNSG;  /* just make it an unsigned 16-bit */
         oldfree = mfree;  /* remember where we started adding locals to
                            * symbol table
                            */
         funord = addat->noff;
         funname = addat - addat->nlen;
         gotlname = 0;
         if ((addparm || addproto) && nested > pNest)
               /* if parameters or prototypes parsed from a different level
                * of #define expansion, backtrack to previous level of
                * expansion to parse params/prototypes
                */
            {
            macxp = macfrom[pNest + 1];
            nested = pNest;
            }
         if (addparm)  /* if we need to parse function args */
            {
            gotlname = 1;
#if CHECK
            if (copt)  /* supply name for locals */  /* add checkout info to object file */
               {
               objb(OLNAME);
               objs(funname);
               }
#endif
            oldnerr = nerr;
            cur = addparm;
            lineinc = 1;
            tokit();
            doparms();  /* parse K&R function passed parameter list */
            if (oldnerr != nerr)  /* if error, try to re-sync with source */
               while (curch != '{' && curch != CONTZ)  /* } */
                  tokit();
            }  /* if (addparm) */
         if (addproto)
               /* addproto says that we have 'types' in the functions
                * (declaraton or prototype) parameters.  i.e.
                *    int q(int x);
                *    x(int y){}
                *
                * save function prototype in symbol table, and setup
                * protohash for later "quick reference"
                */
            {
            struct pargs *pp, **lp;

            oldcur = cur;  /* save current parsing info */
            oldch = curch;
            wp = mfree;
            *wp = protohash[addat & 31];
            protohash[addat & 31] = wp;
               /* addat is currently pointing to the symbol table entry
                * for this function, setup protohash to be a chain to
                * prototype(s) for this function.
                */
            *(wp + 1) = addat;
            *(lp = wp + 2) =
            pp = wp + 3 ;
            if (macproto)  /* if getting prototypes from a #define */
               {
               cur = macproto;
               while (cur != addproto)  /* make sure we are up to correct loc */
                  tokit();
               }
            else
               cur = addproto;
            tokit();
            if (heir == RESERVED && curch == 'v' && *cur == ')')
               goto noargs;  /* void parameter list */
            do
               {
               if (heir == RESERVED && bvalue == RELIPSE)
                  {
                  *lp = -1;
                  tokit();
                  notch(')');  /* make sure elipse is last argument */
                  goto eliparg;
                  }
               addloc = pp->ptype;  /* ptr to type bytes to fill */
               specs(SUNSTOR);
               getab(1);  /* Handle complex type modifiers, like pointers
                           * or pointers to functions
                           */
               if (addtype > 255)  /* array, structure, or union */
                  {
                  char typb;

                  do
                     {
                     *addloc++ =          /* copy type info */
                     typb = *addtype++;
                     if (typb == ARRAY)   /* and array size */
                        {
                        *addloc++ = *addtype++;
                        *addloc++ = *addtype++;
                        }
                     } while (typb >= FUNCTION);
                  if (typb == CSTRUCT)  /* if a struct */
                     {
                     addloc->word = addtype->word;  /* copy structure size */
                     addloc += 2;
                     }
                  }
               else
                  *addloc++ = addtype;  /* just copy simple type byte */
               lp = pp;
               *lp =
               pp = addloc;
               } while (ifch(','));
            notch(')');
noargs:
            *lp = 0;
eliparg:
            oldfree =
            mfree = pp;
            cur = oldcur;
            curch = oldch;
            }  /* if (addproto) */
         if (ifch('{')) /* } */  /* if function declaration (not data or fn prototype) */
            {
            if (addproto)  /* if we need to declare parameters */
               {
               int poff = 0;

               findover = mfree - 1;  /* allow local params with same name as globals */
               if (macproto)  /* if parsing param list from a macro expansion */
                  {
                  cur = macproto;
                  while (cur != addproto)
                     tokit();
                  }
               else
                  cur = addproto;
               macproto =
               addproto = 0;
               cline++;
               lineinc = 0;  /* do not update line number while parsing parameter list */
               tokit();
               if (heir == RESERVED && curch == 'v' && *cur == ')')  /* void list */
                  {
                  tokit();
                  goto voidspec;
                  }
               do
                  {
                  if (specs(SPARM) == 0 || addvar(0) == 0)
                     {
                     error("bad parameter declare");
                     findover = 0;
                     if (addat > mfree)
                        mfree = addat;
                     break;
                     }
                  addat->noff = poff + 4 + (is_big + is_big); /* 4 or 6 bytes from bp */
                                           /* large memory model */
                  atype(addtype);
                  poff += 2;
                  if (addat->ntype[0] == CLONG)
                     poff += 2;

            /* could be a structure */
                  else
                     if (addat->ntype[0] == CSTRUCT)
                        {
                        struct_type = addat->ntype;
                        poff += struct_type->ststagat->staglen - 2;
                        if (wopt)  /* show warnings */
                           warning("structure argument");
                        }
                     else
                        if (is_big && addat->ntype[0] == PTRTO)  /* large memory model */
                           poff += 2;
/* a FLOAT is really a DOUBLE */
                        else
                           {
                           if (addat->ntype[0] == CFLOAT)
                              addat->ntype[0] = CDOUBLE;
                           if (addat->ntype[0] == CDOUBLE)
                              poff += 6;
                           }
                  } while (ifch(','));
voidspec:
               notch(')');
               notch('{');  /* } */
               }  /* if (addproto) */

#if CHECK
            if (copt && gotlname == 0)  /* add checkout info to object file */
               {  /* supply name for locals */
               objb(OLNAME);
               objs(funname);
               }
#endif
           /* funname is to separate statics */
            newfun(oldfree,funord,funline);  /* parse function */
            funname = 0;
            return;  /* end of global(), do not look for additional declarations
                      * of the same type after a function
                      */
            }  /* if (ifch()) */
         funname = 0;
         addproto = 0;
         }  /* if (addat->ntype[0] == FUNCTION) */
      else
         doinit();  /* must be data, init it (if necessary) */
      } while (ifch(','));
   notch(';');
   }

/* parse K&R function passed parameter list */
doparms()
   {
   char *next;
   int  *pchain,bigger,*fchain;

   pchain = 0;
   cur = addparm;  /* point to command parameters to parse */
   addparm = 0;  /* forget this pointer */
   findover = mfree - 1;  /* remember the "separation line" in symbol
                           * table for local versus global variables.
                           */
   tokit();  /* prime the parser "pump" */
   do
      {
      if (heir == OPERAND && nameat < findover)  /* if already a global */
         heir = UNDEF;   /* we can have a local variable of same name */
      if (heir != UNDEF)
         {
         error("duplicate argument");
         findover = 0;
         return;
         }
      newname();
      if (pchain)  /* if we already have at least 1 parameter */
         *pchain = nameat;  /* add new parmeter to end of list */
      else
         fchain = nameat;  /* set chain root */
      pchain = &nameat->nchain;  /* get ptr to chain */
      nameat->opcl = OPERAND;
      nameat->nstor = SAUTO;
      nameat->noff = pbytes + 4 + (is_big + is_big);  /* 4 or 6 bytes from bp */
                                 /* large memory model */
      pbytes += 2;               /* assume parms are two bytes each */
      nameat->ntype[0] = CINT;   /* assume parm is integer */
      for (i = 1; i <= 10; i++)  /* but leave room to grow */
         nameat->ntype[i] = 0xff;
            /* BUG: while this seems like a lot of space, it allows
             *      a declaration like
             *         a(b,c) char ****************b,c; {}
             *      to overflow the reserved space and trash the
             *      symbol table (table entry for c is overwritten).
             */
      mfree = &nameat->ntype[11];  /* finish symbol table allocation */
      tokit();
      } while (ifch(','));

   if (notch(')'))
      return;
   while (specs(SPARM))  /* are types given for parameters? */
      {
      do
         {
         if (curch == ';')
            break;  /* allow struct decls without parms */
               /* Note:  Hmm... this allows files like:
                *           a(i) struct xxx; {}
                *        or
                *           a(i) long; {}
                *        to be valid.  Is this correct?
                */
         if (addvar(0) == 0 || addat > mfree)  /* declaring a varible not in passed list */
            {
            error("bad parameter declare");
            findover = 0;
            if (addat > mfree)  /* keep memory chain intact */
               mfree = addat;
            return;
            }
         addparm = 0;  /* forget any "parms" */
         atype(addtype);

         /* if parm is not 2 bytes long, must adjust offsets */

         bigger = 0;
         if (addat->ntype[0] == CLONG)
            bigger = 2;

         /* could be a structure */
         else
            if (addat->ntype[0] == CSTRUCT)
               {
               struct_type = addat->ntype;
               bigger = struct_type->ststagat->staglen - 2;
               if (wopt)  /* show warnings */
                  warning("structure argument");
               }
            else
               if (is_big && addat->ntype[0] == PTRTO)  /* large memory model */
                  bigger = 2;

         /* a FLOAT is really a DOUBLE */
               else
                  {
                  if (addat->ntype[0] == CFLOAT)
                     addat->ntype[0] = CDOUBLE;
                  if (addat->ntype[0] == CDOUBLE)
                     bigger = 6;
                  }
         if (bigger)
            {
            pbytes += bigger;
            next = addat->nchain;  /* update offset of all following params */
            while (next)
               {
               next->noff += bigger;
               next = next->nchain;
               }
            }
         } while (ifch(','));
      if (notch(';'))
         break;
      }
   findover = 0;  /* don't need parameters "dividing line" anymore */
#if CHECK
   if (copt)  /* add checkout info to object file */
      {
      while (fchain)
         {
         addat = fchain;
         objtype(OLTYPE,addat->noff);
         fchain = ((void *)fchain)->nchain;
         }
      }
#endif
  }

/* parse a function declaration */
newfun(oldfree,funord,funline)
   char *oldfree;
   int  funord,funline;
   {
   int  pstart,rnode[3];
   unsigned retval;

#if LIMITED == 0
   if (sopt && nerr > 0)  /* symbols in pass 1, tree in pass 2 */
      dumpsym();
#endif
   locoff = 0;
   laststmt = 0;
   pstart = compound(oldfree,1);  /* parse { <statment; ...> } */
   if (laststmt != RET)  /* if no explicit RET, "force" one */
      {
      retval = 0;  /* default to NULL return */
      if (fun_ret_len)
         error("must return structure");
      else
         if (funtype == CFLOAT || funtype == CDOUBLE)
            {
            warning("must return float");
            rnode[1] = oconst(0);  /* emit a constant of 0 */
            rnode[0] = CAST + (funtype << 8);  /* cast it to float/double */
            retval = tree2(rnode);  /* get node of cast'ed 0 */
            }
      rnode[0] = RET;
      rnode[1] = retval;
      rnode[2] = tree2(rnode);  /* emit RET, remember node */
#if CHECK
      if (copt)  /* add checkout info to object file */
         {
         rnode[0] = STMT;
         rnode[1] = rnode[2];
         rnode[2] = retnum;
         rnode[2] = tree3(rnode);
         }
#endif
      rnode[0] = LST + 2;   /* two computational nodes */
      rnode[1] = pstart;  /* "root" of function expansion */
                          /* rnode[2] is our forced RET, from above */
      pstart = tree3(rnode);  /* new start node */
      }

   ctlb(2);  /* put FUNCTION block in CTEMP */
   ctlw(funord);  /* function name index in symbol table */
   ctlw(ntree-1);  /* number of nodes in tree */
   ctlw(pstart);   /* root node of tree */
   ctlw(pbytes);   /* bytes passed into function */
   if (locoff & 1)
      locoff--;  /* since locoff is a negative number, this will "add" a
                  * byte to make the stack allocation size even.
                  */
   ctlw(locoff);   /* local bytes allocated */
   ctlw(funline);  /* funline is only set if doing a checkout file */
   }



/* put any storage types into addstor and a pointer to a type in addtype */
/* returns TRUE (1) if a data type, else FALSE (0) */
specs(defstor)
   char defstor;
   {
   char tempstor, got, tempstru;
   int  tempaddat;

   addstor = defstor;  /* default storage class */
   addtype = CINT;     /* integer is always default */
   got = 0;            /* see if have anything */

   while (1)
      {
      if (heir == RESERVED)  /* if a reserved keyword */
         {
         got = 1;  /* default to storage */
         switch (bvalue)
            {
            case RCONST:    /* these keywords are "nops" */
            case RNOALIAS:
            case RVOLATILE:
            case RCDECL:
            case RFAR:
            case RFORTRAN:
            case RHUGE:
            case RINTERRUPT:
            case RNEAR:
            case RPASCAL:
            case RELIPSE:
               break;
            case RAUTO:   /* "nops" unless local variable */
            case RREGISTER:
               if (addstor != SPARM)
                  addstor = SAUTO;
               break;
            case RSTATIC:
               addstor = SSTATIC;
               break;
            case REXTERN:
               addstor = SEXTONLY;
               break;
            case RTYPEDEF:
               addstor = STYPEDEF;
               break;
            case RCHAR:
               if (addtype != CSCHAR)
                  addtype = CCHAR;
               break;
            case RSHORT:
            case RINT:
               if (addtype != CUNSG && addtype != CLONG)
                   /* NOTE: this allows "long short i;", size = 4 bytes */
                  addtype = CINT;
               break;
            case RSIGNED:
               addtype = CSCHAR;
               break;
            case RLONG:
               if (addtype == CUNSG)
                  error("illegal type");
               addtype = CLONG;
               break;
            case RUNSIGNED:
               if (addtype == CLONG)
                  error("illegal type");
               addtype = CUNSG;
               break;
            case RFLOAT:
               addtype = addtype == CLONG ? CDOUBLE : CFLOAT;
               break;
            case RDOUBLE:
               addtype = CDOUBLE;
               break;
            case RSTRUCT:
            case RUNION:
               tempstor = addstor;
               tempstru = in_stru;
               tempaddat = addat;
               stagat.stat = newstru();
               in_stru = tempstru;
               addstor = tempstor;
               addat = tempaddat;
               stagat.cstype = CSTRUCT;
               addtype = &stagat;
               return 1;
            case RENUM:
               newenum();
               return 1;
            default:   /* not a data type */
               return 0;
            }
         tokit();  /* parse next storage qualifier */
         }
      else
         if (heir == OPERAND && nameat->nstor == STYPEDEF)  /* reduce typedef'ed type to base storage */
            {
            got = 1;
            addtype = &nameat->ntype[0];
            tokit();
            }
         else
            return got;
      }
   }


/* add new variable or parameter or function / procedure */
/* returns TRUE (1) if name in symbol table, else FALSE (0) */
/* Warning: recursive routine */
addvar(found)
   char found;  /* has a symbol table entry with this name been found? */
   {
   int  val;

   if (!found)
      {
      if (curch == '*')  /* pointer to ... */
         {
         tokit();
         found = addvar(0);
         if (found)
            *addloc++ = PTRTO;  /* if name found, say pointer to type */
         return found;
         }
      if (ifch('('))  /* type cast, etc */
         {
         found = addvar(0);
         notch(')');
         return (addvar(found));
         }
      while (heir == RESERVED && bvalue >= RCDECL && bvalue <= RPASCAL)
         tokit();  /* ignore "nop" reserved words */
      if (heir == UNDEF)  /* if we haven't seen this name yet */
         newname();   /* new symbol table entry */
      if (heir == OPERAND)
         {
         if (addstor == SPARM)
            {
            if (nameat > findover)  /* passed parameter */
               heir = UNDEF;
            /*************************************************************
             *  BUG12222005
             *
             *  I think the above two lines should be replaced with
             *  if (nameat < findover)
             *     {
             *     heir = UNDEF;
             *     newname();
             *     }
             *
             *************************************************************/
            }
         else
            if (in_stru || nameat < findover || addstor == SEXTONLY ||
                     nameat->nstor == SEXTONLY || nameat->ntype[0] == FUNCTION)
               {
               if (addstor == SEXTERN || addstor == SEXTONLY ||
                                        (addstor == SSTATIC && funname == 0))
                 /* if we're here, we have a previous definition.  i.e.
                  *    extern int i;  int i;
                  * or
                  *    int i;  extern int i;
                  */
               original = nameat;
               newname();
               heir = UNDEF;
               }
         }
      if (heir == UNDEF)
         {
         addat = nameat;
         *nameat = OPERAND;
         addat->nstor = (addstor == SPARM) ? SAUTO : addstor;
         addloc = &nameat->ntype[0];
         found = 1;
         tokit();
         }
      }
   if (found)
      {
      while (curch == '(' || curch == '[')
         {
         if (curch == '(')  /* skip past parms if any */
                            /* declaring a function, i.e. int x(){} or doing
                             * local decl of a function, i.e.
                             *      y(){int x();}
                             */
            {
            pNest = nested;  /* remember nested level of #define expansion
                              * for later param/prototype parsing
                              */
            addparm = cur;  /* remember pointer to (possible) args */
            macproto =
            addproto = 0;
            tokit();
            if ((heir == OPERAND && nameat->nstor == STYPEDEF)
                        || (heir != UNDEF && heir != OPERAND))
               {
               addproto = addparm;
               addparm = 0;
               if (cur > savEnd && cur < macEnd)
                  macproto = lastname;
               }
            if (curch != ')')  /* if not blank parms */
               {
               int rplvl = 1;
               while (rplvl)
                  {
                  if (curch == '(')
                     rplvl++;
                  if (curch == ')')
                     rplvl--;
                  tokit();
                  }
               }
            else
               {
               addparm =   /* no parms, no need to parse them later */
               addproto = 0;
               tokit();
               }
            *addloc++ = FUNCTION;
            }
         else  /* this must be '[' */
            {
            tokit();
            if (heir == CONSTANT || heir == LCONSTANT || curch == '('
                    || (heir == RESERVED && bvalue == RSIZEOF))
               val = constexp();  /* size of array */
            else
               val = -1;
            if (addstor == SPARM && addloc == &addat->ntype[0])
               {
               *addloc++ = PTRTO;
               }
            else
               {
               *addloc++ = ARRAY;
               if (val != -1)
                  {
                  addloc->word = val;
                  }
               else
                  {
                  if (addstor <= SEXTERN || addstor == STYPEDEF)
                     addloc->word = -1;
                  else
                     {
                     error("sorry, must have dimension for locals");
                     addloc->word = 1;
                     }
                  }
               addloc += 2;
               }
            notch(']');
            }
         }
      }
   return found;
   }

atype(addtype)  /* add type byte(s) to end of operand types */
   char *addtype;
   {
   char i,typb;

   if (addtype > 255)  /* if a pointer to type, not a simple type */
      {
      if (addstor == SPARM && *addtype == ARRAY && addloc == &addat->ntype[0])
         {
         addtype += 3;  /* skip ARRAY and dimension */
         *addloc++ = PTRTO;
         }
      do
         {
         *addloc++ =
         typb = *addtype++;
         if (typb == ARRAY)
            {
            *addloc++ = *addtype++;  /* copy dimension */
            *addloc++ = *addtype++;
            }
         } while (typb >= FUNCTION);  /* while ARRAY or PTRTO */
      if (typb == CSTRUCT)
         {
         addloc->word = addtype->word;  /* get STAG */
         addloc += 2;
         }
      }
   else
      *addloc++ = addtype;

   if (mfree < addloc)  /* if new variable, not filling-in already allocated
                         * symbol like a parameter
                         */
      {
      if (addstor <= SEXTERN)  /* EXTERN or STATIC variable */
         {
         if (original)   /* have older definition */
            {
            addat->noff = original->noff;
            i = 0;   /* types must agree */
            while (1)
               {
               typb = original->ntype[i];
               if (typb != addat->ntype[i])
                  {
                  warning("conflicting types");
                  break;
                  }
               if (typb == ARRAY)
                  i += 2;
               else
                  if (typb != FUNCTION && typb != PTRTO)
                     break;
               i++;
               }
            if (addat->ntype[0] == FUNCTION)
               original = 0;
            else
               {
               if (addstor != SEXTONLY)
                  {
                  original = 0;
                  addat->nstor = SSTATIC;
                  was_ext = 1;
                  }
               }
            }
         else
            {
            addat->noff = ++ordinal;  /* add symbol table index number */

            if (addat->ntype[0] == FUNCTION)  /* if this is a function */
               {
               while (addloc <= &addat->ntype[10])
                  *addloc++ = 0xff;  /* add "nulls" for later expansion */
               addat->nchain = hash[32];  /* add to function hash */
               hash[32] = addat;
               ctlb(1);  /* control structure to CTEMP1:  EXTERN */
               ctlb(addstor == SEXTONLY ? SEXTERN : addstor);  /* visibility */
               ctlw(addat->noff);  /* symbol table index number */
               ctls(addat - addat->nlen);  /* send string */
               ctlb(INITFUN);  /* tell GEN.EXE this is a function */
               }
            }
         }
      mfree = addloc;  /* finish symbol table allocation */
      }
   }

newstru()
   {
   char is_stru,*struat,bitsin;
   int  sesize,*memchain;
   char *real_type;

   is_stru = bvalue == RSTRUCT;  /* STRUCT or UNION? */
   tokit();
   if (heir == OPERAND || heir == UNDEF)  /* a named structure type? */
      {
      string[0] |= 0x80;  /* not a normal symbol name */
      bptr = hash[hashno];
      while (bptr)  /* see if structure name already exists */
         {
         j = -1;
         wp = bptr;
         bptr += 2;
         while (*bptr++ == string[++j])
            ;
         if (j > i)
            {
            nameat = wp;
            nameat += i + 3;
            heir = *nameat;
            if (heir == STAG || heir == STREF)
               goto gotSTAG;
            }
         bptr = *wp;
         }
      heir = UNDEF;
      }
gotSTAG:
   in_stru = 1;  /* global flag: we are defining structure elements */
   if (heir == STAG || heir == STREF)  /* if already a structure definition */
      {
      memchain =
      struat = nameat;
      tokit();
      }
   else
      {
      if (heir != UNDEF)  /* create dummy stag */
         {
         mfree->word = anon;
         anon = mfree;
         (mfree+2)->byte = CSTRUCT;
         (mfree+3)->word = mfree + 5;
         memchain =
         struat =
         nameat = mfree + 5;
         }
      else
         {
         string[0] |= 0x80;  /* create entry for named structure */
         newname();
         *wp = hash[hashno];
         hash[hashno] = wp;
         memchain =
         struat = nameat;
         tokit();
         }
      nameat->stagcl = STREF; 
      mfree = &nameat->staglen;
      mfree += 2;
      }

   if (ifch('{')) /* } */  /* allow pointer to forward structure */
      {
      if (*struat != STREF && wopt)  /* show warnings */
         warning("structure redefinition");
      *struat = STAG;
      struat->schain =
      struat->staglen = 0;
      if (is_stru == 0)  /* if a UNION */
         struat->staglen = 2;
      bitsin = 0;
      do
         {
         if (specs(SMEMBER) == 0)  /* if not a recognizeable data type */
            {
            error("bad member declare");
            return struat;
            }
         if (addstor != SMEMBER)  /* if "extern" or "register" or some such */
            {
            error("bad member storage");
            return struat;
            }
         real_type = addtype;
         do
            {
            if (ifch(':'))  /* if a bitfield in the form 'int : 1;' */
               {
               if (heir != CONSTANT || wvalue > 16)  /* must be a constant 0 .. 16*/
                  {
                  error("field needs constant");
                  return struat;
                  }
               if (wvalue == 0 || wvalue+bitsin > 16)  /* need to start at next INT? */
                  {
                  if (is_stru && bitsin)
                     struat->staglen += 2;
                  bitsin = wvalue;
                  }
               else
                  bitsin += wvalue;
               tokit();
               continue;
               }
            if (addvar(0) == 0)  /* if name not in symbol table */
               {
                  /* allow stags in unions */
               if (!is_stru && addtype == &stagat)  /* union && contains structure */
                  {
                  sesize = stagat.stat->staglen;
                  if (sesize == 0)  /* structure not declared */
                     error("undefined structure");
                  goto got_size;
                  }
               error("bad STRUCT declare");
               return struat;
               }
            addparm = 0;
            ((void *)memchain)->nchain = addat;
            memchain = addat;
            if (ifch(':'))  /* if a bitfield in the form 'int i : 1;' */
               {
               if (heir != CONSTANT || wvalue > 16 || wvalue == 0)  /* must be a constant 1 .. 16*/
                  {
                  error("field needs constant");
                  return struat;
                  }
               if (wvalue+bitsin > 16)  /* if this will overflow the current WORD */
                  {
                  if (is_stru)  /* if a struct, move to next element */
                     struat->staglen += 2;
                  bitsin = 0;
                  }
               addtype = BITS + (wvalue-1) * 16 + bitsin;
               bitsin += wvalue;
               if (is_stru)
                  addat->noff = struat->staglen;
               else
                  addat->noff = 0;
               addloc = &addat->ntype[0];  /* forget real type */
               atype(addtype);
               tokit();
               }
            else
               {
               if (bitsin)  /* if previous field was a bitfield */
                  {
                  if (is_stru)  /* if struct, proceede to next element */
                     struat->staglen += 2;
                  bitsin = 0;  /* forget previous bitcount in word */
                  }
               atype(addtype);
#if CHECK
               if (copt)  /* add checkout info to object file */
                  {
                  objtype(OMTYPE,is_stru ? struat->staglen : 0);
                  }
#endif
               sesize = dsize(&addat->ntype[0]);
got_size:
               if (is_stru)
                  {
                  addat->noff = struat->staglen;  /* update structure size */
                  struat->staglen += sesize;
                  }
               else
                  {
                  if (sesize > struat->staglen)  /* update union size */
                     struat->staglen = sesize;
                  addat->noff = 0;
                  }
               }
            addtype = real_type;  /* restore overall type */
            } while (ifch(','));
         notch(';');
/* { */  } while (!ifch('}'));
      if (bitsin && is_stru)  /* if leftover bits, update structure size */
         struat->staglen += 2;
      }
   return struat;
   }


/* NEWENUM -- add enumerated types. treat as defines. */
newenum()
   {
   char *defat;
   int enumnum, num;

   enumnum = 0;
   tokit();
      /* ignore the enum name if present */
   if (heir == OPERAND || heir == UNDEF)
      {
      tokit();
      }

      /* if list of values, treat as #define x 1, #define y 2 etc. */
   if (ifch('{'))  /* } */
      {
      do
         {
         while (ifch(','))
            ;
/* { */ if (ifch('}'))
            return;
         if (heir != UNDEF)
            {
            error_l("duplicate enum");
            return;
            }
         newname();
         defat = nameat;
         tokit();
         if (ifch('='))
            enumnum = constexp();
         defat->defcl = DEFINED;
         defat->dargs = 255;
         defat = &defat->dval;
         *defat++ = DEFSTR;
            /* add number of the enum */
         num = enumnum++;
         if (num > 9999)
            *defat++ = num / 10000 + '0';
         if (num > 999)
            *defat++ = num / 1000 + '0';
         if (num > 99)
            *defat++ = num / 100 + '0';
         if (num > 9)
            *defat++ = (num / 10) % 10 + '0';
         *defat++ = num % 10 + '0';
         *defat++ = LF;
         *defat = DEFEND;
         mfree = defat + 1;
/* { */  } while (!ifch('}'));
      }
   }
