/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  DeSmet C is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; either version 2 of the License, or any
 *  later version.
 *
 *  DeSmet C is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/* GEN4.C part 4 of generation */
/*
 *  Modification history:
 *    William Couture    11/18/05   Reformat/cleanup/comments
 */

#include "PASS2.H"
#include "NODES.H"
#include "inst86.h"


/* recursive via genarith() -> get2() -> getval() -> genarith() */
genarith(node,vtype)
   int  node, vtype[];
   {
   int  lnode[7];
   char type;

   type = tree[node];
   get2(node,vtype,lnode);

/* take care of float first */

   if (floater(vtype))
      {
      switch (type)
         {
         case ADD:
            builtin(_FADD);
            break;
         case SUB:
            builtin(_FSUB);
            break;
         case MUL:
            builtin(_FMUL);
            break;
         case DIV:
            builtin(_FDIV);
         }
      return;
      }

/* take care of long next */
/* do not allow long + pointer */

   if (lnode[VT] == PTRTO && vtype[VT] == CLONG)
      forceint(vtype);
   if (vtype[VT] == PTRTO && lnode[VT] == CLONG)
      forceint(lnode);

   if (lnode[VT] == CLONG || vtype[VT] == CLONG)  /* if two longs */
      {
      forceacx(vtype,lnode);  /* force vtype into DX:AX and lnode into BX:CX */
      if (type == ADD)  /* handle 32-bit ADD */
         {
         asm_add(ADD86,toreg(AX),toreg2(CX));
         asm_add(ADC86,toreg(DX),toreg2(BX));
         }
      else
         if (type == SUB)  /* 32-bit SUB */
            {
            asm_add(SUB86,toreg(AX),toreg2(CX));
            asm_add(SBB86,toreg(DX),toreg2(BX));
            }
         else
            builtin(type-MUL+_MUL4);   /* call _MUL4 or _DIV4 */
      freev(lnode);/* if regs used by lnode, mark them as available */
      return;
      }

   if (type == ADD)
      genadd(vtype,lnode);
   else
      if (type == SUB)
         gensub(vtype,lnode);
      else
         muldiv(type,vtype,lnode);
   }


muldiv(type,vtype,lnode)
   char type;
   int  vtype[], lnode[];
   {
   char reg, byt;
      {  /* Huh?  WAC 11/21/2005 */
      int  mtype;
      char c_typ;

      if (type == MUL && lnode[VIS] == REGV) 
         if (lnode[VVAL] == AX)
            flip(lnode,vtype);  /* swap lnode, vtype references and contents */
      if (vtype[VIS] == CONSTV && lnode[VIS] == CONSTV)  /* constant { * / % } constant */
         {
         vtype[VT] = CINT;
         if (type != MUL && lnode[VVAL] == 0)
            error("divide by zero");
         switch (type)
            {
            case MUL:
               vtype[VVAL] *= lnode[VVAL];
               break;
            case DIV:
               vtype[VVAL] /= lnode[VVAL];
               break;
            case MOD:
               vtype[VVAL] %= lnode[VVAL];
            }
         return;
         }
      if (lnode[VIS] == CONSTV && type != MOD)  /* mul or div by constant */
         {
         if (lnode[VVAL] == 1)  /* mul or div by 1, that's a NOP */
            return;
         if (lnode[VVAL] == 2 || lnode[VVAL] == 4 || lnode[VVAL] == 8)
                   /* mul or div by power of two, and it's more efficient to
                    * do SHIFT's instead of real MUL or DIV
                    */
            {

/* V3.1 03/07/88 - in case we have been called from IND and need value in AX */

            prefer_si = 0;

            force(vtype,AXPAT);
            alterv(vtype);  /* optimizer call */
            if (type == MUL)
               mtype = SHL86;
            else
               if (vtype[VT] == CINT)  /* signed int */
                  {
                  mtype = SAR86;  /* signed shift */
                  need(DXPAT);
                  asm_cwd();   /* get 0 / -1 sign extension */
                  if (lnode[VVAL] != 2)  /* if we need to round */
                     asm_add(XOR86,toreg(AX),toreg2(DX));  /* if AX < 0, NOT AX */
                  asm_add(SUB86,toreg(AX),toreg2(DX));  /* add 0 / 1 to AX */
                     /* NOTE: for constant != 2, this gives abs(AX)
                      *       if constant == 2, AX > 0 we have AX,
                      *                         AX < 0 we have AX+1
                      *       and we are setup for proper rounding
                      */
                  }
               else
                  mtype = SHR86;  /* unsigned int, logical shift */
            asm_shift_1(mtype,vtype);
            if (lnode[VVAL] >= 4)  /* need more than 1-bit shift */
               {
               asm_shift_1(mtype,vtype);
               if (lnode[VVAL] == 8)  /* need 3-bit shift */
                  {
                  asm_shift_1(mtype,vtype);
                  }
               if (type != MUL && vtype[VT] == CINT)  /* div by signed constant */
                  {
                  if (lnode[VVAL] != 2)  /* if statement not needed, we're inside "if (lnode[VVAL] >= 4)" */
                     asm_add(XOR86,toreg(AX),toreg2(DX));
                  asm_add(SUB86,toreg(AX),toreg2(DX));
                      /* if AX < 0, get -AX */
                  }
               }
            return;
            }
         }
            /* LHS == Left Hand Side,  RHS == Right Hand Side */
      force(vtype,AXPAT);  /* force LHS into AX */
      alterv(vtype);  /* optimizer call */
      byt = vtype[VT] == CCHAR;  /* if LHS an unsigned char type */
      if (byt)
         {
         if (lnode[VT] != CCHAR)  /* if RHS is not an unsigned char */
            {
            if (lnode[VT] == CSCHAR)  /* if RHS signed char */
               asm_cbw();   /* makde LHS signed, sign extend into AH */
            else
               asm_move(toreglow(AH),tocon(0));  /* else set AH to 0 */
            vtype[VT] = CINT;  /* type is now an INT */
            byt = 0;  /* no longer a BYTE type */
            }
         else
            if (type != MUL)
               asm_move(toreglow(AH),tocon(0));  /* set AH to 0 */
         }
      if (!byt)
         {
         need(DXPAT);  /* need DX for high word of MUL / DIV / MOD */
         regat[DX] = 1;  /* mark DX in use, though '1' is not a valid vtype[] */
         if (type != MUL)  /* if DIV or MOD */
            if (((c_typ = vtype[VT]) == CINT || c_typ == CSCHAR)
                  && ((c_typ = lnode[VT]) == CINT || c_typ == CCHAR || c_typ == CSCHAR))
                           /* integer promotion rules */
               asm_cwd();  /* AX winds up SIGNED */
            else
               asm_add(XOR86,toreg(DX),toreg2(DX));  /* AX winds up UNSIGNED */
         forceint(lnode);  /* make lnode an INT */
         }
      forcemul(lnode);  /* force lnode[] into a reg */
      if (is_big)
         forcees(lnode);
      if (regat[DX] == 1)  /* our signature? */
         regat[DX] = 0;  /* yes, mark DX as free */
      if (type == MUL)  /* if a MUL */
         {
         mtype = MUL86;  /* generate a MUL instruction */
         if (byt)
            vtype[VT] = CINT;  /* result will be an INT */
         }
      else
         mtype = DIV86;  /* use DIV for both DIV and MOD */
      /* turn into IMUL and IDIV */
      if (((c_typ = vtype[VT]) == CINT || c_typ == CSCHAR)
            && ((c_typ = lnode[VT]) == CINT || c_typ == CSCHAR))
               /* integer promotion rules */
         mtype++;  /* MUL --> IMUL, DIV --> IDIV */
      asm_not(mtype,lnode);  /* generate MUL / DIV instruction */
      freev(lnode);  /* don't need this anymore */
      if (type == MOD)
         {
         if (byt)
            {
            asm_move(toreglow(AL),toreglow2(AH));  /* want remainder from AH */
            }
         else
            {
            vtype[VVAL] = DX;  /* set our result to DX, not AX */
            regat[DX] = regat[AX];
            regat[AX] = 0;
            }
         }
      }  /* Huh?  WAC 11/21/2005 */
   }

genadd(vtype,lnode)
   int  vtype[], lnode[];
   {
   int  used;

   if (lnode[VIS] == CONSTV)  /* if adding a constant */
      {
      if (lnode[VVAL] == 0)  /* adding 0, do nothing */
         return;
      if (vtype[VIS] == CONSTV)  /* adding constant to constant */
         {
         vtype[VVAL] = vtype[VVAL] + lnode[VVAL];  /* sum of constants */
         if (vtype[VT] == CCHAR || vtype[VT] == CSCHAR)  /* promote char to int */
            vtype[VT] = CINT;
         return;
         }
      if (vtype[VIS] == OFFV)  /* adding constant to offset */
         {
         vtype[VOFF] = vtype[VOFF] + lnode[VVAL];  /* just update the offset */
         return;
         }
      }
   if (lnode[VIS] == OFFV && vtype[VIS] != OFFV)
      flip(lnode,vtype);  /* swap lnode, vtype references and contents */
   if (vtype[VIS] == OFFV)
      {
      used = regused[vtype[VVAL]];
      if (used != 8)  /* if not direct address */
         {
         forceint(lnode);
         if (is_big)
            forcees(lnode);
         asm_add(ADD86,toreg(used),lnode);
         freev(lnode);
         }
      else
         {
         forceind(lnode);
         regat[lnode[VVAL]] = vtype;
         if (vtype[VVAL] == 8)
            vtype[VVAL] = (lnode[VVAL] == SI) ? 4 : 5;  /* was vtype[1] = ... */
         else
            if (vtype[VVAL] == 7)
               {
               vtype[VVAL] = (lnode[VVAL] == SI) ? 0 : 1;
               }
            else
               vtype[VVAL] = (lnode[VVAL] == SI) ? 2 : 3;  /* was vtype[1] = ... */
         }
      }
   else
      {
      if (lnode[VT] == PTRTO)
         flip(lnode,vtype);  /* swap lnode, vtype references and contents */
      if (vtype[VT] != lnode[VT])
         {
         if (vtype[VT] == CCHAR || vtype[VT] == CSCHAR)
            forceint(vtype);
         else
            if (lnode[VT] == CCHAR || lnode[VT] == CSCHAR)
               forceint(lnode);
         }
      forcerm(lnode);
      forcereg(vtype);
      alterv(vtype);  /* optimizer call */
      if (is_big)
         forcees(lnode);
      asm_add(ADD86,vtype,lnode);
      freev(lnode);
      }
   }

gensub(vtype,lnode)
   int  vtype[], lnode[];
   {
   if (lnode[VIS] == CONSTV)
      {
      if (lnode[VVAL] == 0)
         return;
      if (vtype[VIS] == CONSTV)
         {
         vtype[VVAL] = vtype[VVAL] - lnode[VVAL];
         if (vtype[VT] == CCHAR || vtype[VT] == CSCHAR)
            vtype[VT] = CINT;
         return;
         }
      if (vtype[VIS] == OFFV)
         {
         vtype[VOFF] = vtype[VOFF] - lnode[VVAL];
         return;
         }
      }
   if (lnode[VT] == PTRTO)
      flip(lnode,vtype);  /* swap lnode, vtype references and contents */
   if (lnode[VT] != vtype[VT])
      if (lnode[VT] == CCHAR || lnode[VT] == CSCHAR)
         forceint(lnode);
      else
         if (vtype[VT] == CCHAR || vtype[VT] == CSCHAR)
            forceint(vtype);
   forcerm(lnode);
   forcereg(vtype);
   alterv(vtype);  /* optimizer call */
   if (is_big)
      forcees(lnode);
   asm_add(SUB86,vtype,lnode);
   freev(lnode);
   }

/* recursive via genbin() -> get2same() -> get2() -> getval() -> genbin() */
genbin(node,vtype)
   int  node, vtype[];
   {
   int  lnode[7], type86;
   char type;

   type = tree[node];
   get2same(node,vtype,lnode);
   if (lnode[VIS] == CONSTV && vtype[VIS] == CONSTV && vtype[VT] != CLONG)
      {
      if (type == OR)
         vtype[VVAL] |= lnode[VVAL];
      else
         if (type == AND)
            vtype[VVAL] &= lnode[VVAL];
         else
            vtype[VVAL] ^= lnode[VVAL];
      if (vtype[VT] == CCHAR || vtype[VT] == CSCHAR)
         vtype[VT] = CINT;
      return;
      }
   forcerm(lnode);
   forcereg(vtype);
   if (is_big)
      forcees(lnode);
   alterv(vtype);  /* optimizer call */
   if (type == OR)
      type86 = OR86;
   else
      if (type == AND)
         type86 = AND86;
      else
         type86 = XOR86;
   asm_add(type86,vtype,lnode);
   freev(lnode);
   if (vtype[VT] == CLONG)
      {
      asm_add(type86,tormh(vtype),tormh2(lnode));
      }
   }


/* recursive via getval() */
gentest(node,vtype)  /* conditional statement ? : */
   int  node, vtype[];
   {
   int  lab, lab2;
   char reg, reglong, ntype, i;

   ntype = tree[node] >> 8;
   if (ntype == FUNCTION)
      ntype = is_big ? PTRTO : CUNSG;
   /* sadly, must get rid of all register variables. */
   for (i = 0; i < 8; i++)
      if (regat[i])
         pushv(regat[i],1);  /* 1 == need vtype to address variable on stack */

   gencond(tree[node+1],0,lab = ++nextlab,0);
   getval(tree[node+2],vtype);
   if (ntype == PTRTO)
      {
      if (vtype[VT] != PTRTO)
         forcel(vtype);
      force(vtype, INDPAT);
      force_real_es(vtype);
      reg = vtype[VVAL];
      }
   else
      if (ntype >= CFLOAT)
         loadf(vtype);
      else
         {
         if (vtype[VIS] == CONSTV && (vtype[VT] == CCHAR || vtype[VT] == CSCHAR))
            vtype[VT] = CINT;
         force(vtype,AXPAT);
         if (ntype == CLONG)
            forcel(vtype);
         else
            forceint(vtype);
         }
   freev(vtype);

   lab2 = ojmp(10,++nextlab);
   genlab(lab);
   getval(tree[node+3],vtype);
   if (ntype == PTRTO)
      {
      if (vtype[VT] != PTRTO)
         forcel(vtype);
      force(vtype, regpat[reg]);
      force_real_es(vtype);
      }
   else
      if (ntype >= CFLOAT)
         loadf(vtype);
      else
         {
         if (vtype[VIS] == CONSTV && (vtype[VT] == CCHAR || vtype[VT] == CSCHAR))
            vtype[VT] = CINT;
         force(vtype,AXPAT);
         if (ntype == CLONG)
            forcel(vtype);
         else
            forceint(vtype);
         }
   genlab(lab2);
   }

/* warning: recursive subroutine */
/* come here when generating code for IFE, WHIL, DOW, FORS, gentest(),
 * genlog()
 */
gencond(node,true,lab,keep)
   int  node, lab;
   int *keep;
   char true;
   {
   char type, otype, jtype;
   int  dolab, reg, vtype[7];
   unsigned cur_loc;

   type = tree[node];
   otype = optype[type];
   if (otype == 7)  /* logical -- LOR, LAND */
      {
      if (type+true == LAND)
                 /* LAND+(0) or LOR+(1) */
         {
         gencond(tree[node+1],true,lab,keep);  /* left node */
         gencond(tree[node+2],true,lab,keep);  /* right node */
            /* both must be true for entire expression to be true */
         }
      else
         {
         dolab = ++nextlab;  /* label for intermediate results */
         gencond(tree[node+1],!true,dolab,keep);  /* left node */
         gencond(tree[node+2],true,lab,keep);  /* right node */
            /* either can be true and entire expression is true */
         genlab(dolab);
         }
      }
   else
      if (otype == 8)  /* comparison -- EQ, NE, GE, LT, GT, LE */
         {
         reg = keep[VVAL];
            /* BUG: keep is often the NULL pointer */
         cur_loc = curoff;
         jtype = compare(node);  /* generate compare */
         if (true)  /* if we need to logically invert compare */
            jtype ^= 1;  /* do it (see to_short[] in gen5.c */
         cleanup(cur_loc);  /* remove "extra" locals */
         if (keep && reg != keep[VVAL])
                /* at least they check for keep == NULL before using reg */
            force(keep,regpat[reg]);
         ojmp(jtype,lab);
         }
      else
         if (otype == 13)  /* NULL */
            {
            if (true)
               {
               asm_jump(JMP86,lab);
               }
            }
         else
            if (type == NOT)  /* logical NOT of expression */
               {
               gencond(tree[node+1],!true,lab,keep);
               return;
               }
            else
               {
               reg = keep[VVAL];
                 /* BUG: keep is often the NULL pointer */
               cur_loc = curoff;
               getval(node,vtype);
               forcerm(vtype);
               if (keep && reg != keep[VVAL])
                /* at least they check for keep == NULL before using reg */
                  force(keep,regpat[reg]);
               if (vtype[VIS] == CONSTV && !floater(vtype))
                  {
                  if (vtype[VVAL] || vtype[VNAME])
                     true = !true;
                  if (true == 0)
                     {
                     cleanup(cur_loc);  /* remove "extra" locals */
                     asm_jump(JMP86,lab);
                     }
                  }
               else
                  {
                  if (floater(vtype))
                     {
                     loadf(vtype);
                     builtin(_FIS);
                     }
                  else
                     if (vtype[VT] == CLONG || vtype[VT] == PTRTO)
                        {
                        if (vtype[VIS] == VARV || vtype[VT] == PTRTO)
                           {
                           reg = need(ANYPAT);
                           if (is_big)
                              forcees(vtype);
                           asm_move(toreg(reg),tormh(vtype));
                           }
                        else
                           reg = vtype[VVAL] + 2;
                        asm_add(OR86,toreg(reg),vtype);
                        freev(vtype);
                        }
                     else
                        {
                        if (vtype[VIS] == CONSTV || vtype[VIS] == OFFV)
                           forcereg(vtype);
                        if (is_big)
                           forcees(vtype);
                        if (vtype[VIS] == REGV && vtype[VT] != CCHAR && vtype[VT] != CSCHAR)
                           asm_add(OR86,vtype,vtype);
                        else
                           asm_add(CMP86,vtype,tocon(0));
                        freev(vtype);
                        }
                  cleanup(cur_loc);  /* remove "extra" locals */
                  asm_jump( (true ? JNZ86 : JZ86),lab);
                  }
               }
   }

/* remove "extra" locals */
cleanup(cur_loc)
   unsigned cur_loc;
   {
   if (cur_loc < curoff)
      {
      asm_lea(SP,tobpm(cur_loc));
      curoff = cur_loc;
      }
   }

/* recursive via genlog() -> gencond() -> getval() -> genlog() */
genlog(node,vtype)
   int  node, vtype[];
   {
   int  lab, i;
   char reg;

   /* sadly, must get rid of all register variables. */
   for (i = 0; i < 8; i++)
      if (regat[i])
         pushv(regat[i],1);  /* 1 == need vtype to address variable on stack */

   needzreg(vtype);
   gencond(node,0,lab = ++nextlab,vtype);
   asm_inc(INC86,vtype);
   genlab(lab);
   }

flip(lnode,vtype)  /* swap lnode, vtype references and contents */
   int  *lnode, *vtype;
   {
   char i;
   int  temp;

   for (i = 0; i < 9; i++)  /* AX .. DI */
      {
      if (regat[i] == vtype)  /* if reg is used by vtype */
         regat[i] = lnode;  /* set usage to lnode */
      else
         if (regat[i] == lnode)  /* if reg is used by lnode */
            regat[i] = vtype;  /* set usage to vtype */
      }
   i = 0;
   while (i <= VFROM)  /* swap lnode, vtype contents */
      {
      temp = *vtype;
      *vtype++ = *lnode;
      *lnode++ = temp;
      i++;
      }
   }

/* return true if vtype addresses a floating point number */

floater(vtype)
   int  vtype[];
   {
   return (vtype[VT] == CFLOAT || vtype[VT] == CDOUBLE)
          && vtype[VIS] != OFFV;
   }
