/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  DeSmet C is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; either version 2 of the License, or any
 *  later version.
 *
 *  DeSmet C is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/* TYPES GO INTO THE DTYP FIELD OF NODES THAT NEED A TYPE */
/*
 *  Modification history:
 *    William Couture    10/12/05   Reformat/cleanup/comments
 */

  /* DTYP's */
#define CCHAR       1
#define CINT        2
#define CUNSG       3
#define CLONG       4
#define CFLOAT      5
#define CDOUBLE     6
#define CSCHAR      7
#define CSHORT      8
#define CULONG      9
#define CLABEL     10
#define CSTRUCT    11    /* FOLLOWED BY POINTER TO STAG IN SYMBOL TABLE.    */
#define BITS       12    /* dtype is BITS+(#bits-1)*16+bit offset */

/* THE FOLLOWING ARE THE TYPE MODIFIERS */
#define FPTR2      251
#define HPTR2      252
#define FUNCTION   253
#define ARRAY      254  /* FOLLOWED BY DIMENSION */
#define PTRTO      255

/* Types of initilization records (CTEMP1 contents) */

#define INITEND 10
#define INITDB  11  /* int number of bytes to zero */
#define INITRB  12  /* int number of bytes to RB */
#define INITVAL 13  /* type byte, node to keep, nodes needed, start node */
#define INITSTR 14  /* byte number of bytes that follow */
#define INITFUN 15  /* nothing follows - not even INITEND */
/* Define Node Types */

struct nod
   {
   char ntyp;  /* node type: OPND, CONST, LAB, CAST, TA, IND, IFE, ... */
   char dtyp;  /* data type: CCHAR, CINT, CUNSG, CLONG, CDOUBLE, ... */
   int  subn1;  /* LEFT node number */
   int  subn2;  /* RIGHT node number */
   char onext;
   char pat;
   };

#define OPND     0  /* operand.         DTYPE, BP offset or ordinal,
                                        structure offset, zero. */
  /* struct opnd_node { char ntyp; char dtyp; int varname; int offset; int in_union; }
   *   varname is 0 for local, else index of variable name (CTEMP1)
   *   offset is offset from BP for local, or offset from base address for
   *      array / structure / union
   *   un_union is a flag: 1 == in union / structure
   *                            (offset 0 from base addr) else 0
   */
#define CONST    1  /* constant.        DTYPE (1, 2 or 4), value. */
                                        /* value 2 words if CLONG */
  /* struct const_node { char ntyp; char dtyp; unsigned value; }
   * struct const_lnode { char ntyp; char dtyp; unsigned value_low; unsigned value_high; }
   * struct const_fnode { char ntyp; char dtyp; unsigned value_low; unsigned value_high; }
   *   const_dnode is used for dtype == 6 (CDOUBLE, float double, 64-bits)
   *   else
   *      const_lnode is used for dtype == 4 (CLONG, 4-byte constant)
   *   else const_node
   *   Note that dtyp will be 3 (CUNSG) for a pointer
   */
#define LAB      2  /* label.           node, ordinal of label. */
  /* struct lab_node { char ntyp; char dtyp; int stmtnode; int varname; }
   *   stmtnode is the node of the statement after the label, varname is
   *   the index of the variable name (CTEMP1)
   */
#define CAST     3  /* cast.            DTYPE, operand. */
  /* struct cast_node { char ntyp; char dtyp; int nodenum; }
   *   dtyp is target type, nodenum is node number of value to cast.
   */
#define TA       4  /* take address (&).operand. */
  /* struct ta_node { char ntyp; char dtyp; int nodenum; }
   *   nodenum is node number to take address of
   */
#define IND      5  /* indirect(*).     DTYPE, operand. */
  /* struct ind_node { char ntyp; char dtyp; int nodenum; }
   *   dtype is type of pointer reference
   */
#define IFE      6  /* if then else.    test, then stmt, else stmt or 0 */
  /* struct ife_node { char ntyp; char dtyp; int testnode; int truenode; int elsenode; }
   *   Note that 0 for node numbers represents a blank statement.  i.e.
   *       if () ; else ;
   *   becomes
   *       { 6, 0, 0, 0, 0 }
   */
#define WHIL     7  /* while.           test, stmt, 0, 0, 0, 0 */
  /* struct whil_node { char ntyp; char dtyp; int testnode; int stmtnode; int zeros[3]; }
   *   note that there are 3 0's at the end of the node, not the 4 shown
   *   in the original comment.  A 0 for node numbers indicates a "null"
   *   statment.  i.e. 
   *        while() ;
   *   becomes
   *        { 7, 0, 0, 0, 0, 0, 0 }
   */
#define DOW      8  /* do while.        stmt, test, 0, 0, 0, 0 */
  /* struct dow_node { char ntyp; char dtyp; int stmtnode; int testnode; int zeros[3]; }
   *   Note that there are 3 0's at the end of the node, not the 4 shown
   *   in the original comment.  A 0 for node numbers indicates a "null"
   *   statment.  i.e. 
   *        do ; while ();
   *   becomes
   *        { 8, 0, 0, 0, 0, 0, 0 }
   */
#define SWIT     9  /* switch.          test, stmt. */
  /* struct swit_node { char ntyp; char dtyp; int valuenode; int stmtnode; }
   *   A 0 for node numbers indicates a "null"
   *   statment.  i.e. 
   *        switch () ;
   *   becomes
   *        { 9, 0, 0, 0 }
   *   although this code does trigger a pass 2 error.
   */
#define FORS    10  /* for.             pre, test, post, stmt, 0, 0, 0, 0*/
  /* struct fors_node { char ntyp; char dtyp; int prenode; int testnode;
   *                    int postnode; int stmtnode; int zeros[3]; }
   *   prenode is the pre-loop setup statement, testnode is the termination
   *   condition node, postnode is the loop-end processing statement; stmtnode
   *   is the "in-loop" statement.  A 0 for any of the nodes indicates an
   *   empty statement  i.e.
   *           for ( ; ; ) ;
   *   becomes
   *           { 10, 0, 0, 0, 0, 0, 0, 0, 0 }
   *   Note that there are 3 0's at the end of the node, not the 4 shown in
   *   the original comment.
   */
#define RET     11  /* return.          value or zero. */
  /* struct ret_node { char ntyp; char dtyp; int valuenode; }
   *   valuenode == 0 indicates an empty statement.  i.e.
   *      return;
   *   becomes
   *      { 11, 0, 0 }
   */
#define GOTOS   12  /* goto.            label node. */
  /* struct gotos_node { char ntyp; char dtyp; int labelnode; }
   *   NOTE: labelnode doesn't seem to be what I expect.  I have to
   *         look at this more closely.
   */
#define CONT    13  /* continue.        */
  /* struct cont_node { char ntyp; char dtyp; }
   */
#define BRK     14  /* break.           */
  /* struct brk_node { char ntyp; char dtyp; }
   */
#define LST     (15-2) /* 13 */  /* nodes 15 to 19 are list nodes with 2 to 6 subn */
  /* 15: struct nodes2_node { char ntyp; char dtyp; int nodes2[2]; }
   * 16: struct nodes3_node { char ntyp; char dtyp; int nodes3[3]; }
   * 17: struct nodes4_node { char ntyp; char dtyp; int nodes4[4]; }
   * 18: struct nodes5_node { char ntyp; char dtyp; int nodes5[5]; }
   * 19: struct nodes6_node { char ntyp; char dtyp; int nodes6[6]; }
   */
#define CAS     20  /* case label.      node, value. */
  /* struct cas_node { char ntyp; char dtyp; int stmtnode; int value; }
   */
#define DFLT    21  /* default label    node. */
  /* struct dflt_node { char ntyp; char dtyp; int stmtnode; }
   */
#define CALL    22  /* call.            dtype,function ord, arguments or 0,
                                        return structure length or 0. */
  /* struct call_node { char ntyp; char dtyp; int fnnamenode; int argsnode; int strctlen; }
   *   if no args, argsnode is 0.  If the function does not return a structure,
   *   strctlen is 0 else it is the size of the returned structure.
   */
#define TEST    23  /* conditional(? :).dtype, test, true val, false val. */
  /* struct test_node { char ntyp; char dtyp; int testnode; int truenode; int falsenode; }
   */

  /* NOTE: the order of the nodes ADD .. XOR *MUST* be the same as the
   *       order of the nodes AADD .. AXOR.  This is used in opeq() in
   *       C6.C
   */
#define ADD     24  /* add.             left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT + RIGHT
   */
#define SUB     25  /* subtract.        left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT - RIGHT
   */
#define MUL     26  /* multiply.        left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT * RIGHT
   */
#define DIV     27  /* divide.          left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT / RIGHT
   */
#define MOD     28  /* modulo.          left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT % RIGHT
   */
#define SHR     29  /* shift right.     left node, right node. */
  /* struct match_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT >> RIGHT
   */
#define SHL     30  /* shift left.      left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT << RIGHT
   */
#define AND     31  /* binary and (&).  left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT & RIGHT
   */
#define OR      32  /* binary or (|).   left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT | RIGHT
   */
#define XOR     33  /* exclusive or.    left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT ^ RIGHT
   */
#define LOR     34  /* logical or (||). left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT || RIGHT
   */
#define LAND    35  /* logical and (&&).left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT && RIGHT
   * Note:  LAND *MUST* be the node after LOR (i.e. LAND == LOR+1)
   */
#define EQ      36  /* equality (==).   left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT == RIGHT
   */
#define NE      37  /* not equal (!=).  left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT != RIGHT
   */
#define GE      38  /* greater or equal.left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT >= RIGHT
   */
#define LT      39  /* less than.       left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT < RIGHT
   */
#define GT      40  /* greater than.    left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT > RIGHT
   */
#define LE      41  /* less or equal.   left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT <= RIGHT
   */
#define NEG     42  /* negate.          node. */
  /* struct unary_node { chr ntyp; char dtyp; int opnode; }
   */
#define NOT     43  /* not.             node. */
  /* struct unary_node { chr ntyp; char dtyp; int opnode; }
   */
#define COMP    44  /* compliment.      node. */
  /* struct unary_node { chr ntyp; char dtyp; int opnode; }
   */
#define PREI    45  /* pre increment(++).  node, add value. */
  /* struct prei_node { char ntyp; char dtyp; int varnode; int incvalue; }
   */
#define PRED    46  /* pre decrement(--). node, subtract value. */
  /* struct pred_node { char ntyp; char dtyp; int varnode; int decvalue; }
   */
    /* Original comment was wrong:
     *              post decrement(++). node, subtract value.
     */
#define POSTI   47  /* post increment(++).  node, add value. */
  /* struct posti_node { char ntyp; char dtyp; int varnode; int incvalue; }
   */
    /* Original comment was wrong:
     *              pre increment(--).  node, add value.
     */
#define POSTD   48  /* post decrement(--).  node, subtract value. */
  /* struct postd_node { char ntyp; char dtyp; int varnode; int decvalue; }
   */
    /* Original comment was wrong:
     *              pre decrement(--).  node, subtract value.
     */

/* THE FOLLOWING ARE THE ASSIGNMENT NODES */

#define ASGN    49  /* assignment.      left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT = RIGHT
   */

  /* NOTE: the order of the nodes ADD .. XOR *MUST* be the same as the
   *       order of the nodes AADD .. AXOR.  This is used in opeq() in
   *       C6.C
   */
#define AADD    50  /* add (+=).        left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT += RIGHT
   */
#define ASUB    51  /* subtract (-=).   left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT -= RIGHT
   */
#define AMUL    52  /* multiply (*=).   left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT *= RIGHT
   */
#define ADIV    53  /* divide (/=).     left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT /= RIGHT
   */
#define AMOD    54  /* modulo (%=).     left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT %= RIGHT
   */
#define ASHR    55  /* shift right (>>=). left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT >>= RIGHT
   */
#define ASHL    56  /* shift left (<<=).left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT <<= RIGHT
   */
#define AAND    57  /* binary and (&=). left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT &= RIGHT
   */
#define AOR     58  /* binary or (|=).  left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT |= RIGHT
   */
#define AXOR    59  /* exclusive or.    left node, right node. */
  /* struct math_node { char ntyp; char dtyp; int leftnode; int rightnode; }
   *    LEFT ^= RIGHT
   */

#define NULL    60  /* used as null type in generation */
  /* I don't know how to make C88 generate a NULL node, so I don't know
   * if there are any associated fields.
   */
#define OUTASM  61  /* assembler follows. ends with byte=255. */
  /* Notes:  if a line does not contain an even number of characters, it is
   *         padded with a binary 0 to make it an even number.  Also, although
   *         C88 emits ntyp == 255 && dtyp == 255 to signal the end of
   *         assembly, GEN takes a single 255 anywhere (such as a character
   *         constant or comment) to be end of assembly.
   */
#define STMT    62  /* line number follows */
  /* struct stmt_node { char ntyp; char dtyp; int linenode; int linenumber; }
   * linenode is the "root node" corresponding to the line number
   */
#define MOVE    63  /* move structure.   left,right,length. right=0 for push.*/
  /* struct move_node { char ntyp; char dtyp; int leftnode; int rightnode; int length; }
   *    LEFT = RIGHT  {* structure assignment *
   * NOTE:  I think the original comment is wrong.  The code
   *           struct { int i; } a, b;  main() { a = b; }
   *        produces a symbol table of:
   *             a -- index == 1
   *             b -- index == 2
   *        and a tree of:
   *             0800  0001  0000  0000  0304  0001  0800  0002                    
   * (tree indx) 1     2     3     4     5     6     7     8
   *             STUCT a                 addressof <struct a>
   *                                                 STRUCT b
   *             0000  0000  0304  0007  003F  000B  0005  0002                    
   * (tree indx) 9     A     B     C     D     E     F     10
   *                         addressof <struct b>
   *                                     MOVE  addrof(b)   sizeof(struct)
   *                                                 addrof(a)
   *             000B  0000  000F  000D  0011                                      
   */


#ifdef NEVERDEF

   The output of the C88 compiler is two files, CTEMP1 and CTEMP2 which
   are passed to GEN to finish the compilation process.

   CTEMP1 is the "control" file, which has the symbol information and
   tells GEN which objects to compile.

   CTEMP2 is the "nodes" file, which has the compuational information
   stored in a tree structure.

   Note that string constants are stored in CTEMP1 (i.e. char str[4]="abc";)
   while numeric constants have their names stored in CTEMP1 with pointers
   into CTEMP2 which contains the numeric value (i.e. int i = 1;)

   CTEMP1 is a list of control structures, each starting with a single
   byte identifying the structure type.  They are:
      0 -- EOF  followed by 2 ints:
               int nwarnings
               int usage (percentage)
      1 -- EXTERN (variable name) followed by:
              char visibility SEXTERN == 2, global visibility (int i;)
                   (nstor)    SSTATIC == 1, only visible in this module
                                                (static int i;)
                              SEXTONLY == 0, global visibility from another
                                             module (extern int i;)
              int  symbol table entry number (starts at 1)
              char*N symbol name, '\0' terminated.
           additional substructures may follow, starting with a single byte
           to identify the substructure:
              INITEND (== 10) end of variable initialization
              INITDB  (== 11) followed by
                 int number of bytes, 0 initialization
              INITRB  (== 12) followed by
                 int number of bytes, no initialization
              INITVAL (== 13) numeric initialization, followed by
                 char dtyp (CCHAR, CINT, CUNSG, CLONG, CFLOAT, CDOUBLE)
                 int  "node to keep" -- where to chop off tree to remove
                      one-time init value from function tree
                 int  nodes_needed (node in current tree in CTEMP2 to read up
                      to for constant expression)
                 int  start_node (node at which to start tree processing)
              INITSTR (== 14) string constant, followed by
                 char strlen  (bytes in string)
                 char*n string value (does *NOT* include normal C EOS '\0',
                        though it may include embedded '\0' padding.  i.e.
                        char a[3] = "ab" will have a string value of "ab",
                        since the 3rd byte would be the normal EOS '\0', but
                        char a[4] = "ab" will have a string value of "ab\0"
                        to pad the string to LEN-1, leaving room for the
                        normal EOS of '\0').
                 char == 0xFF EOS marker
                         Note that it is legal for 0xff to be a valid
                         string initalization byte.  i.e.
                            char a[4] = "a\377b";
                         has a INITSTR of
                            0x0e==INITSTR 0x03 0x61 0xff 0x62 0xff 0x0a==INITEND
              INITFUN (== 15) function declaration
           Note that substructures may be strung together to initialize
           complex variables like strucutres. i.e.
              struct { char a; char s[3]; int i; int j; } x = { 1, "ab", 1024 };
      2 -- FUNCTION (follows INITFUN), followed by:
              int  funord index of function name in symbol table
              int  num_nodes (number of nodes in CTEMP2 corresponding to
                   function)
              int  pstart (root node in tree processing)
              int  pbytes (number of bytes passed to function on stack)
              int  locoff (number of bytes to allocate on stack to locals)
              int  funline (if CHECKOUT file, line number of function)
      3 -- FLUSH, clear nodes (end of function)

   CTEMP2 is a list of nodes that form trees, and represent the "computational
   objects" (stuff that emits code) in the source program.  Nodes are are
   a set of 16-bit values, with the structures given in nodes.h.  Note that
   the first WORD of each node structure is actually two bytes, ntyp and
   dtyp (node type, data type).  The ntyp values and their associated
   structures are given in NODES.H.

   Nodes are numbered starting at 1 for each tree (logically, no node
   numbers area kept in the CTEMP2 file).  There are no "seperator" fields
   in CTEMP2.

   So, for example, the simple program

      int i, j;
      a() { i += j; }
      int b() { a(); return i; }
      main() { printf("b() = %d\n",b()); }

   would produce the following CTEMP1
    OFFSET  data <hex>                                        data <ascii>
    ======  ===============================================   ================
      0000  01 02 01 00 69 00 0C 02 00 01 02 02 00 6A 00 0C   ....i........j..
      0010  02 00 01 02 03 00 61 00 0F 02 03 00 10 00 0E 00   ......a.........
      0020  00 00 00 00 00 00 03 01 02 04 00 62 00 0F 02 04   ...........b....
      0030  00 13 00 11 00 00 00 00 00 00 00 03 01 02 05 00   ................
      0040  6D 61 69 6E 00 0F 01 02 06 00 70 72 69 6E 74 66   main......printf
      0050  00 0F 01 01 07 00 00 0E 09 62 28 29 20 3D 20 25   .........b() = %
      0060  64 0A FF 0A 02 05 00 1C 00 1A 00 00 00 00 00 00   d...............
      0070  00 03 00 00 00 00 00                              .......
   and CTEMP2
    OFFSET  data <hex>
    ======  ==============================================
      0000  0200  0001  0000  0000  0200  0002  0000  0000
      0010  0032  0001  0005  000B  0000  000F  0009  000C
      0020  FD00  0003  0000  0000  0216  0001  0000  0000
      0030  0200  0001  0000  0000  0203  0009  000B  000D
      0040  000F  0005  000F  FD00  0006  0000  0000  FE00
      0050  0007  0000  0000  FD00  0004  0000  0000  0216
      0060  0009  0000  0000  000F  000D  0005  0216  0001
      0070  0011  0000  000B  0000  000F  0014  0018

   Parsing CTEMP1 according to the rules presented above, we have:
      Offset 0:
         01        EXTERN
         02        visibility = SEXTERN
         0001      symbol table entry number
         "i" 00    symbol name
         0C 0002   INITDB, 2 bytes
      Offset 9:
         01        EXTERN
         02        visibility = SEXTERN
         0002      symbol table entry number
         "j" 00    symbol name
         0C 0002   INITDB, 2 bytes
      Offset 0x12:
         01        EXTERN
         02        visibility = SEXTERN
         0003      symbol table entry number
         "a" 00    symbol name
         0F        INITFUN
      Offset 0x19:
         02        FUNCTION
         0003      funord (index of function name in symbol table)
         0010      num_nodes (number of nodes in function tree)
         000E      pstart (root node in tree parsing)
         0000      pbytes (number of bytes passed to function on stack)
         0000      locoff (number of bytes to allocate on stack to locals)
         0000      funline (if CHECKOUT file, line number of function)
      Offset 0x26:
         03        FLUSH
      Offset 0x27:
         01        EXTERN
         02        visibility = SEXTERN
         0004      symbol table entry number
         "b" 00    symbol name
         0F        INITFUN
      Offset 0x2e:
         02        FUNCTION
         0004      funord (index of function name in symbol table)
         0013      num_nodes (number of nodes in function tree)
         0011      pstart (root node in tree parsing)
         0000      pbytes (number of bytes passed to function on stack)
         0000      locoff (number of bytes to allocate on stack to locals)
         0000      funline (if CHECKOUT file, line number of function)
      Offset 0x3c:
         03        FLUSH
      Offset 0x3d:
         01        EXTERN
         02        visibility = SEXTERN
         0005      symbol table entry number
         "main" 00 symbol name
         0F        INITFUN
      Offset 0x46:
         01        EXTERN
         02        visiblity = SEXTERN
         0006      symbol table entry number
         "printf" 00  symbol name
         0F        INITFUN
      Offset 0x52:
            01     EXTERN
            01     visibility = SSTATIC
            0007   symbol table entry number
            00     symbol name == NULL (compiler generated symbol)
            0E     INITSTR
      Offset 0x58:
               09  STRLEN
               "b() = %d" 0A  string (not \0 terminated)
               FF  End-Of-String
      Offset 0x63:
         0A        INITEND
      Offset 0x64:
         02        FUNCTION
         0005      funord (index of function name in symbol table)
         001C      num_nodes (number of nodes in function tree)
         001A      pstart (root node in tree parsing)
         0000      pbytes (number of bytes passed to function on stack)
         0000      locoff (number of bytes to allocate on stack to locals)
         0000      funline (if CHECKOUT file, line number of function)
      Offset 0x71:
         03        FLUSH
      Offset 0x72:
         00        EOF
         0000      nwarnings
         0000      usage

   Parsing CTEMP2 according to the rules presented above, we have:
      Offset 0:
         Node #:
             1    00 02  0001  0000  0000
                     ntyp == OPND
                     dtyp == CINT
                     varname == 1 (index of "i" in symbol table)
                     offset == 0 (global variable)
                     in_union == 0
             5    00 02  0002  0000  0000
                     ntyp == OPND
                     dtyp == CINT
                     varname == 2 (index of "j" in symbol table)
                     offset == 0 (global variable)
                     in_union == 0
             9    32 00  0001  0005
                     ntyp == AADD
                     dtyp == NULL (not used)
                     leftnode == 1 (OPND reference to "i")
                     rightnode == 5 (OPND reference to "j")
          0x0c    0B 00  0000
                     ntyp == RET
                     dtyp == NULL (not used)
                     nodevalue == NULL (nothing returned)
          0x0e    0F 00  0009  000C
             pstart (root node) of tree for function a()
                     ntyp == LST (2 nodes)
                     dtyp == NULL (not used)
                     nodes2[0] == 9 (AADD node above)
                     nodes2[1] == 0x0c (null RET above)

      Offset 0x20:
         Node #:
             1    00 FD  0003  0000  0000
                     ntyp == OPND
                     dtyp == FUNCTION
                     varname == 3 (index of "a" in symbol table)
                     offset == 0 (global)
                     in_union == 0
             5    16 02  0001  0000  0000
                     ntyp == CALL
                     dtyp == CINT (default for non-typed function)
                     fnnamenode == 1 (OPND reference to "a")
                     argsnode == NULL (no arguments)
                     strctlen == 0 (no structure returned)
             9    00 02  0001  0000  0000
                     ntyp == OPND
                     dtyp == CINT
                     varname == 1 (index of "i" in symbol table)
                     offset == 0 (global)
                     in_union == 0
          0x0d    03 02  0009
                     ntyp == CAST
                     dtyp == CINT
                     nodenum == 9 (OPND reference to "i")
                        Make sure type of return is correct
          0x0f    0B 00  000D
                     ntyp == RET
                     dtyp == NULL (not used)
                     valuenode == 0x0d (CAST node above)
                        Return properly typecast'ed value
          0x11    0F 00  0005  000F
             pstart (root node) of tree for function b()
                     ntyp == LST (2 nodes)
                     dtyp == NULL (not used)
                     nodes2[0] == 5 (CALL node above)
                     nodes2[1] == 0x0f (RET node above)

      Offset 0x46:
         Node #:
             1    00 FD  0006  0000  0000
                     ntyp == OPND
                     dtyp == FUNCTION
                     varname == 6 (index of "printf" in symbol table)
                     offset == 0 (global)
                     in_union == 0
             5    00 FE  0007  0000  0000
                     ntyp == OPND
                     dtyp == ARRAY
                     varname == 7 (index of "b() = %d" 0A in symbol table)
                     offset == 0 (global)
                     in_union == 0
             9    00 FD  0004  0000  0000
                     ntyp == OPND
                     dtyp == FUNCTION
                     varname == 4 (index of "b" in symbol table)
                     offset == 0 (global)
                     in_union == 0
          0x0d    16 02  0009  0000  0000
                     ntyp == CALL
                     dtyp == CINT
                     fnnamenode == 9 (OPND reference to "b")
                     argsnode == NULL (no arguments)
                     strctlen == 0 (no structure returned)
          0x11    0F 00  000D  0005
                     ntyp == LST (2 nodes)
                     dtyp == NULL (not used)
                     nodes2[0] == 0x0d (CALL "b" node above)
                     nodes2[1] == 5 (OPND ARRAY node above)
          0x14    16 02  0001  0011  0000
                     ntyp == CALL
                     dtyp == CINT (default for non-typed function)
                     fnnamenode == 1 (OPND reference to "printf")
                     argsnode == 0x11 (LST for 2nodes, we have 2 arguments)
                     strctlen == 0 (no structure returned)
          0x18    0B 00  0000
                     ntyp == RET
                     dtyp == NULL (not used)
                     nodevalue == NULL (nothing returned)
          0x1a    0F 00  0014  0018
             pstart (root node) of tree for function main()
                     ntyp == LST (2 nodes)
                     dtyp == NULL (not used)
                     nodes2[0] == 0x14 (CALL "printf" node above)
                     nodes2[1] == 0x18 (RET node above)

#endif
