/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  DeSmet C is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; either version 2 of the License, or any
 *  later version.
 *
 *  DeSmet C is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/* PASS1.H            INCLUDE FOR PASS1 */
/*
 *  Modification history:
 *    William Couture    10/12/05   Reformat/cleanup/comments
 *                                  changed 'heir' magic numbers to names
 *    William Couture    12/28/05   Change skipl() to SKIPL() so it's more
 *                                  obvious that it is a macro.
 */
/*
 *  Note the discussion of the C88 output files (CTEMP1, CTEMP2) at the
 *  end of this file.
 */

#define CHECK   1            /* true if generating checkout compiler */
#define LIMITED 0

/* LTYPE VALUES */

#define LETTER 1
#define DIGIT  2
#define SPACE  3

#define DEFCPY  251  /* copy macro value literal */
#define DEFPAS  252  /* #define concatination */
#define DEFSFY  253  /* #define stringify */
#define DEFSTR  254  /* #define <text> */
#define DEFEND  255  /* end of #define expansion */

/* symbol table types */

/* Note:  These structures are used with a "DeSmet-ism":  the pointer
 *        does not need to be a pointer to the structure type, the member
 *        reference is enough to create the correct memory reference.
 *
 *        For example:
 *            char *nameat;
 *        (from pass1.h)
 *            nameat->nchain = 0;
 *        (from newname() in c6.c, using the "OPERAND" structure)
 */

#define RESERVED    1       /* BYTE CONTAINING RESERVED VALUE */
    struct
       {
       char rescl, rvalue;
       };

#define STAG        2       /* forward chain, size of structure */
    struct
       {
       char stagcl;
       int  schain,   /* forward chain of members, type really should be
                       * char *
                       */
            staglen;  /* len (size) of structure */
       };

#define DEFINED     3       /* #args, (arg # or DEFSTR chars... LF)
                                ...DEFEND */
    struct
       {
       char defcl,    /* define class */
            *dchain,  /* forward chain of members */
            dnlen,    /* len */
            dargs,
            dval;
       };

#define DEFPARM     4       /* number of argument */
    struct
       {
       char dpcl,      /* class */
            dpnum;     /* number of this #define argument */
       };

#define OPERAND    26       /* op class, forward chain of members,
                             length of name, storage, offset or ordinal */
    struct
       {
       char opcl,     /* operator class */
            *nchain,  /* forward chain of members */
            nlen,     /* length of name + EOS */
            nstor;    /* storage (SEXTONLY, SSTATIC, SAUTO, SPARM, STYPEDEF, etc) */
       int  noff;     /* offset / ordinal */
       char ntype[1];
       };


/* named heirarchy values */
#define HEIR_COMMA     11  /* , */
#define HEIR_OPEQ      12  /* %= &= *= += -= /= <<= >>= ^= |= = */
#define HEIR_CONDITIONAL 13  /* ? : */
#define HEIR_LOG_OR    14  /* || */
#define HEIR_LOG_AND   15  /* && */
#define HEIR_BIT_OR    16  /* | */
#define HEIR_BIT_XOR   17  /* ^ */
#define HEIR_BIT_AND   18  /* & */
#define HEIR_EQUALITY  19  /* != or == */
#define HEIR_COMPARE   20  /* < or <= or > or >= */
#define HEIR_SHIFT     21  /* << or >> */
#define HEIR_ADDSUB    22  /* + or - */
#define HEIR_MULDIV    23  /* * or / or % */
#define HEIR_PREFIX    24  /* ! or ++<var> or --<var> */
#define HEIR_STRUCTREF 25  /* -> or . */

#define CONSTANT    27
#define LCONSTANT   28
#define FFCONSTANT  29
#define FDCONSTANT  30
#define STRNG       31
#define UNDEF       99
#define STREF      100

/* storage classes. in nstor field. */

#define SEXTONLY    0
#define SSTATIC     1
#define SEXTERN     2
#define SUNSTOR     3
#define SMEMBER     4
#define SAUTO       5
#define SPARM       6
#define STYPEDEF    7

#define LF          10
#define CR          13
#define CONTZ       26

/* the reserved words */
/* v3.1 03/08/88 - resorted reserved word in ANSI order */

#define RAUTO        1
#define RBREAK       2
#define RCASE        3
#define RCHAR        4
#define RCONST       5
#define RCONTINUE    6
#define RDEFAULT     7
#define RDO          8
#define RDOUBLE      9
#define RELSE       10
#define RENUM       11
#define REXTERN     12
#define RFLOAT      13
#define RFOR        14
#define RGOTO       15
#define RIF         16
#define RINT        17
#define RLONG       18
#define RNOALIAS    19
#define RREGISTER   20
#define RRETURN     21
#define RSHORT      22
#define RSIGNED     23
#define RSIZEOF     24
#define RSTATIC     25
#define RSTRUCT     26
#define RSWITCH     27
#define RTYPEDEF    28
#define RUNION      29
#define RUNSIGNED   30
#define RVOID       31
#define RVOLATILE   32
#define RWHILE      33

/* predefined macro names */

#define R_LINE      34
#define R_FILE      35
#define R_DATE      36
#define R_TIME      37
#define R_STDC      38

/* extended keywords + MSC keywords */

#define RASM        39
#define RCDECL      40
#define RFAR        41
#define RFORTRAN    42
#define RHUGE       43
#define RINTERRUPT  44
#define RNEAR       45
#define RPASCAL     46
#define RDOTDOT     47
#define RELIPSE     48

/* NOTE: This structure is defined so that it has the same size as
 *       the following data area, so that the compile context can be
 *       saved when processing an include file.
 *
 *       However, it has the contents of the data area out of order
 *       (*cur is after *lineBeg and *lineEnd, but *_cur is before).
 *       This does not make a difference, fortunately.
 *
 *       It is important that the data area remain in a single block.
 *       Later I may get around to makeing it an actual structure.
 */
struct savearea
   {
   char _name[65];
   char *_cur, *_lineBeg, *_lineEnd, *_bufEnd, *_eofptr;
   unsigned _cline, _lineinc, _file;
   char _lineBuf[512], _fileBuf[2048];
   };
/* Note: The savearea does not include savEnd[2].  This is used by the
 *       parser to include fence values (LF & Ctrl-Z) to flag the need
 *       to re-load fileBuf[].  However, once these values are set, they
 *       are never cleared, so "popping" a full fileBuf[] will not overrun
 *       due to lack of fence values.
 *
 *       I wonder, though, if I can get the buffer logic confused, so that
 *       before the "push", savEnd[] was { LF, CTRL_Z }, but it ends up
 *       as { CTRL_Z, ... } after the "pop".  Would the compiler think it
 *       encountered the end of the program?
 */

#define SAVSIZE sizeof(struct savearea)

char name[65];
char *lineBeg,  /* ptr to beginning of line in buffer */
     *lineEnd,
     *cur,  /* pointer to "next" character to be parsed */
     *bufEnd,
     *eofptr;
unsigned cline, lineinc, file;
char lineBuf[512], fileBuf[2048], savEnd[2];
  /* Note:  lineBuf[] *MUST* preceede fileBuf[]!  In dolf(), the compiler uses
  *         fileBuf[-lsiz] to address a line at the end of lineBuf[] (i.e.,
  *         fileBuf[] is used as a "end of buffer + 1" pointer).
  */
  /* Note:  savEnd[2] is never used as storage.  Instead it is a "high water
   *        mark" after fileBuf[] to indicate that it's time to read a fresh
   *        fileBuf[].
   *
   *        Also, it is used as a "seperator" to flag parsing file input
   *        versus parsing a #define expansion.  If cur > savEnd, then
   *        a macro expansion is being parsed.
   */

char macExp[2048], macEnd[1], *macxp;

char trigraphs;  /* process trigraphs */
unsigned lastline;

unsigned tree,
         control,
         ntree,  /* number of nodes in this tree */
         *intree;  /* pointer into treebuf[] */
unsigned maxmem,minmem,hiwater;
int  last,retnum;
int  *wp,     /* pointer to "WORD" (INT, 2-byte, 16-bit) value */
     findover, /* start of symbol table for local symbols */
               /* NOTE: this should be char *findover; */
     nerr,    /* number of errors */
     nwarn,   /* number of warnings */
     locoff,
     locplus;
int  ordinal; /* index number of item in symbol table.  pre-inc, so indices
               * start at 1.
               */
int  pbytes,
     wvalue;  /* WORD (16-bit int) equivalent of dvalue (value found by number()) */
char curch,   /* current character being parsed */
     *tokat,  /* starting location of current token */
     heir,    /* "heirarchy" of current token (nameat->opcl == RESERVED, OPERAND, etc) */
     bvalue,
     *funname,
     *original,  /* set if a preveious definition i.e. extern i; int i; */
     *mfree,
     *hash[33];  /* table of pointers to symbol chains in memory */
                 /* hash[32] is chain of labels */
char c_type[1],int_type[1],u_type[1],l_type[1],ff_type[1],fd_type[1],p_type[2];
char treebuf[512],ctlbuf[512];
char string[257],  /* current token */
     nested,
     *bptr,  /* general use pointer to byte */
     i,
     j,
     ininit,
     was_ext;

char *addat,
     *addtype,
        /* Note:  addtype must be beyond first 255 bytes of memory, because
         *        a value <= 255 is taken to be a data storage type (CCHAR,
         *        CINT, etc), not a pointer.
         */
     addstor,  /* data type of "storage" being added (SEXTERN, SSTATIC, etc) */
     *addloc,  /* ptr to fill location in ntype[] in symbol table */
     *addparm,  /* ptr to params in function declaration -- i.e. points to
                 * "i" in x(i) int i; {}
                 */
     *addproto,  /* ptr to prototypes in function declaration -- i.e. points
                  * to "int i" in x(int i){}
                  */
     hashno,  /* token hash: (sum of characters + len) & 0x1f */
     defhash,
     laststmt;

char *inctl,
     *initnode,
     *nameat,
       /* pointer to symbol table entry in memory:
        *             xxyy         hash link
        *             'abc...'00   symbol (0 terminated, length varies)
        *   nameat->  cc           class (1=RESERVED KEYWORD, 2=STRUCTURE/UNION,
        *                                 3=DEFINE, 4=PARAMETER, 26=OPERAND)
        *             ....         varies depending on class, above.
        *
        * For RESERVED, the entry is:
        *             xxyy         hash link
        *             'abc...'00   symbol (0 terminated, length varies)
        *   nameat->  01           rescl (class) = RESERVED
        *             wxyz         reserved value (keyword index)
        *
        * For OPERAND (variable or function), the entry is:
        *             xxyy         hash link
        *             'abc...'00   symbol (0 terminated, length varies)
        *   nameat->  1A           opcl (class) = OPERAND
        *             0000         nchain (initial value is 0)
        *             ll           strlen of symbol, including EOS
        *             ss           nstor == storage (SEXTONLY, SSTATIC, SAUTO, SPARM, STYPEDEF, etc)
        *             iiii         noff (index in symbol table)
        *             cc           ntype[x] values, length varies default is
        *                          ntype[0] == CINT
        */
     in_stru,  /* currently parsing a strucutre declaration */
     funtype,
     longconst;

char is_big,  /* large memory model */
     eopt,  /* tokens in pass 1, debug info in pass 2 */
     istring[40],  /* include path, if any */
     sopt,  /* symbols in pass 1, tree in pass 2 */
     soopt,  /* syntax check only */
     topt,  /* temporaries drive */
     xopt,  /* do not chain, do not delete temporaries */
     wopt,  /* show warnings */
     have_asm,  /* flag: have #asm in file, must compile as if 'a' option given.
                 * i.e. GEN does not compile directly to .O
                 */
     see_exit;

long dvalue,   /* long ("double") (32-bit) value found by number() */
     constexp();
unsigned fvalue[4],
         dsize(),  /* returns bytes in data type */
         fun_ret_len;  /* if function returns a structure, this is non-zero == sizeof(structure) */
int  objnum;  /* handle to .O file for CHKOUT info */
char *inobj,
      newline,
      aopt,  /* compile to assembly */
      copt,  /* add checkout info to object file */
      objname[65];
extern char ltype[256];
   /* increased from 128 to 256 so "extended ASCII" in source file will be a
    * valid index into this array.  WAC 11/2/05
    */
#if CHECK
char objbuf[512];  /* buffer for CHKOUT info */
#endif

#define MAXINC  20
unsigned incPara[MAXINC];
char incnext;

#define MAXNEST 60
int  nestfrom[MAXNEST];  /* individual pieces of a #define expansion */

struct stag
   {
   char cstype,
        *stat;
   int  staglen;
   } stagat;

struct
   {
   char sttype;
   struct stag *ststagat;
   } *struct_type;

union
   {
   char byte;
   int word;
   } ;  /* for "DeSmet-ism" such as addat->byte or addat->word */

struct
   {
   unsigned loword,
            hiword;
   } ;  /* for "DeSmet-ism" of ((void *)&dvalue)->hiword;
         * Note that loword is never used.
         */

char mdate[12],
     mtime[9],
     mactokn,  /* flag: currently parsing an #if expression */
     *macname[MAXNEST],
     *macfrom[MAXNEST],
     *machash[32],
     macOp,  /* last #define expansion op.  Although only used in addnest(),
              * this is a global to keep its value between addnest() levels
              * (addnest() is recursive)
              */
     blevel,  /* "brace level", if non-zero (in function), do ??? with #define & hash[32] */
     *anon;  /* ptr to list of "anonymous" struct/union definitions (i.e. struct { int i; } j;) */
char *protohash[32],  /* hash pointers for prototype matching */
     *labhash[32],  /* hash pointers for labels */
     xkwd,  /* recognize extended keywords */
     __stdc;  /* only standard C */

char *lastname, *macproto;

struct pargs
   {
   struct pargs * plink;
   char ptype[1];
   };

#define SKIPL() cur = lineEnd - 1

#ifdef MacHiWater

char *MacHiWaterP, maxNest;

#endif


#ifdef NEVERDEF

   The output of the C88 compiler is two files, CTEMP1 and CTEMP2 which
   are passed to GEN to finish the compilation process.

   CTEMP1 is the "control" file, which has the symbol information and
   tells GEN which objects to compile.

   CTEMP2 is the "nodes" file, which has the compuational information
   stored in a tree structure.

   Note that string constants are stored in CTEMP1 (i.e. char str[4]="abc";)
   while numeric constants have their names stored in CTEMP1 with pointers
   into CTEMP2 which contains the numeric value (i.e. int i = 1;)

   CTEMP1 is a list of control structures, each starting with a single
   byte identifying the structure type.  They are:
      0 -- EOF  followed by 2 ints:
               int nwarnings
               int usage (percentage)
      1 -- EXTERN (variable name) followed by:
              char visibility SEXTERN == 2, global visibility (int i;)
                   (nstor)    SSTATIC == 1, only visible in this module
                                                (static int i;)
                              SEXTONLY == 0, global visibility from another
                                             module (extern int i;)
              int  symbol table entry number (starts at 1)
              char*N symbol name, '\0' terminated.
           additional substructures may follow, starting with a single byte
           to identify the substructure:
              INITEND (== 10) end of variable initialization
              INITDB  (== 11) followed by
                 int number of bytes, 0 initialization
              INITRB  (== 12) followed by
                 int number of bytes, no initialization
              INITVAL (== 13) numeric initialization, followed by
                 char dtyp (CCHAR, CINT, CUNSG, CLONG, CFLOAT, CDOUBLE)
                 int  "node to keep" -- where to chop off tree to remove
                      one-time init value from function tree
                 int  nodes_needed (node in current tree in CTEMP2 to read up
                      to for constant expression)
                 int  start_node (node at which to start tree processing)
              INITSTR (== 14) string constant, followed by
                 char strlen  (bytes in string)
                 char*n string value (does *NOT* include normal C EOS '\0',
                        though it may include embedded '\0' padding.  i.e.
                        char a[3] = "ab" will have a string value of "ab",
                        since the 3rd byte would be the normal EOS '\0', but
                        char a[4] = "ab" will have a string value of "ab\0"
                        to pad the string to LEN-1, leaving room for the
                        normal EOS of '\0').
                 char == 0xFF EOS marker
                         Note that it is legal for 0xff to be a valid
                         string initalization byte.  i.e.
                            char a[4] = "a\377b";
                         has a INITSTR of
                            0x0e==INITSTR 0x03 0x61 0xff 0x62 0xff 0x0a==INITEND
              INITFUN (== 15) function declaration
           Note that substructures may be strung together to initialize
           complex variables like strucutres. i.e.
              struct { char a; char s[3]; int i; int j; } x = { 1, "ab", 1024 };
      2 -- FUNCTION (follows INITFUN), followed by:
              int  funord index of function name in symbol table
              int  num_nodes (number of nodes in CTEMP2 corresponding to
                   function)
              int  pstart (root node in tree processing)
              int  pbytes (number of bytes passed to function on stack)
              int  locoff (number of bytes to allocate on stack to locals)
              int  funline (if CHECKOUT file, line number of function)
      3 -- FLUSH, clear nodes (end of function)

   CTEMP2 is a list of nodes that form trees, and represent the "computational
   objects" (stuff that emits code) in the source program.  Nodes are are
   a set of 16-bit values, with the structures given in nodes.h.  Note that
   the first WORD of each node structure is actually two bytes, ntyp and
   dtyp (node type, data type).  The ntyp values and their associated
   structures are given in NODES.H.

   Nodes are numbered starting at 1 for each tree (logically, no node
   numbers area kept in the CTEMP2 file).  There are no "seperator" fields
   in CTEMP2.

   So, for example, the simple program

      int i, j;
      a() { i += j; }
      int b() { a(); return i; }
      main() { printf("b() = %d\n",b()); }

   would produce the following CTEMP1
    OFFSET  data <hex>                                        data <ascii>
    ======  ===============================================   ================
      0000  01 02 01 00 69 00 0C 02 00 01 02 02 00 6A 00 0C   ....i........j..
      0010  02 00 01 02 03 00 61 00 0F 02 03 00 10 00 0E 00   ......a.........
      0020  00 00 00 00 00 00 03 01 02 04 00 62 00 0F 02 04   ...........b....
      0030  00 13 00 11 00 00 00 00 00 00 00 03 01 02 05 00   ................
      0040  6D 61 69 6E 00 0F 01 02 06 00 70 72 69 6E 74 66   main......printf
      0050  00 0F 01 01 07 00 00 0E 09 62 28 29 20 3D 20 25   .........b() = %
      0060  64 0A FF 0A 02 05 00 1C 00 1A 00 00 00 00 00 00   d...............
      0070  00 03 00 00 00 00 00                              .......
   and CTEMP2
    OFFSET  data <hex>
    ======  ==============================================
      0000  0200  0001  0000  0000  0200  0002  0000  0000
      0010  0032  0001  0005  000B  0000  000F  0009  000C
      0020  FD00  0003  0000  0000  0216  0001  0000  0000
      0030  0200  0001  0000  0000  0203  0009  000B  000D
      0040  000F  0005  000F  FD00  0006  0000  0000  FE00
      0050  0007  0000  0000  FD00  0004  0000  0000  0216
      0060  0009  0000  0000  000F  000D  0005  0216  0001
      0070  0011  0000  000B  0000  000F  0014  0018

   Parsing CTEMP1 according to the rules presented above, we have:
      Offset 0:
         01        EXTERN
         02        visibility = SEXTERN
         0001      symbol table entry number
         "i" 00    symbol name
         0C 0002   INITDB, 2 bytes
      Offset 9:
         01        EXTERN
         02        visibility = SEXTERN
         0002      symbol table entry number
         "j" 00    symbol name
         0C 0002   INITDB, 2 bytes
      Offset 0x12:
         01        EXTERN
         02        visibility = SEXTERN
         0003      symbol table entry number
         "a" 00    symbol name
         0F        INITFUN
      Offset 0x19:
         02        FUNCTION
         0003      funord (index of function name in symbol table)
         0010      num_nodes (number of nodes in function tree)
         000E      pstart (root node in tree parsing)
         0000      pbytes (number of bytes passed to function on stack)
         0000      locoff (number of bytes to allocate on stack to locals)
         0000      funline (if CHECKOUT file, line number of function)
      Offset 0x26:
         03        FLUSH
      Offset 0x27:
         01        EXTERN
         02        visibility = SEXTERN
         0004      symbol table entry number
         "b" 00    symbol name
         0F        INITFUN
      Offset 0x2e:
         02        FUNCTION
         0004      funord (index of function name in symbol table)
         0013      num_nodes (number of nodes in function tree)
         0011      pstart (root node in tree parsing)
         0000      pbytes (number of bytes passed to function on stack)
         0000      locoff (number of bytes to allocate on stack to locals)
         0000      funline (if CHECKOUT file, line number of function)
      Offset 0x3c:
         03        FLUSH
      Offset 0x3d:
         01        EXTERN
         02        visibility = SEXTERN
         0005      symbol table entry number
         "main" 00 symbol name
         0F        INITFUN
      Offset 0x46:
         01        EXTERN
         02        visiblity = SEXTERN
         0006      symbol table entry number
         "printf" 00  symbol name
         0F        INITFUN
      Offset 0x52:
            01     EXTERN
            01     visibility = SSTATIC
            0007   symbol table entry number
            00     symbol name == NULL (compiler generated symbol)
            0E     INITSTR
      Offset 0x58:
               09  STRLEN
               "b() = %d" 0A  string (not \0 terminated)
               FF  End-Of-String
      Offset 0x63:
         0A        INITEND
      Offset 0x64:
         02        FUNCTION
         0005      funord (index of function name in symbol table)
         001C      num_nodes (number of nodes in function tree)
         001A      pstart (root node in tree parsing)
         0000      pbytes (number of bytes passed to function on stack)
         0000      locoff (number of bytes to allocate on stack to locals)
         0000      funline (if CHECKOUT file, line number of function)
      Offset 0x71:
         03        FLUSH
      Offset 0x72:
         00        EOF
         0000      nwarnings
         0000      usage

   Parsing CTEMP2 according to the rules presented above, we have:
      Offset 0:
         Node #:
             1    00 02  0001  0000  0000
                     ntyp == OPND
                     dtyp == CINT
                     varname == 1 (index of "i" in symbol table)
                     offset == 0 (global variable)
                     in_union == 0
             5    00 02  0002  0000  0000
                     ntyp == OPND
                     dtyp == CINT
                     varname == 2 (index of "j" in symbol table)
                     offset == 0 (global variable)
                     in_union == 0
             9    32 00  0001  0005
                     ntyp == AADD
                     dtyp == NULL (not used)
                     leftnode == 1 (OPND reference to "i")
                     rightnode == 5 (OPND reference to "j")
          0x0c    0B 00  0000
                     ntyp == RET
                     dtyp == NULL (not used)
                     nodevalue == NULL (nothing returned)
          0x0e    0F 00  0009  000C
             pstart (root node) of tree for function a()
                     ntyp == LST (2 nodes)
                     dtyp == NULL (not used)
                     nodes2[0] == 9 (AADD node above)
                     nodes2[1] == 0x0c (null RET above)

      Offset 0x20:
         Node #:
             1    00 FD  0003  0000  0000
                     ntyp == OPND
                     dtyp == FUNCTION
                     varname == 3 (index of "a" in symbol table)
                     offset == 0 (global)
                     in_union == 0
             5    16 02  0001  0000  0000
                     ntyp == CALL
                     dtyp == CINT (default for non-typed function)
                     fnnamenode == 1 (OPND reference to "a")
                     argsnode == NULL (no arguments)
                     strctlen == 0 (no structure returned)
             9    00 02  0001  0000  0000
                     ntyp == OPND
                     dtyp == CINT
                     varname == 1 (index of "i" in symbol table)
                     offset == 0 (global)
                     in_union == 0
          0x0d    03 02  0009
                     ntyp == CAST
                     dtyp == CINT
                     nodenum == 9 (OPND reference to "i")
                        Make sure type of return is correct
          0x0f    0B 00  000D
                     ntyp == RET
                     dtyp == NULL (not used)
                     valuenode == 0x0d (CAST node above)
                        Return properly typecast'ed value
          0x11    0F 00  0005  000F
             pstart (root node) of tree for function b()
                     ntyp == LST (2 nodes)
                     dtyp == NULL (not used)
                     nodes2[0] == 5 (CALL node above)
                     nodes2[1] == 0x0f (RET node above)

      Offset 0x46:
         Node #:
             1    00 FD  0006  0000  0000
                     ntyp == OPND
                     dtyp == FUNCTION
                     varname == 6 (index of "printf" in symbol table)
                     offset == 0 (global)
                     in_union == 0
             5    00 FE  0007  0000  0000
                     ntyp == OPND
                     dtyp == ARRAY
                     varname == 7 (index of "b() = %d" 0A in symbol table)
                     offset == 0 (global)
                     in_union == 0
             9    00 FD  0004  0000  0000
                     ntyp == OPND
                     dtyp == FUNCTION
                     varname == 4 (index of "b" in symbol table)
                     offset == 0 (global)
                     in_union == 0
          0x0d    16 02  0009  0000  0000
                     ntyp == CALL
                     dtyp == CINT
                     fnnamenode == 9 (OPND reference to "b")
                     argsnode == NULL (no arguments)
                     strctlen == 0 (no structure returned)
          0x11    0F 00  000D  0005
                     ntyp == LST (2 nodes)
                     dtyp == NULL (not used)
                     nodes2[0] == 0x0d (CALL "b" node above)
                     nodes2[1] == 5 (OPND ARRAY node above)
          0x14    16 02  0001  0011  0000
                     ntyp == CALL
                     dtyp == CINT (default for non-typed function)
                     fnnamenode == 1 (OPND reference to "printf")
                     argsnode == 0x11 (LST for 2nodes, we have 2 arguments)
                     strctlen == 0 (no structure returned)
          0x18    0B 00  0000
                     ntyp == RET
                     dtyp == NULL (not used)
                     nodevalue == NULL (nothing returned)
          0x1a    0F 00  0014  0018
             pstart (root node) of tree for function main()
                     ntyp == LST (2 nodes)
                     dtyp == NULL (not used)
                     nodes2[0] == 0x14 (CALL "printf" node above)
                     nodes2[1] == 0x18 (RET node above)

#endif
