/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  DeSmet C is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; either version 2 of the License, or any
 *  later version.
 *
 *  DeSmet C is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/*
 *  Modification history:
 *    William Couture    11/18/05   Reformat/cleanup/comments
 */
#define OP      0       /* on if want optimize code     */
#define DEBUG   1       /* on if want debug code        */
#define CHECK   1       /* on if want checkout option   */
#define LIMITED 0       /* on for limited version       */

/* vtype contains an operand */

#define CONSTV  0
#define OFFV    1
#define REGV    2
#define SEGRV   3
#define VARV    4
#define FLOATV  5

/* the fields of vtype */

#define VIS     0  /* CONST, OFFV, REGV, SEGRV, VARV, FLOATV */
#define VVAL    1  /* if REGV, this is AX, CX, DX, BX, etc */
                   /* if CONST, this is (low word of) value */
#define VNAME   2  /* name index in symbol table (0 for local) */
#define VOFF    3  /* offset (from BP or base address) */
#define VMORE   4  /* 4th word of CONST value (FP double) */
#define VT      5  /* contains DTYP */
#define VFROM   6

/* vtype[VT] is CCHAR, CINT, CUNSG, CLONG, BITS+, CFLOAT, CDOUBLE, 
 *           OR CLABEL.
 */
/* vtype[VSEG] (VARV only) is ES,CS,SS,DS or -1 (default) prefix. */
/* NOTE: I think he meant vtype[VMORE] (VARV only) is ES,CS,SS,DS or -1 (default) prefix. */
/*

vtype[VIS==0]  [VVAL==1]      [VNAME==2]     [VOFF==3]     [VMORE==4]

CONSTV         value                         constant. value follows in as many words
                                             as needed for the type.
OFFV                                         same as VARV except offset of.
REGV           value                         byte or word register.
SEGRV          value                         segment register.
VARV           8             name    value   name[value].
               0              or             name[BX+SI+value].
               1              0              name[BX+DI+value].
               2                             [BP+SI+value].
               3                             [BP+DI+value].
               4                             name[SI+value].
               5                             name[DI+value].
               6                             [BP+value].
               7                             name[BX+value].
 Note:  the values 0 .. 7 map to the R/M values in the EA addressing byte
        of the x86.
        EA byte:  MOD REG R/M
                  00  ... 000 [BX+SI]
                  00  ... 001 [BX+DI]
                  00  ... 010 [BP+SI]
                  00  ... 011 [BP+DI]
                  00  ... 100 [SI]
                  00  ... 101 [DI]
                  00  ... 110 [address]
                  00  ... 111 [BX]
                  01  ... 000 [BX+SI+bb]
                  01  ... 001 [BX+DI+bb]
                  01  ... 010 [BP+SI+bb]
                  01  ... 011 [BP+DI+bb]
                  01  ... 100 [SI+bb]
                  01  ... 101 [DI+bb]
                  01  ... 110 [BP+bb]
                  01  ... 111 [BX+bb]
                  10  ... 000 [BX+SI+wwww]
                  10  ... 001 [BX+DI+wwww]
                  10  ... 010 [BP+SI+wwww]
                  10  ... 011 [BP+DI+wwww]
                  10  ... 100 [SI+wwww]
                  10  ... 101 [DI+wwww]
                  10  ... 110 [BP+wwww]
                  10  ... 111 [BX+wwww]
            bb = signed BYTE offset following EA byte
          wwww = WORD offset following EA byte
        If MOD == 11, both REG and R/M are register indices (BYTE/WORD
        depends on instruction byte)

        VVAL == 8 is not a valid R/M value, and tells GEN to generate
        a direct address (i.e. invoke the special x86 case of MOD == 0
        and R/M == 110, since there is no [BP] addressing mode.)
FLOATV                              the floating value is on the float stack.

for the BIG model and the REGV, VARV and OFFV VIS types, the VMORE field is
used to indicate the segment or segment register:
VMORE:
   -1       normal
   0-7      AX thru DI holds the ES value
    8       ES register
    9       DS register
   10       SS register
   11       CS register
   12       will need to use ES
*/
#define NEED_ES      12
#define CFOREIGN    251
#define CSEGMENT    252


/* registers */

#define AX   0
#define CX   1
#define DX   2
#define BX   3
#define SP   4
#define BP   5
#define SI   6
#define DI   7
#define ES   8
#define DS   9
#define SS  10
#define CS  11

#define AL   0
#define CL   1
#define DL   2
#define BL   3
#define AH   4
#define CH   5
#define DH   6
#define BH   7

/* paterns for register clases */
/* see regpat[] */
#define ESPAT       256  /* 1 0000 0000 */
#define ANYPAT      207  /*   1100 1111 */
#define NOTCXPAT    205  /*   1100 1101 */
#define INDPAT      192  /*   1100 0000 */
#define SIPAT        64  /*   0100 0000 */
#define DIPAT       128  /*   1000 0000 */
#define BYTEPAT      15  /*   0000 1111 */
#define BXPAT         8  /*   0000 1000 */
#define DXPAT         4  /*   0000 0100 */
#define AXCXPAT       3  /*   0000 0011 */
#define CXPAT         2  /*   0000 0010 */
#define AXPAT         1  /*   0000 0001 */

/* names of the builtin calls */

#define _SWITCH     0
#define _CMP4       1
#define _SHR4       2
#define _SHL4       3
#define _MUL4       4
#define _DIV4       5
#define _MOD4       6
#define _FLOADD     7  /* load double precision number */
#define _FLOADE     8  /* load single precision number */
#define _FLOADL     9  /* load long from DX:DX into floating point accum */
#define _FSTORED   10  /* store double precision number */
#define _FSTOREE   11  /* store single precision number */
#define _FSTOREL   12  /* FP to long in DX:AX */
#define _FADD      13  /* FP ADD: TOS-1 = TOS-1 + TOS */
#define _FSUB      14  /* FP SUB: TOS-1 = TOS-1 - TOS */
#define _FMUL      15  /* FP MUL: TOS-1 = TOS-1 * TOS */
#define _FDIV      16  /* FP DIV: TOS-1 = TOS-1 / TOS */
#define _FCMP      17  /* compare TOS-1 to TOS, clean both from stack, return flags */
#define _FNEG      18  /* FP Negate: TOS = -TOS */
#define _FIS       19  /* return NZ if TOS is non-zero, clear TOS */
#define _FNOT      20  /* return 1 if TOS is 0, return flags, clear TOS */
#define _FDEC      21  /* FP DEC:  TOS = TOS - 1 */
#define _FINC      22  /* FP INC:  TOS = TOS + 1 */
#define _FPUSH     23  /* Push a double onto stack */
#define _FXCH      24  /* exchange FAC and FAC+1 */
#define _FCLEAR    25  /* clear element from floating stack */
#define _MOVE      26  /* move memory */


/* globals for pass2 of Medium C Compiler  */

char aopt,  /* A option: leave .A file, do not invoke assembler */
     copt,  /* C option: build CHKOUT info */
     eopt,  /* Ve option: display debug info */
     mopt,  /* M option: Intel (Microsoft) .OBJ files */
     sopt,  /* Vs option: display tree */
     topt,  /* T option: drive for temporaries */
     xopt,  /* X option: do not delete temporaries, do not invoke assembler */
     zopt,  /* Z option: compile directly to object code, not assembly (DEFAULT from C88) */
     subn[64],  /* number of subnodes, indexed by node types from NODES.H */
     *nextext;
char optype[64],  /* type of node, indexed by node types from NODES.H */
     commut[64],  /* does this node type commute?  indexed by node types from NODES.H */
     flipped[64],
     prefer_si,
     in_post;
char regpat[12],  /* {  1,  2,  4,  8, 16, 32, 64, 128, 256, 512, 1024, 2048 } */
                  /*   AX  CX  DX  BX  SP  BP  SI   DI   ES   DS    SS    CS */
     numswit,  /* number of switches */
     asmname[65],
     objname[65],
     nopre,
     nopost,
     isasgn;
int  nodesin,
     nextlab,
     lastlab,
     numlit;
int  funord,
     treelen,
     pstart,
     pbytes,
     level;
#if DEBUG
int  namet[8],  /* names of data types */
     namen[64];  /* names of operands */
#endif
int  jumps[11],  /* actually, this should be char *jumps[11] */
     regname[12],  /* actually, this should be char *regname[12] */
     reglow[4],  /* actually, this should be char *reglow[4] */
     reghigh[4];  /* actually, this should be char *reghigh[4] */
int  *regat[13],  /* indexed by register, pointers to vtype[] of expression using this reg */
     bltname[27],  /* actually, this should be char *bltname[27] */
     numfloat;
char regused[9],
     is_big,  /* large memory model */
     bltused[27];
unsigned util,
         locoff,  /* bytes allocated to locals */
         curoff,  /* current stack allocation for locals */
         line_num;

#define MAXFOR  16
int  numwhil,
     whilcont[MAXFOR],
     whilbrk[MAXFOR],
     switnext[MAXFOR];

#define MAXCASE 128
int  swits[MAXCASE],
     switto[MAXCASE];

#define BUFSIZE 512
int  treef,
     control,
     output;
int  *intree;  /* pointer into treebuf */
char *inctl,  /* pointer into ctlbuf */
     *inout,  /* pointer into outbuf */
     see_exit;
unsigned see_off,
         see_seg,
         see_ret,
         see_bp;
char see_msg[80],
     see_index;

int  ctlblk;  /* "block" (size == BUFSIZE) in CTEMP1 control file,
               * for "unparsing" in findstr()
               */
char treebuf[BUFSIZE],  /* tree (nodes) (CTEMP2) buffer */
     ctlbuf[BUFSIZE],  /* control info (CTEMP1) buffer */
     outbuf[BUFSIZE];  /* output (.O / .A) buffer */
extern char nextpgm[77];

#if OPT
#define NUMREC  10
/* bit mask for reg_dirty */
#define USED    1
#define DIRTY   2
extern char perfect;
char better,
     reg_dirty[9];
int  reg_from[9],
     reg_weight[9];
unsigned down,
         recat[NUMREC],
         recent[NUMREC],
         weight[NUMREC];
#endif

#define TREEMAX 8000
int  tree[TREEMAX];

#define EXTMAX  5000
char ext[EXTMAX];  /* variable visibility (SEXTERN, SSTATIC, SEXTONLY) & names */

#define EXTATMAX 500  /* number of variables */
int  numext,
     asmext;
char *extat[EXTATMAX];  /* pointers to variable visibility & names in ext[] */

union
   {
   int word;
   char byte;
   } ;  /* for "DeSmet-ism" such as addat->byte or addat->word */

#define SEXTONLY 0
#define SSTATIC 1
#define SEXTERN 2

#define CR 13
#define LF 10
